/* The SREhttp/2 content management system (CMS) program  

   This module will generate "output files" associated with a CMS dataseries
  1) first, it will provide a pick list of cms dataseries (for this host)
  2) given the chosen dataseries, it will present a form with checkboxes for each
     template
  3) generate output file(s), given the chosen templates

Todo:
  Some .CMS procs 

note on globals:

  INFO. = meta-data on this template
  OPTNAME. = options provided in the html request
  CINFO.  = record specific information (and some meta-data)
  REP_STUFF. = content of a repeat block, and temporary storage used in repetitive processing
  STUFF.   = contents of a CMS definition file
  GERRS.   = errors (found when processing template file)
  RECIDS.  = id7 and id of all records in the dataseries
  CSHDATA. = mid-sized cache of records and all their values    
*/

/******** User changeable parameters ***/
/* root directory for the SREhttp/2 CMS addon.
   Leave blank to use default (CMS subdirectory of srehttp/2 addon directory */
cmsdir=''


/******** end of user changeable parameters ***/


parse arg list,servername,verb,tempfile,,
          prog_file,reqnum,verbose,client_ip,privset,,
          uri,host_nickname,idinfo,aiter,attribs,seluse2
crlf='0d0a'x


if verb=" " then do
   say "This SREhttp/2 procedure is not meant to be run in standalone mode "
   exit
end  

signal on error name cmserr
signal on syntax name cmserr
signal on novalue name cmserr

verb=strip(translate(verb))
if verb='UPDATE' then do
   cmsdir=servername
   update_verb=1
end
else do
  update_verb=0
  addon_dir=sre_value('H2_ADDON_DIR_DEFAULT',,'SRE')
  if cmsdir='' then cmsdir=strip(addon_dir,,'\')||'\CMS'
end

foo=dosisdir(cmsdir)
if foo=0 then return 'String Error:unable to find CMS directory: ('cmsdir

cmsdir_data=cmsdir||'\DATA'
okk=dosisdir(cmsdir_data)
if okk=0 then do
  return 'string unable to find CMS data directory: 'cmsdir_data
end

cmsdir_lib=cmsdir||'\LIB'
okk=dosisdir(cmsdir_lib)
if okk=0 then do
  return 'string unable to find CMS lib directory: 'cmsdir_lib
end

cmsdir_pol=cmsdir||'\LIB\POL'
okk=dosisdir(cmsdir_pol)
if okk=0 then do
  return 'string unable to find CMS index directory: 'cmsdir_pol
end


cmsdir_files=cmsdir||'\FILES'
okk=dosisdir(cmsdir_files)
if okk=0 then do
  return 'string unable to find CMS input/output files directory: 'cmsdir_files
end

cmsdir_uploads=cmsdir||'\UPLOADS'
okk=dosisdir(cmsdir_uploads)
if okk=0 then do
  return 'string unable to find CMS upload files directory: 'cmsdir_uploads
end


cmsdir_index=cmsdir||'\INDEX'
okk=dosisdir(cmsdir_index)
if okk=0 then do
  return 'string unable to find CMS index directory: 'cmsdir_index
end

vers=sre_cache_data(' ','INFO',,'VERSION')
parse var vers . '0d0a'x aversion ; aversion=strip(translate(aversion,' ','09'x))
if aversion<1.19 then do
  return 'string SRE2003 version ('||c2x(aversion)||') is too old (must at least support SRE_DCSH version 1.19) '
end

parse var uri thisaddon '?' .
jj=lastpos('/',thisaddon)
mainaddon=left(uri,jj)||'cms.cmd'
root=left(uri,jj)


optname.=''
m=0
if list='' then list='action=1'


do until list=""
    parse var list opt1 '&' list
    m=m+1
    parse var opt1 avar '=' aval ; avar=strip(translate(avar)) 
    optname.0=optname.0||' '||avar
    aval=translate(aval,' ', '+')
    aval=sre_packur(aval)
    avar1='!!'avar
    if optname.avar1<>'' then
       optname.avar1=optname.avar1||' '||aval
    else
       optname.avar1=aval
    avar2='!'avar
    aval2=strip(translate(aval,' ','090a0d'x))
    optname.avar2=aval2
end


if update_verb=1 then do        /* check how_update */
     select
        when optname.!how_update=2 then update_verb=2     /* update or create (use instructions in template */
        when optname.!how_update=3 then update_verb=3     /* always create */
        otherwise  nop   /* retain value of 1 -- always update */
     end 
end 

/***** End of GENERIC CMS addon code *****/

if verb='UPDATE' then do
  update_files=0 ; update_bytes=0;
  optname.!action=3
end


if optname.!action='1' then do   /* find the list of dataseries defined for this host*/
  call show_datafiles
  foo=result
  return foo
end

if optname.!action='2' then do   /* choose a template */
  call ask_template
  foo=result
  return foo
end


if optname.!action='3' then do   /* generate output files */
  call generate_output
  foo=result
  return foo
end

if optname.!action='5' then do   
  call view_file
  foo=result
  return foo
end

if optname.!action='61' then do   
  call view_file_out 'TEXT'
  foo=result
  return foo
end

if optname.!action='6' then do   
  call view_file_out 
  foo=result
  return foo
end

return 'ERROR'

cmserr:
call sre_pmprintf("CMS_MAKE error at "sigl' rc='rc)
exit 


/*******************/
/* pick one of the current cms datafiles (for this host) */
show_datafiles:

call lineout1 '<html><head><title>CMS: Choose a CMS datafile</title> '


call lineout1 '<font size=-1><a href="'||mainaddon||'?action=1">Another CMS action</a></font><br>'

call lineout1 '<h3>Select a CMS dataseries</h3><em>from which you want to generate output files...</em><p> '

ahost=host_nickname

/* look for files, in cmsDIR_index dir, of the form: ahost_nnnn.DEF
   These .DEF files define "CMS indices" --- nnnn is a number from 0 to 9999
*/

cookie_value=sreh2_get_cookie('CMS') 

call lineout1 '<FORM name="" action="'||thisaddon||'" method="POST">'
call lineout1 '<input type="hidden" name="action" value="2">'
ff=cmsdir_index||'\'||ahost||'*.DEF'
goo=sysfiletree(ff,'ff2','FO')

call lineout1 'For host: <Tt>'||ahost||'</tt>, the following CMS dataseries have been created'


call lineout1 '<table border=1>'
do jj=1 to ff2.0
   daname=filespec('n',ff2.jj)
   call read_deffile ff2.jj
   call lineout1 '<tr><th align="left">' 
   if daname=cookie_value  | (cookie_value='' & jj=1) then
      call lineout1 '<INPUT TYPE="RADIO" checked name="deffile" value="'||daname||'">' 
   else
      call lineout1 '<INPUT TYPE="RADIO"  name="deffile" value="'||daname||'">' 
   gob=strip(stuff.!series)||'</th>'
   call lineout1 gob||'<td> <em>&nbsp;'||stuff.!desc||'</em></td>'
end

call lineout1 '</table>'


call lineout1 ''
call lineout1 '<p><INPUT name="sub2" TYPE="submit" VALUE="Choose a CMS dataseries">'
call lineout1 '</form>'
call lineout1 '</body> </html>'
call lineout1 '1a'x 
rcode=sre_command('FILE ETAG_AUTO EXPIRES 30s NOCACHE NOIF ERASE TYPE text/html NAME 'tempfile,,idinfo)
return rcode


/*******************/
/* pick template(s) to generate output files from */
ask_template:
stuff.=''

call lineout1 '<html><head><title>CMS: Which output to generate</title> '

call lineout1 '</head>'
call lineout1 '<font size=-1><a href="javascript:history.go(-1)">Back</a> || <a href="'||mainaddon||'?action=1">Another CMS action</a></font><br>'

call lineout1 '<h3>Generate output files</h3> '

daindx=optname.!deffile
ff=cmsdir_index||'\'||daindx
goo=stream(ff,'c','query exists')
if goo='' then do
   call lineout1 'Error: could not find index file: '||ff
   signal gooble
end

call read_deffile ff

astat=sreh2_set_cookie('CMS',daindx,1,root,,1) 

/* check for access rights */
rcode=check_access(stuff.!OWNERS,stuff.!SERIES)
if rcode<>1 then return rcode

a=sre_cache_data(stuff.!series,'OPEN  INDEX',stuff.!file)
if a>0 then do
   amess=sre_cache_data('OPEN','ERROR',a,stuff.!file)
   call lineout1 '<b>Error:</b> '||amess
   signal gooble
end


a=sre_cache_data(stuff.!series,'INFO',,'MODIFIED RECORDS VARS DESCRIPTION')
parse var a astat a ; astat=strip(astat)
if astat<>'0' then do
   amess=sre_cache_data('INFO','ERROR',astat)
   call lineout1 '<b>Error retrieving info:</b> '||amess
   signal gooble
end

parse var a . '0d0a'x . '09'x modified  '=' . '09'x nrecords '09'x  varlist '09'x  desc

call lineout1 '<FORM name="dogenerate" action="'||thisaddon||'" method="POST">'
call lineout1 '<input type="hidden" name="action" value="3">'
call lineout1 '<input type="hidden" name="deffile" value="'||daindx||'">'

call lineout1 '<em>For the CMS dataseries: </eM><b>'||stuff.!series||'</b><br>'
call lineout1 '<table>'
call lineout1 '<tr><th align="left">Last modified:</th><td>'||modified||'</td>'
call lineout1 '<tr><th align="left"># of records currently in dataseries:</th><td>'||nrecords||'</td>'
call lineout1 '<tr><th align="left">Description:</th><td>'||desc||'</td>'
call lineout1 '<tr><th align="left">Target directory:</th><td>'||stuff.!filedir||'</td>'

call lineout1 '</table>'

if stuff.!template='' then do
   call lineout1 '<b>No outputs specified</b> for this CMS dataseries '
   signal gooble

end 

call lineout1 '<br><table bgcolor="#a0a9ad"><tr><td>'
call lineout1 '<p>Select one, or more, output files to generate.<br>'

call lineout1 '<table border=1 cellpadding=5>'

tt=stuff.!template
do until tt=''
    parse var tt aline '0d0a'x tt
    parse var aline ofile tfile ',' desc ; tfile=strip(tfile)
    call lineout1 '<tr><td><input type="checkbox" name="tfile" value="'||tfile||'">'    
    hh='<a target="viewer" href="'||thisaddon||'?action=5&deffile='||optname.!deffile||'&tplfile='||tfile||'">'||tfile||'</a>'
     call lineout1 '<b>'||ofile||'</b><br><tt>('||hh||')</tt>'
    call lineout1 '</td><td><em>'||desc||'</em>'
    call lineout1 '</td>'
end 
call lineout1 '</table>'


call lineout1 '<p><table><tr><td bgcolor="#aaaaaa">'
call lineout1 '<tr><td colspan=3>Optional <em>argument</em> string: <input type="text" name="arg" size=60></td>'

if stuff.!specials<>'' then do
  call lineout1 '<tr><td>Sort variable:</td>'
  call lineout1 '<td><select name="SORTIT_VAR">'
  call lineout1 '<option selected value="">Entry date'
  do jj=1 to words(stuff.!specials)
      aww=strip(translate(word(stuff.!specials,jj)))
      call lineout1 '<option value="'||aww||'">'||aww
  end
  call lineout1 '</select></td>'
  call lineout1 '<td><input type="checkbox" name="SORTIT_ORDER" value="D">Descending order <em>unchecked means ascending</em></td>'
end

call lineout1 '<tr><td>Display recent:'
  call lineout1 '<input type="hidden" name="SELECTVAL" value="RECENT"></td>'
  call lineout1 '<td><select name="SELECT">'
      call lineout1 '<option  selected value="0">No (display all records)'
      call lineout1 '<option    value="HOUR">Last hour'
      call lineout1 '<option  value="DAY">Last day'
      call lineout1 '<option  value="WEEK">Last week'
      call lineout1 '<option  value="MONTH">Last month'
      call lineout1 '<option  value="YEAR">Last year'
      if stuff.!specials<>'' then 
         call lineout1 '<option  value="SORT">Last sorted-variable'
call lineout1 '</select></td>'

call lineout1 '<tr><td colspan=3>'
  call lineout1 '<input type="checkbox" name="updatearchive" value="1" checked>'
call lineout1 '<em>uncheck to (re)create all archive, and possibly comment, files</em>'
call lineout1 '</td>'

call lineout1 '<tr><td bgcolor="#aaaaaa"><input type="checkbox" name="backup" value=1>Backup old version of output files?</td>' 
call lineout1 '<td bgcolor="#eeaeae">Report errors: '
call lineout1 '<select name="DEBUG" size=1>'
call lineout1 '<Option value="-1">Ultra-terse'
call lineout1 '<Option value="0">None'
call lineout1 '<Option value="1"  >Major errors'
call lineout1 '<Option SELECTED  value="2">Verbose'
call lineout1 '<Option value="3">Very Verbose'

call lineout1 '</select></td>'
call lineout1 '<td><input name="sub2" type="submit" value="Generate the chosen output files"></td>'
call lineout1 '</table></form>'

call lineout1 '</td></table>'
call lineout1 '<p><table><tr><th>Notes</th>'
call lineout1 '<td><ul>'
call lineout1 '<li> The <u>template file</u> contains CMS instructions used to generate each output file.<p>'
call lineout1 '<li>If the <tt>update</tt> checkbox is checked, then CMS will update files.'
call lineout1 '<menu><li><u>If it is unchecked,</u> CMS will generate all the files specified in the template file.'
call lineout1 '<br>This includes the main output file, subset archives, singleton archives, and comment files.'
call lineout1 '<li><u>If it is checked,</u> CMS will: '
call lineout1 '<ol type="i">'
call lineout1 '<li>Update the main output file '
call lineout1 '<li>Update the most recent subset-archive (if specified), but not older subset archives'
call lineout1 '<li>Create new singleton archives (do not overwrite existing singleton archives,<b>even if</b> the contents of the record have changed)'
call lineout1 '<li>Update comment-files if any comments have been added to a record (or if a first comment has been added). '
call lineout1 '<br>If the contents of an existing comment have been modified, but no comments have been added, the comment file (for the record) will <b>not</b> be updated.'
call lineout1 '</ol></menu>'
call lineout1 '</ul></td><table>'

gooble:         /* jump here on error */
a=sre_cache_data(stuff.!series,'CLOSE')

call lineout1 '</body> </html>'
call lineout1 '1a'x 

rcode=sre_command('FILE ETAG_AUTO EXPIRES 30s NOCACHE NOIF ERASE TYPE text/html NAME 'tempfile,,idinfo)
return rcode



/*************/
/* read a definition file */
read_deffile:
parse arg afile
stuff.=''
stuff=sre_read_file(afile,2,2,1)

do until stuff=''
   parse var stuff aline '0d0a'x stuff ; aline=strip(aline)
   if aline='' then iterate
   if abbrev(aline,';')=1 then iterate
   parse var aline aname ':' avalue
   aname0=strip(translate(aname))
   aname='!'||aname0
   if stuff.aname='' then
      stuff.aname=strip(avalue)
   else
      stuff.aname=stuff.aname||'0d0a'x||strip(avalue)
   stuff.0=stuff.0||' '||aname0
end

return 1



/*******************/
/* view a stored file */
view_file:

daindx=optname.!deffile
ff=cmsdir_index||'\'||daindx
goo=stream(ff,'c','query exists')
if goo='' then do
   call lineout1 '<html><head><title>CMS: Viewing a template file</title></head> '
   call lineout1 'Error: could not find CMS-dataseries index file: '||ff
   signal gooble
end

chlist=strip(translate(sre_packur(optname.!!tplfile)))

call read_deffile ff
/* check for access rights */
rcode=check_access(stuff.!OWNERS,stuff.!SERIES)
if rcode<>1 then return rcode

thefile=''
do until stuff.!TEMPLATE = ''
   parse var stuff.!TEMPLATE aline '0d0a'x stuff.!template
   parse upper var aline incfile tplfile ',' .
   tplfile=strip(tplfile)
   if tplfile=chlist then do
      thefile=tplfile
      leave
   end
end

if thefile='' then do
   call lineout1 '<html><head><title>CMS: Viewing a template file</title></head> '
   call lineout1 'Error: template file not known: '||chlist
   signal gooble
end

/* if no path info, check in FILEDIR path (if specified). If no FILEDIR path, use CMS\FILES
   directory
*/
if abbrev(chlist,'\')=0 & pos(':',chlist)=0 then do     /* relative filename */
   if stuff.!FILEDIR<>'' then do
        defdir=stuff.!FILEDIR
   end 
   else do
         defdir=cmsdir_files
   end
   defdir=strip(defdir,,'\')||'\'
   chlist=defdir||chlist
end

if stream(chlist,'c','query exists')='' then do
   call lineout1 '<html><head><title>CMS: Viewing a template file</title></head> '
   call lineout1 'Error: template file does not exist: '||chlist
   signal gooble
end

dtype='text/plain'
rcode=sre_command('FILE ETAG_AUTO EXPIRES 30s NOCACHE NOIF TYPE text/plain '||dtype||' NAME 'chlist,,idinfo)
return rcode

gooble:         /* jump here on error */


call lineout1 '</body> </html>'
call lineout1 '1a'x 

rcode=sre_command('FILE ETAG_AUTO EXPIRES 30s NOCACHE NOIF ERASE TYPE text/html NAME 'tempfile,,idinfo)
return rcode




/*******************/
/* view an output  file */
view_file_out:
parse arg astext

daindx=optname.!deffile
ff=cmsdir_index||'\'||daindx
goo=stream(ff,'c','query exists')
if goo='' then do
   call lineout1 '<html><head><title>CMS: Viewing an output file</title></head> '
   call lineout1 'Error: could not find CMS-dataseries index file: '||ff
   signal gooble
end

chlist=strip(translate(sre_packur(optname.!!incfile)))

call read_deffile ff
/* check for access rights */
rcode=check_access(stuff.!OWNERS,stuff.!SERIES)
if rcode<>1 then return rcode

thefile=''
do until stuff.!TEMPLATE = ''
   parse var stuff.!TEMPLATE aline '0d0a'x stuff.!template
   parse upper var aline incfile tplfile ',' ddesc
   incfile=strip(incfile)
   if incfile=chlist then do
      thefile=incfile
      leave
   end
end

if thefile='' then do
   call lineout1 '<html><head><title>CMS: Viewing a template file</title></head> '
   call lineout1 'Error: a template file not known: '||chlist
   signal gooble
end

/* if no path info, check in FILEDIR path (if specified). If no FILEDIR path, use CMS\DATA
   directory
*/


if abbrev(chlist,'\')=0 & pos(':',chlist)=0 then do     /* relative filename */
   if stuff.!FILEDIR<>'' then do
        defdir=stuff.!FILEDIR
   end 
   else do
         defdir=cmsdir_files
   end
   defdir=strip(defdir,,'\')||'\'
   chlist=defdir||chlist
end

if stream(chlist,'c','query exists')='' then do
   call lineout1 '<html><head><title>CMS: Viewing an output file</title></head> '
   call lineout1 'Error: output file does not exist: '||chlist
   signal gooble
end

dtype='text/html '
if astext='TEXT' then dtype='text/plain '
rcode=sre_command('FILE ETAG_AUTO EXPIRES 30s NOCACHE NOIF TYPE '||dtype||' NAME 'chlist,,idinfo)
return rcode

gooble:         /* jump here on error */


call lineout1 '</body> </html>'
call lineout1 '1a'x 

rcode=sre_command('FILE ETAG_AUTO EXPIRES 30s NOCACHE NOIF ERASE TYPE text/html NAME 'tempfile,,idinfo)
return rcode



/*******************/
/* use selected template files to make output files */
generate_output:

stuff.=''

call lineout1 '<html><head><title>CMS: Generate output files from a CMS dataseries</title></head> '

call lineout1 '<font size=-1><a href="'||mainaddon||'?action=1">Another CMS action</a></font><br>'

call lineout1 '<h3>Generating output files from a CMS dataseries</h3> '

daindx=optname.!deffile
ff=cmsdir_index||'\'||daindx
goo=stream(ff,'c','query exists')
if goo='' then do
   call lineout1 'Error: could not find index file: '||ff
   signal gooble
end

call read_deffile ff
if update_verb=0 then do
/* check for access rights */
  rcode=check_access(stuff.!OWNERS,stuff.!SERIES)
  if rcode<>1 then return rcode
end

a=sre_cache_data(stuff.!series,'OPEN  INDEX',stuff.!file)
if a>0 then do
   amess=sre_cache_data('OPEN','ERROR',a,stuff.!file)
   call lineout1 '<b>Error:</b> '||amess
   signal gooble
end

recids.0=0
cshdata.='01'x

info.=''
info.!nprocs=0
info.!SERIES=stuff.!series
info.!CMSDIR=cmsdir
info.!sorted=''
a=sre_cache_data(stuff.!series,'INFO',,'CREATION MODIFIED RECORDS VARS DESCRIPTION')
parse var a astat a ; astat=strip(astat)
if astat<>'0' then do
   amess=sre_cache_data('INFO','ERROR',astat)
   call lineout1 '<b>Error retrieving info:</b> '||amess
   signal gooble
end

info.!ARG=optname.!ARG

parse var a . '0d0a'x . '09'x info.!series_created '09'x info.!series_modified '09'x ,
               info.!series_records '09'x  info.!series_CMSvarlist '09'x  info.!series_desc

info.!series_varnames=stuff.!varnames
info.!series_sortables=stuff.!specials
info.!series_pack=stuff.!pack
/* create a lookup-list (names of packed vars) */
ith=0
do mm=1 to words(info.!series_pack)
   ith=ith+1
   aw='!'||strip(translate(word(info.!series_pack,mm)))
   info.!cmpvars.ith=aw
end 
info.!cmpvars.0=ith
/* create another lookup list (names of special vars */
ith=0
do mm=1 to words(info.!series_sortables)
   ith=ith+1
   aw='!'||strip(translate(word(info.!series_sortables,mm)))
   info.!spvars.ith=aw
end
info.!spvars.0=ith
/* create another lookup list (names of both kinds of vars */
ith=0
do mm=1 to words(info.!series_varnames)
   ith=ith+1
   aw='!'||strip(translate(word(info.!series_varnames,mm)))
   info.!bothvars.ith=aw
end 
info.!bothvars.0=ith

do mm=1 to words(info.!series_varnames)
   aname=strip(translate(word(info.!series_varnames,mm)))
   info.!series_varname.mm=aname
   info.!series_vartype.mm=strip(translate(word(stuff.!vartypes,mm)))
   if  wordpos(aname,translate(stuff.!series_sortables))>0 then  info.!series_sortables.mm=1
end
info.!series_varname.0=words(info.!series_varnames)

call lineout1 '<tt>From the CMS dataseries: </tt><b>'||stuff.!series||'</b>'
call lineout1 '<menu><li><em>current # of records='info.!series_records'</em>'
call lineout1 '<li>Default output directory: <tt>'stuff.!Filedir||'</tt>'
call lineout1 '<li>Default upload directory: <tt>'stuff.!uploaddir||'</tt>'
if optname.!UPDATEARCHIVE=1 then 
  call lineout1 '<li>Update box checked: Yes '
else
  call lineout1 '<li>Update box checked: No'

call lineout1 '</menu>'

chlist=strip(translate(sre_packur(optname.!!tfile)))
if chlist='' then do
   call lineout1 '<b>You did not select any output files! </b>'
   call lineout1 'Please <a href="javascript:history.go(-1)">go back and select one!</a>'
   signal gooble
end

call lineout1 '<br><b>Using input argument</b>:<nobr><tt>'||info.!ARG||'</tt></nobr><p>'


/* make a list of template instructions.
  This can be added to, by ARCHIVE and COMMENT keywords */

cmt_open=''
dolist_1=''
dolist_c=''

do until stuff.!TEMPLATE = ''
   parse var stuff.!TEMPLATE aline '0d0a'x stuff.!template
   parse  var aline incfile tplfile ',' desc_tpl ;
   tplfile=translate(strip(tplfile))
   zid=strip(word(optname.!SORTIT_VAR,1)) ; if zid='' then zid='!ID' /* var name */
   zid2=strip(word(optname.!SORTIT_ORDER,1)) ; if zid2='' then zid2='A' /* A/D order */
   if wordpos(tplfile,chlist)=0 then iterate
   dolist_1=dolist_1||tplfile||' '||incfile||' '||zid||' '||zid2||' I 0,'|| ,
            OPTNAME.!SELECT||'09'x||desc_tpl||'09'x||info.!arg||'09'x||' '||'09'x||optname.!SELECTVAL||'0d0a'x  
end

incfile_name=filespec('n',incfile)

call lineout1 "<ul>"

/* now start processing template files (starting with the invoked ones, but with many
others possibly added, such as comment and archive files */

do until dolist_1=''
   parse var dolist_1 aline '0d0a'x dolist_1
   if aline='' then iterate
   parse var aline tplfile incfile sortvar sortorder sorttype  isarc ',' ,
     vSELECT '09'x desc_tpl '09'x da_arg  '09'x aaurl  '09'x vSELECTVAL

   info.!arg=da_arg

   tplfile=strip(tplfile); incfile=strip(incfile)
   if abbrev(incfile,'+')=1 then do
      incfile=substr(incfile,2)
      foo=sysfiledelete(incfile)
   end
   info.!template_file=tplfile
   info.!template_desc=desc_tpl
   if update_verb<>0 & (isarc <>0 & isarc<>"") then do
      if update_verb=1  then isarc=2    /* update only! */
      if update_verb=3  then isarc=1   /* always recreate */
/*update_verb=2 means "use instructions in template file as is */
   end 
   select
     when  isarc=1  then do
         if update_verb<>0 & optname.!do_arc<>1 then iterate   /* update mode, but not archives? */
         call lineout1 '<li type="square">Generating archive file'
         hh='<a target="viewer" href="'||aaurl||'">'||filespec('n',incfile)||'</a>'
     end
     when  isarc=2 then do
         if update_verb<>0 & optname.!do_arc<>1 then iterate   /* update mode, but not archives? */
         call lineout1 '<li type="square">Updating archive file'
         hh='<a target="viewer" href="'||aaurl||'">'||filespec('n',incfile)||'</a>'
     end

     otherwise do
        call lineout1 '<li><strong>Generating</strong>'
        hh='<a target="viewer" href="'||thisaddon||'?action=6&deffile='||optname.!deffile||'&incfile='||incfile||'">'||incfile||'</a>'
        hh=hh||' <tt>(as <a target="viewer" href="'||thisaddon||'?action=61&deffile='||optname.!deffile||'&incfile='||incfile||'">text</a>)'

     end
   end
   call lineout1 hh||'<em>('desc_tpl||')</em>'

/*  !!! NOW, make the output file, using the template file !!! */

   astat=make_inc_file(tplfile,incfile,optname.!BACKUP,optname.!DEBUG, ,
                       sortvar,sortorder,sorttype,vselect,vselectval,isarc)

   foo=report_status(tempfile,astat)

end
call lineout1 "</ul>"


/* are there comment files to create */
if dolist_c<>'' then do
    if (update_verb=0) | (update_verb<>0 & optname.!do_cmt=1) then do   /* update mode, but not comments? */
      a=sre_cache_data(stuff.!series,'CLOSE')  
      call generate_comments
   end
end 
call lineout1 '<hr><em>Records, comments, and archives have been processed.</em>'

if optname.!DEBUG=-1 & update_verb=0 then do    /* oops, do ultra terse response */
   call lineout1 '1a'x 
   foo=sysfiledelete(tempfile)
   call lineout1 '<Html><head><title>Updating dataseries </title></head>'
   call lineout1 '<body>'
   call lineout1 '<font size=-1><a href="javascript:close()">close window</a></font><p>'
   call lineout1 'The output files for the <b>'||stuff.!series||'</b> dataseries <em>('||stuff.!desc||')</em> have been <b>updated</b>.' 
   call lineout1 ''
end


gooble:         /* jump here on error */

a=sre_cache_data(stuff.!series,'CLOSE')  /* might be comment, might be records dataseries */

gooble2: 

call lineout1 '</body> </html>'
call lineout1 '1a'x 

if update_verb=0 then do
  rcode=sre_command('FILE ETAG_AUTO EXPIRES 30s NOCACHE NOIF ERASE TYPE text/html NAME 'tempfile,,idinfo)
  return rcode
end

amess=' # of files written='||update_files||', # of bytes written='||update_bytes
return amess


/*******************/
/* generate comment files */
generate_comments:

call lineout1 '<h3>Generating comment files</h3> '
a=sre_cache_data(stuff.!series,'OPEN  INDEX',stuff.!commentfile)
if a>0 then do
   amess=sre_cache_data('OPEN','ERROR',a,stuff.!commentfile)
   call lineout1 '<b>Error:</b> '||amess
   signal gooble
end

drop recids.
drop cshdata.
recids.0=0
cshdata.='01'x

drop info.
info.=''
info.!nprocs=0
info.!SERIES=stuff.!series
info.!CMSDIR=cmsdir
info.!sorted=''
info.!COMMENTS_MODE=1

a=sre_cache_data(stuff.!series,'INFO',,'CREATION MODIFIED RECORDS VARS DESCRIPTION')
parse var a astat a ; astat=strip(astat)
if astat<>'0' then do
   amess=sre_cache_data('INFO','ERROR',astat)
   call lineout1 '<b>Error retrieving info:</b> '||amess
   signal gooble
end

parse var a . '0d0a'x . '09'x info.!series_created '09'x info.!series_modified '09'x ,
               info.!series_records '09'x  info.!series_varnames '09'x  info.!series_desc

/* all variables are sortables, no packed variables used */
info.!series_varnames=stuff.!commentvars
info.!series_sortables=info.!series_varnames
info.!series_pack=''
info.!cmpvars.0=-1
/* create lookup list (names of special vars */
ith=0
do mm=1 to words(info.!series_sortables)
   ith=ith+1
   aw='!'||strip(translate(word(info.!series_sortables,mm)))
   info.!spvars.ith=aw
   info.!bothvars.ith=aw
   info.!series_sortables.ith=1
end
info.!spvars.0=ith
info.!bothvars.0=ith

info.!series_varname.0=words(info.!series_varnames)
do mm=1 to info.!series_varname.0
   aname=strip(translate(word(info.!series_varnames,mm)))
   info.!series_varname.mm=aname
   if abbrev(aname,'_')=0 then
     info.!series_vartype.mm='1'
   else
     info.!series_vartype.mm='M'
end

call lineout1 '<em>(current # of comment records='||info.!series_records-3||')</em><br>'

call lineout1 '<ol type="i">'
/* now start processing template files (starting with the invoked ones, but with many
others possibly added, such as comment and archive files */
do until dolist_c=''
   parse var dolist_c aline '0d0a'x dolist_c
   if aline='' then iterate
   parse var aline tplfile incfile sortvar sortorder sorttype  isarc ',' ,
     vSELECT '09'x desc_tpl '09'x da_arg  '09'x aaurl  '09'x vSELECTVAL ,
     '09'x retainval

   if update_verb=1 then isarc=12   /* update only  */
   if update_verb=3 then isarc=13   /* always re-create */

   parse var desc_tpl info.!comments_mode.!status ';' info.!comments_mode.!abstract

   info.!arg=da_arg

   vv='!$RETAIN'
   info.vv=retainval
   tplfile=strip(tplfile); incfile=strip(incfile)
   if abbrev(incfile,'+')=1 then do
      incfile=substr(incfile,2)
      foo=sysfiledelete(incfile)
   end

   info.!template_file=tplfile
   info.!template_desc=desc_tpl
   select
     when  isarc=12 then do     /* update comments file */
         call lineout1 '<li type="disc">Updating comments file'
         hh='<a target="viewer" href="'||aaurl||'">'||filespec('n',incfile)||'</a>'
     end
     when  isarc=13 then do  /* create comments file */
         call lineout1 '<li type="disc">Creating comments file'
         hh='<a target="viewer" href="'||aaurl||'">'||filespec('n',incfile)||'</a>'
     end
     otherwise do
        call lineout1 ' Error in comment generation: 'isarc
        return ''
     end
   end
   call lineout1 hh||'<em>('|| info.!comments_mode.!abstract||')</em>'

/*  --- NOW, make the output file, using the template file  ----- */

   astat=make_inc_file(tplfile,incfile,optname.!BACKUP,optname.!DEBUG,sortvar,sortorder,sorttype,vselect,vselectval,isarc)

   foo=report_status(tempfile,astat)
end
call lineout1 "</ol>"

return ''


/**********************************************/
/* report status of output file construction */
report_status:procedure expose gerrs. info. recids. update_verb 
parse arg tempfile,astat
call lineout1 '<BR>'||astat


call lineout1 '<br>'
e1='' ; e2=''
if gerrs.!E>0 then do
   e1='<font color="#bb1122" ><u>' ; e2='</u></font>'
end 
call lineout1 e1||'# Errors: '||gerrs.!E||e2

e1='' ; e2=''
if gerrs.!W>0 then do
   e1='<em>' ; e2='</em>'
end 
call lineout1 ', '||e1||'# of Warnings: '||gerrs.!W||e2

e1='' ; e2=''
if gerrs.!M>0 then do
   e1='' ; e2=''
end 
call lineout1 ', '||e1||'# of Messages: '||gerrs.!M||e2

dlevel=gerrs.!DO
if dlevel=0 then return 1   /* no specifics */

nams.!E='<u>Error</u>. '
nams.!W='<em>Warning</em>. '
nams.!M='Message. '
call lineout1 '<pre>'
do mm=1 to gerrs.0
   atype=gerrs.mm.!TYPE
   if atype='W' & dlevel<2 then iterate
   if atype='M' & dlevel<2 then iterate

   vtype='!'||atype
   amessage=nams.vtype||gerrs.mm
   
   if dlevel>2 then do
       if lengtH(gerrs.mm.!line)<50 then
            gline=gerrs.mm.!line
       else
            gline=left(gerrs.mm.!line,50)||' ...  '
       gg2=''
       if gerrs.mm.!line<>'' & gerrs.mm.!noloc<>1 then gg2=' ['||gline||']'
       amessage=amessage||gg2
   end
   call lineout1 '   '||amessage
end
call lineout1 '</PRE>'

return 1

/*
gerrs.ii=message
gerrs.ii.!LINE=gerrs.!CURRENT
gerrs.ii.!TYPE=severe
gerrs.!DO=debug   /* -1, very terse , 0 or ''=do not report errors, 1=major errors, 2=verbose errors, 3=too verbose */
*/



/**********************************************/
/* ===========================================*/
/**********************************************/


/* generate an output file, using a template file and data from the
   CMS dataseries */
make_inc_file:procedure expose optname. stuff. cmsdir_files cmsdir_data info. recids. cshdata. tempfile   ,
                                daindx gerrs. dolist_1 dolist_c root thisaddon  cmsdir_uploads cmsdir_lib cmsdir cmsdir_pol ,
                                update_verb update_files update_bytes
parse arg tplfile0,outfile0,backout,debug,def_sortvar,def_sortorder,def_sorttype,def_select,def_selectval,isarchive
gerrs.=''
if wordpos(debug,'-1 0 1 2 3 4')=0 then debug=0
gerrs.!DO=debug   /* 0 or ''=do not report errors, 1=major errors, 2=verbose errors */
gerrs.0=0
gerrs.!W=0
gerrs.!E=0
gerrs.!M=0

recids.=''
recids.0=0

info.!!count=0

if stuff.!FILEDIR<>'' then do
        defdir=stuff.!FILEDIR
end 
else do
         defdir=cmsdir_files
end
defdir=strip(defdir,,'\')||'\'

if stuff.!UPLOADDIR<>'' then do
        defdir_up=stuff.!UPLOADDIR
end 
else do
         defdir_up=cmsdir_uploads
end
defdir_up=strip(defdir_up,,'\')||'\'


if abbrev(tplfile0,'\')=0 & pos(':',tplfile0)=0 then do     /* relative filename */
   tplfile=defdir||tplfile0
   if stream(tplfile,'c','query exists')='' then
     return 'Error: could not find template file: 'tplfile0','tplfile
end
else do
   tplfile=tplfile0
end 

bakfile=''
outfile=outfile0

if isarchive=12 | update_verb<>0 then backout=0      /* if update mode, do NOT backup old comment files */

if abbrev(outfile0,'\')=0 & pos(':',outfile0)=0 then do     /* relative filename */
     if stuff.!FILEDIR<>'' then do
        defdir=stuff.!FILEDIR
     end 
     else do
         defdir=cmsdir_files
     end
     defdir=strip(defdir,,'\')||'\'
     outfile=defdir||outfile0
end
else do
      outfile=outfile0
end

isfile=stream(outfile,'c','query exists')

oldstamp=0
if isfile<>"" then  foo=sysgetea(outfile,'CMS_STAMP','OLDSTAMP')
if datatype(oldstamp)<>'NUM' then oldstamp=0

/* backup old version of outfile */
if isfile<>'' & backout=1 then do
       d1=filespec('d',outfile)||filespec('p',outfile)
       d2=filespec('n',outfile)
       ii=lastpos('.',d2)
       if ii=0 then ii=length(d2)+1
       bakfile=sre_tempfile(left(d2,ii)||'BK?','?',d1)
       foo=dosrename(outfile,bakfile)
       if foo=0 then return 'Error. unable to rename old version of output file: 'outfile
end

call lineout1 '<br>Using template file:<tt>'||tplfile||'</tt><br>'

tplstuff=sre_read_file(tplfile,2,2,1)
if tplstuff='' then return 'Error. Empty or inaccessible template file: '||tplfile

info.!template_file=tplfile

CINFO.=''
CINFO.0='!SERIES  !ARG !COUNT !COUNT_BLOCK !FILEDIR !UPLOADIR !DEFFILE '
cinfo.!0=cinfo.0   /* original special vars */
goose='!!SERIES'
CINFO.goose=info.!series

CINFO.!!FILEDIR=stuff.!filedir     /* these can be used by callable procedures */
cinfo.!!uploaddir=stuff.!!uploaddir
cinfo.!!DEFFILE=daindx

cinfo.!!ARG=info.!arg

fpp=load_recids(def_sortvar,def_sortorder,def_sorttype,1)   /* this loads the "id7"s, possibly sorted, with deselection */

/* if archive update (isarchive=12), check to see if current age of file exceeds newest
  comment */

if isarchive=12 & isfile<>'' then do   /* update comments mode, with an existing file */
/* check recids, see if any are larger then OLDSTAMP */
   doit1=0
   foo2=info.!comments_mode.!status 
   parse upper var foo2 . 'CMT_MOD=' recentmod . ;recentmod=strip(recentmod)
   if recentmod<oldstamp then do
       fpp=note_error('M','No new comments for record '||def_selectval)
       return 'Ok (no new comments, so nothing written)'
   end 
   
end 

fpp=note_error('M','# of records used= '||recids.0)

subroutines.=''

/* now we are ready to start parsing */
repstat=0
outit=''
dodefine=0
do until tplstuff=''
   parse var tplstuff a1 '<' tplstuff
   if tplstuff='' then do
      a2a=''
   end
   else do
     i0=verify(tplstuff,' '||'0d0a09'x)        /* find first non-blank character */
     a2a=substr(tplstuff,i0,3)
   end

   if abbrev(a2a,'!--')=1  then do  /* start of comment, find end */
      do forever
         i1=pos('--',tplstuff,i0+3)
         if i1=0 then do           /* no end of comment */
               a1=a1||'<'||tplstuff
               fpp=note_error('W','Open comment ')
               tplstuff=''
               a2=''
               leave
          end
          i2=pos('>',tplstuff,i1)
          if i2=0 then do               /* no end of comment */
                a1=a1||'<'||tplstuff
                a2=''
                tplstuff=''
                fpp=note_error('W','Open comment ')
                leave
          end
          doug=substr(tplstuff,i1,1+i2-i1)
          doug=space(translate(doug,' ','0d0a09'x),0)
          if doug='-->' then do                 /* got end of comment */
             a1=a1||'<'||left(tplstuff,i2)
             a2=''
             tplstuff=substr(tplstuff,i2+1)
             leave
          end
          i0=i2
      end     /* do forever (search for -->) */
   end          /* <!-- found */        
   else do              /* not beginning with <!-- */
      parse var tplstuff a2 '>' tplstuff
   end

/* note that DEFINE and /DEFINE are not allowed in REPEAT */

   if repstat=0 then do   /* not in a REPEAT  */
         if dodefine=0 then
              outit=outit||a1   
         else
              defineblock=defineblock||a1
         if a2='' then iterate
         a2a=translate(a2,' ','0d0a09'x)
         if abbrev(strip(a2a),'?')=0 then do  /* this is NOT a <? cmscommand > */
            if dodefine=0 then
               outit=outit||'<'||a2||'>'
            else
               defineblock=defineblock||'<'||a2||'>'
            iterate
         end 

         parse var a2a . '?' keyword options
         keyword=strip(translate(keyword))

         if keyword='DEFINE' then do
             if dodefine=1 then do
               fpp=note_error('W','DEFINE not allowed in a DEFINE .. /DEFINE block')
               tplstuff=''
               a2=''
               leave
             end
             dodefine=1
             defineblock=''
             parse upper var options define_name define_arg .
             if define_name='' then do
                fpp=note_error('W','DEFINE did not specify a name')
                tplstuff=''
                a2=''
                leave
             end
             iterate
         end 
         if keyword='/DEFINE' then do
            vv='!'||strip(define_name)
            subroutines.vv=defineblock
            subroutines.vv.!arg=define_arg
            dodefine=0
            drop defineblock
            iterate
         end
         if dodefine=1 then do
               defineblock=defineblock||'<'||a2||'>'
               iterate
         end 

/* not starting,in, or ending a DEFINE. So process it */

         if keyword='REPEAT' then do
            repstat='REPEAT '||strip(options)
            rep_stuff.=''
            rep_stuff.0=0
            iterate
         end

         if keyword='GOSUB' then do             /* gosub is just a textual insert */
             parse  var options sub_name argval  
             argval=strip(argval)
             if abbrev(argval,'"')=1 then do
                parse var argval '"' argval '"' .
             end 
             vv='!'||translate(strip(sub_name))
             zzstuff=subroutines.vv
             argname=subroutines.vv.!arg
             if argname<>'' then zzstuff=sre_replacestrg(zzstuff,argname,argval,'ALL')
             tplstuff=zzstuff||tplstuff
             iterate
          end

         if abbrev(keyword,'CALL') | abbrev(keyword,'INT') then do    /* get options, retaining crlfs */
              parse var a2 . '?' zz 
              jj=verify(zz,' '||'0d0a09'x)  /* jj is start of CALL */
              zz2=substr(zz,jj)             /* CALL $nn stuff */
              parse var zz2 . zz2           /* strip off first word (the CALL) */
              parse upper var zz2 w1 .  
              w1=space(translate(w1,' ','0d0a09'x),0)         /* check for $CODE or $FILE  */
              select 
                when w1='$CODE' then do
                  parse var zz2 w1 options
                  options='$CODE '||options
               end 
               when w1='$FILE' then do
                  parse var zz2 w1 options
                 options='$FILE '||options
               end
               otherwise do
                  if abbrev(keyword,'INT')=1 then 
                     options='$CODE '||zz2
                   else
                     options='$FILE '||zz2
               end
               keyword='CALL'
             end  /* select */
         end

         parse var a2a '?' a2b          /* obtain the approriate CMS data */
         gerrs.!CURRENT=keyword||' '||options
         a3=process_cms(keyword,options,,,optname.!updatearchive)  /* a3 might be '' (i.e.; if an IF is true) */
         outit=outit||a3
         iterate
    end 

/* otherwise, in a REPEAT. Accumulate REP_STUFF until <? /REPEAT >  is found  */
   
   irep=rep_stuff.0+1
   rep_stuff.irep=a1
   rep_stuff.irep.!cms=0
   rep_stuff.0=irep
   if a2='' then iterate

   a2a=translate(a2,' ','0d0a09'x)
   if abbrev(strip(a2a),'?')=0 then do  /* this is NOT a <? cmscommand > */
        irep=rep_stuff.0+1
        rep_stuff.irep='<'||a2||'>'
        rep_stuff.irep.!cms=0
        rep_stuff.0=irep
        iterate
    end 
    parse var a2a . '?' keyword options


    keyword=strip(translate(keyword))
    if keyword='/REPEAT' then do                /* repeat section has been read. do it! */
         a3=process_cms_repeat(repstat)  /* also uses exposed rep_stuff. and info. */
         outit=outit||a3      
         repstat=0
         iterate
    end 

    if keyword='GOSUB' then do             /* gosub is just a textual insert */
             parse  var options sub_name argval  
             argval=strip(argval)
             if abbrev(argval,'"')=1 then do
                parse var argval '"' argval '"' .
             end 
             vv='!'||translate(strip(sub_name))
             zzstuff=subroutines.vv
             argname=subroutines.vv.!arg
             if argname<>'' then zzstuff=sre_replacestrg(zzstuff,argname,argval,'ALL')
             tplstuff=zzstuff||tplstuff
             iterate
     end

    if abbrev(keyword,'CALL') | abbrev(keyword,'INT') then do    /* get options, retaining crlfs */
        parse var a2 . '?' zz 
        jj=verify(zz,' '||'0d0a09'x)  /* jj is start of CALL */
        zz2=substr(zz,jj)             /* CALL $nn stuff */
        parse var zz2 . zz2           /* strip off first word (the CALL) */
        parse upper var zz2 w1 .  
        w1=space(translate(w1,' ','0d0a09'x),0)         /* check for $CODE or $FILE  */
        select 
           when w1='$CODE' then do
             parse var zz2 w1 options
             options='$CODE '||options
          end 
          when w1='$FILE' then do
             parse var zz2 w1 options
             options='$FILE '||options
          end
          otherwise do
            if abbrev(keyword,'INT')=1 then
               options='$CODE '||zz2
            else
               options='$FILE '||zz2
          end
        end  
        keyword='CALL'

    end

/* add to rep_stuff "repeat these" list */
    irep=rep_stuff.0+1
    rep_stuff.irep.1=keyword
    if keyword='IF' then do
       aaa=parse_if(options,irep)
    end
    else do
       rep_stuff.irep.2=options
       rep_stuff.irep.!cms=1
       rep_stuff.0=irep
    end

end             /* get next piece of TPLSTUFF */

eewrite=1
if (isarchive=0 | isarchive="") & (update_verb<>0) then do /*check optname.!DO_MAIN */
   if optname.!DO_MAIN<>1 then eewrite=0
end

if eewrite=1 then do
  foo=sre_append_file(outfile,outit,1,,1)
  select 
     when foo='-1' then return 'Error. Could not open file: 'outfile
     when foo='-2' then return 'Error. Problem writing to file: 'outfile
     otherwise nop
  end 
/* add an ea for use as timestamp */
  nowtime=sre_gmt(,,'JS',0)
  goo=sysputea(outfile,'CMS_STAMP',nowtime)
  if update_verb<>0 then do
       update_files=update_files+1
       update_bytes=update_bytes+foo
  end
  if bakfile<>'' & backout<>1 then do 
     fpp=sysfiledelete(bakfile)
     if fpp<>0 then fpp=note_error('M',"Unable to delete temp file: "bakfile)
  end
end


return 'Ok (bytes written= '||foo||',  records written='||info.!!count||')'



/*********************/
/* parse an ?IF element -- check for THEN ELSE conditions. If
   THEN ELSE exists, add several items (not just one) to rep_stuff */
parse_if:procedure expose rep_stuff. tplstuff  update_verb 
parse arg options,irep
topts=translate(options)

ithen=pos(' THEN ',topts)
if ithen=0 then do                   /* no THEN  (or ELSE)-- add this IF and return */
  rep_stuff.irep.2=options
  rep_stuff.irep.!cms=1
  rep_stuff.0=irep
  return 1
end

p1=left(options,ithen-1)
rep_stuff.irep.2=p1
rep_stuff.irep.!cms=1
rep_stuff.0=irep

/* now add stuff to TPLSTUFF */

options=substr(options,ithen)
topts=translate(options)

ielse=pos(' ELSE ',topts)
if ielse=0 then do
   parse var options . p1       /* get rid of the THEN */
   p1=translate(p1,'<>','{}')
   tplstuff=p1||' <? /IF > '||tplstuff
   return 1
end 

p1=left(options,ielse-1)
parse var p1 . p1
p1=translate(p1,'<>','{}')

p2=substr(options,ielse)
parse var p2 . p2
p2=translate(p2,'<>','{}')

tadd=p1||' <? /IF > <? ELSE > '||p2||' <? /ELSE> '

tplstuff=tadd||tplstuff

return 1



/************/
/* replace a <? cms ...> element */
process_cms:procedure expose stuff. info. cshdata. defdir_up defdir CINFO. gerrs. recids. tplfile outfile  update_verb   ,
                              def_select def_selectval dolist_1  dolist_c isarchive optname. cmsdir_files cmsdir_uploads cmsdir_lib cmsdir cmsdir_pol
parse arg keyword0,options,useid7,varname,updatearchive

keyword=strip(translate(keyword0))
opt1=strip(translate(options))
findid=''


select
   when keyword="!--" then return '<!-- '

   when keyword="--" then return ' --> '

   when keyword="<" then return '<'

   when keyword="/<" then return '>'

   WHEN abbrev(keyword,'CALL')=1 then do   /* CALL */
      spvars='!DEFFILE !NUMREC !UPLOADDIR !FILEDIR  '

      parse var options ctype options
      if ctype='$FILE' then do
         parse var options dafile userargs ; dafile=strip(dafile)
         options=read_interp_file(dafile,info.!cmsdir)
      end 
      k1=info.!nprocs+1 
      info.!nprocs=k1
      daresult=''
      astr=make_interpret_string(options,info.!series_varnames)  /* expose usevars and useval */

      astat=rexxrun('S',astr,daresult,usevars,useval,userargs,info.!series_varnames,k1)
      retstuff=''
      if astat=0 then do                /* success */
           parse var daresult rstat ',' rhd '0d0a'x rval
           if rstat=0 then do                   /* error in CALLed code */
                parse var rhd rcode rline ; rline=rline-29   /* 29 lines of prep material */
                foo=note_error('E','Error 'rcode' in CALLed procedure at line 'rline)
           end 
           else do
             parse var rhd . nres rhd
             if nres<>0 then retstuff=left(rval,nres)
             iat=nres+1
             if rhd<>'' then do
             do until rhd=''
                parse var rhd rrname rrbytes rhd ; rrname=strip(rrname) ; rrbytes=strip(rrbytes)
                if wordpos(translate(rrname),spvars)>0 then iterate /* don't touch special vars */

                if rrbytes=0 then do
                   CINFO.rrname=''
                end
                else do
                  CINFO.rrname=substr(rval,iat,rrbytes)
                  iat=iat+rrbytes
                  gg=substr(rrname,2)
                  if wordpos(gg,cinfo.0)=0 then cinfo.0=cinfo.0||' '||gg
                end     
             end 
             end
           end
      end
      else do
          foo=note_error('E','Error ('||astat||' calling procedure ')
      end

      return retstuff
   end


   when abbrev(keyword,'REM')=1 then return ''   


/* archive expects FILE=template_file  
                   TYPE=nnn|HOUR|DAY|WEEK|WEEK2|MONTH|MONTH2|YEAR|SORT 
                   LIST=ul|ol|table|ulA|olA|tableA
                   OUTPUT=output_file_pattern
                   URL=base_url
                   MESSAGE="a message"
                   UPDATE=0|1|AUTO
                   #FIELD="field contents"

           Template_file, if not fully specified, is relative to the current
           template file. It will be called repetitively, once for each "set",
           where the set is specified by the TYPE option (say, if WEEK is chosen,
           a set will consist of all records entered in a given week).

           Note that the template file can also specify a SORT -- this will be a sorting of
           the records in the "set" specific to this archive 
           (say, a descending sort of the records in a week).

           The SORT type means "by value of current sort variable"
           If no sort variable currently selected, use of SORT yields an error 

           The nnn type means "every nnn'th record" (nnn must be an integer >0),
           using a descending sort of time !ID (time of entry).

           The DAY and DAY2 types are  calendar day. DAY2 yields a
           more compact listing.

           The WEEK and WEEK2 types are for Sunday-Saturday weeks. WEEK2 yields a
           more compact listing.

           The MONTH and MONTH2 are for calendar month (MONTH2 yields a more compact listing)

           For all of the types, NO observations are de-selected.

           List specifies how to create the links to the archive files.
           ul,ol, and table are "descending" (most recent at top of list).
           ulA,olA,and tableA are ascending.

           Output_file_Pattern must contain a * (the * will be replaced by
           a number). If not specified, a suitable modification of the current
           output file will be used. If specified, and relative, a file
           relative to the current output file will be created.

           If UPDATE=1,  or if UPDATE not specified, then update files (do
           not attempt overwrite files that already exist).
           UPDATE=0, then always create new file names.
           UPDATE=AUTO, use the UPDATEARCHIVE form option

           Base_url, if specified, will be used when forming links to the "archive"
           files. If not specified, not path information will be used in the link.

           "A message" will be the INFO.!ARG variable of the template_file.

           #FIELD can be any "field".

*/

   WHEN keyword="/ARCHIVE"  | keyword='/LINK' | keyword='/COMMENT' then do
         return '</a>'
   end

   WHEN keyword="ARCHIVE" then do

       if isarchive='12' | isarchive='13' then do
            foo=note_Error('E','ARCHIVE: archive not permitted when creating comment files ')
            return ''
       end

       call parse_options(options)  

       arctfile=strip(translate(ropts.!FILE))
       arctfile=fix_file(arctfile,tplfile)
       if arctfile='' then do
            foo=note_Error('E','ARCHIVE: the template file not specified: 'arctfile)
            return ''
       end
       if stream(arctfile,'c','query exists')='' then do
            foo=note_Error('E','ARCHIVE: template file does not exist: 'arctfile)
            return ''
       end

       arcofile=strip(translate(ropts.!OUTPUT))
       if arcofile='' then do
          iip=lastpos('.',outfile)
          if iip=0  then iip=length(outfile)+1
          p1=left(outfile,iip-1)
          p2=substr(outfile,iip+1)
          arcofile=p1||'$A*.'||p2
       end
       if pos('*',arcofile)=0 then do
            foo=note_Error('E','ARCHIVE; output file does not contain * : 'arcofile)
            return ''
       end
       arcofile=fix_file(arcofile,outfile,1)

       t1=strip(filespec('d',arcofile)||filespec('p',arcofile),,'\')
       if right(t1,1)=':' then t1=t1||'\'
       if dosisdir(t1)=0 then do
            foo=note_Error('E','ARCHIVE: i) no such directory for output files: 't1)
            return ''
       end

       arcurl=strip(translate(ropts.!URL))
       arcurl=strip(arcurl,'t','/')||'/'

       arclist=strip(translate(ropts.!LIST))
       if arclist='' then arclist='UL'
       if wordpos(arclist,'UL OL TABLE ULA OLA TABLEA')=0 then do
            foo=note_Error('E','ARCHIVE; list type not UL,ULA, OL,OLA, TABLE, or TABLEA: 'arclist)
            return ''
       end

       arcmessage=strip(ropts.!MESSAGE)

       arctype=strip(translate(ropts.!TYPE))
       select 
        when arctype='SORT' then do
          iis=info.!sorted 
          if iis='' | iis='!ID' then do
            foo=note_Error('E','ARCHIVE type=SORT, but SORT variable not specified.')
            return ''
          end
        end
        when datatype(arctype)='NUM' then do
            arctype=trunc(arctype)
            if arctype <1 then do
               foo=note_Error('E','ARCHIVE: type=nnn must be integer > 0: 'arctype)
               return ''
            end
        end 
        otherwise do
          if wordpos(arctype,'HOUR DAY DAY2 WEEK WEEK2 MONTH MONTH2 YEAR')=0 then do
               foo=note_Error('E','ARCHIVE: unknown type: 'arctype)
               return ''
          end 
        end
       end

       doupdate=strip(translate(ropts.!UPDATE))
       if doupdate='AUTO' | doupdate='' then doupdate=updatearchive
       if doupdate<>1 then doupdate=0

       if update_verb=1 then doupdate=1   /* always update */
       if update_verb=3 then doupdate=0  /* always create */

/* now we are ready to make a list of archive file creation instructions */
  
       arcdolist=load_arcfiles(doupdate+1)  /*1=create,2=update */


/* update archives selected? */
       if doupdate<>1 then do
            dadd=arcdolist
       end 
       else do
          dadd=check_dolist(arcdolist)
       end 

       dolist_1=dadd||'0d0a'x||dolist_1  /* add to list of template files to process */

/* make a "list of links" for inclusion in the output file */
       showlist=make_arc_links(arcdolist,ropts.!2)

       return showlist

/* arcdo, and arcmessage, are used to run new templates */

   end

   WHEN keyword="INC" then do
        parse upper var options lname  avalue .
        lname=translate(space(lname,0))
        jj1=wordpos(lname,'!ID !SERIES '||info.!series_varnames)
        if abbrev(lname,'!')=1 | jj1>0 then do
            foo=note_Error('W','Attempting to INCrement CMS variable ('lname)
            return '' /* not allowed to modify  variables defined in dataseries */
        end
        vv='!'||strip(lname)
        val1=CINFO.vv
        if datatype(val1)<>'NUM' then do
            foo=note_error('E','Attempting to INCrement undefined variable ('||lname)
            return ''  /* this accounts for "un SET" local variables */
        end
        avalue=strip(avalue)
        if avalue='' then avalue=1
        if datatype(avalue)<>'NUM' then do
            foo=note_error('W','Attempting to INCrement non-numeric variable ('||lname)
            return ''
        end
        CINFO.vv=CINFO.vv+avalue

        return ''
   end

   WHEN keyword="SET" then do
        parse var options lname '=' lval 
        if useid7<>'' then lval=useid7  /* this is used by VAR .. ASSIGN=xyz */
        if lval='' then do
          if pos('=',options)=0 then do
             parse var options lname lval
          end
          lval=strip(lval)
        end
        if lval='' then lval=0

        lname=translate(space(lname,0))
        jj1=wordpos(lname,info.!series_varnames)
        if abbrev(lname,'!')=1 | jj1>0 then do
            fpp=note_error('E',' Not allowed to set CMS variable: 'lname)
            return ''   /* not allowed to redefine variables defined in dataseries */
        end
        lval=strip(lval)
        if useid7='' then do
          if abbrev(lval,'"')=1 then do /* quotes ? then it is literal */
             parse var lval . '"' lval '"' .
          end
          else do
            if pos('(',lval)>0  then do /* a function on a value or a variable */
              lval=fig_set_function(lval)
            end
            else do             /* not a fuction, use first word */
               parse var lval lval .
            end
          end
        end

        vv='!'||lname
        CINFO.vv=lval
        if wordpos(lname,CINFO.0)=0 then   CINFO.0=CINFO.0||' '||lname
        return ''
   end 

   WHEN keyword="READFILE" then do   /* specify sort parameters */
        val1=''
        if opt1='' then do
             fpp=note_error('W','You did not specify a varname in a READFILE')
             return '' 
        end
        if words(opt1)=1 then do        /* simple form */
             fvarname=translate(strip(opt1))
             if pos('.',fvarname)>0 then val1=fvarname
             vsave=''; nodisplay=0
             usefdir=strip(stuff.!filedir,,'\')||'\'
        end 
        else do
           call parse_options(opt1)  /* name=, order= */
           fvarname=strip(ropts.!NAME)
           if fvarname='' then do
             fpp=note_error('W','You did not specify a NAME= in a READFILE')
             return '' 
           end
           if pos('.',fvarname)>0 then val1=fvarname
           fdir=strip(ropts.!DIR)
           if pos(':',fdir)>0 | abbrev(fdir,'\')=1 then do  /* fully qualified */
               usefdir=fdir
           end 
           else do
                usefdir=strip(stuff.!filedir,,'\')||'\'||fdir
           end 
           usefdir=strip(usefdir,,'\')||'\'
           nodisplay=strip(ropts.!NODISPLAY)
           vsave=translate(strip(ropts.!SAVE))
           if vsave='' then
              vsave=translate(strip(ropts.!ASSIGN))
        end
/* get varname's value */
        if val1='' then do
          jj1=wordpos(fvarname,info.!series_varnames)
          if jj1>0 then do
             fpp=note_error('W','Attempting to READFILE a CMS variable outside of a repeat block: 'fvarname)
             return ''       
          end
          jj2=wordpos(fvarname,cinfo.0)
          if jj2=0 then do
             fpp=note_error('W','Attempting to READFILE an undefined variable: 'fvarname)
             return ''     
          end
          vv='!'||strip(fvarname)
          val1=strip(CINFO.vv)
        end

        if pos(':',val1)>0 | abbrev(val1,'\')=1 then do  /* fully qualified */
           ffile=val1
        end
        else do
           ffile=usefdir||val1
        end 
        fstuff=sre_read_file(ffile,2,2,1)
        if vsave<>'' then do
          vv='!'||vsave
          CINFO.vv=fstuff
          if wordpos(vsave,CINFO.0)=0 then   CINFO.0=CINFO.0||' '||vsave
        end
        if nodisplay<>1  then return fstuff
        return ''
   end

   WHEN keyword="SORT" then do   /* specify sort parameters */
       call parse_options(opt1)  /* name=, order= */
       ann=strip(translate(ropts.!NAME))
       if ann<>'!ID' & info.!comments_mode=1 then do
            fpp=note_error('E','Not allowed to SORT on a variable when processing comments')
            return ''   
       end 

       if wordpos(ann,Translate('!ID '||stuff.!SPECIALS))=0 then do /* can only sort "sortable" variables */
            fpp=note_error('E','SORT variable ('ann') must be one of the "sortable" variables:' stuff.!specials)
            return ''  
       end 
       info.!sortit_name=ann           /* if no NAME=, then no sort will occur */
       info.!sortit_order=strip(translate(ropts.!order))  /* default is ascending */
       info.!sortit_type='I'    /* default is case insensitive */
       if ann<>'!ID' then do
         jj1=wordpos(ann,stuff.!VARNAMES)
         if jj1=0 then do /* should never happen */
            fpp=note_error('E','SORT variable ('ann') not in variable name list')
            return ''   
         end 
         info.!sortit_type=translate(strip(word(stuff.!vartypes,jj1)))
       end
/* note: when here, a first sort (with possible deselection) will have taken place */
       fpp=load_recids(info.!sortit_name,info.!sortit_order,info.!sortit_type)   /* this loads, sorts, & selects the "id7"s */
       return ''                                        /* note that load_recids checks for no change in sortorder */
   end 


   WHEN keyword="SHOW" then do
        parse upper var options lname lopts 
        lname=translate(space(lname,0))

        if lname='!ARCHIVE_TITLE' then return   CINFO.!!ARG||': '||info.!template_desc

        if lname='!DESC' then return info.!template_desc

        if lname='!MESSAGE' then lname='!ARG'
        if abbrev(lname,'$')=1 then do          /* this is a "system" variable? */
            tt=fig_sysvar(lname,lopts)
            return tt
        end 

        jj1=wordpos(lname,'!ID  '||info.!series_varnames)
        if jj1>0 then do
           fpp=note_error('W','Attempting to show CMS variable (use VAR instead): 'lname)
           return ''            /* use VAR or SERIES to display cms vars, ID, or SERIES name */
        end
        jj2=wordpos(lname,cinfo.0)
        if jj2=0 then do
           fpp=note_error('W','Attempting to show undefined variable: 'lname)
           return ''            /* use VAR or SERIES to display cms vars, ID, or SERIES name */
        end
        vv='!'||strip(lname)
        val1=CINFO.vv
        return val1
   end


   when abbrev(keyword,"SER")=1 then do
      select
        when abbrev(opt1,'NAME')=1 then return info.!series
        when abbrev(opt1,'DESC')=1 then return info.!series_desc
        when abbrev(opt1,'DEFFILE')=1 then return filespec('n',optname.!DEFFILE)
        when abbrev(opt1,'DEFFILE_FULL')=1 then return optname.!DEFFILE

        when abbrev(opt1,'VAR')=1 then do 
            return info.!series_varnames
        end 
        when abbrev(opt1,'SOR')=1 then do 
            return stuff.!Specials
        end 

        when abbrev(opt1,'REC')=1 then do
          if info.!COMMENTS_MODE<>1 then
             bbb=info.!series_records-4
          else
             bbb=info.!series_records-3
          return bbb
        end

        when abbrev(opt1,'MOD')=1 then do
          parse var info.!series_modified a1 '=' . ; return a1
        end
        when abbrev(opt1,'CRE')=1 then return info.!series_created
        otherwise do
           fpp=note_error('W','No such SERIES option: '||opt1)
           return ' '
        end
     end
   end 

   when (abbrev(keyword,"TEM")+abbrev(keyword,'TPL'))>0 then do
      select
        when abbrev(opt1,'DESC')=1 then return info.!template_desc
        when abbrev(opt1,'NAME')=1 then return filespec('n',info.!template_file)
        when abbrev(opt1,'FULLNAME')=1 then return info.!template_file
        when abbrev(opt1,'FILE')=1 then return info.!template_file
        otherwise do
           fpp=note_error('W','No such TEMPLATE option: '||opt1)
           return ' '
        end
     end
   end

   WHEN keyword="POLL" then do
       call do_poll_keyword 1
       ares=result
       return ares
   end

   WHEN keyword="VAR" then do
       call do_var_keyword 1
       ares=result
       return ares
   end

/* the next several are only allowed in REPEAT blocks */

   WHEN keyword="IF" then do  
        foo=note_error('W',' IF only allowed in REPEAT block ')
        return foo
   end

   WHEN keyword="/IF" then do  /* should never happen */
        foo=note_error('W',' /IF only allowed in REPEAT block ')
        return foo
   end 

   WHEN keyword="/ELSE"  | keyword="ELSE" then do  /* should never happen */
        foo=note_error('W',' /ELSE and ELSE only allowed in REPEAT block ')
        return foo
   end 


   WHEN keyword="A" then do  /* Error. A only allowed in REPEAT blocks */
        foo=note_error('W',' A only allowed in REPEAT block ')
        return foo
   end 

   WHEN keyword="LINK" then do  /* Error. LINK only allowed in REPEAT blocks */
        foo=note_error('W',' LINK only allowed in REPEAT block ')
        return foo
   end 

   WHEN keyword="COMMENT" then do  /* Error. COMMENT only allowed in REPEAT blocks */
        foo=note_error('W',' COMMENT only allowed in REPEAT block ')
        return foo
   end 

   WHEN keyword="LABEL" then do  /* Error. Labels only allowed in REPEAT blocks */
        foo=note_error('W',' LABEL only allowed in REPEAT block ')
        return foo
   end 

   WHEN keyword="GOTO" then do  /* Error. GOTO only allowed in REPEAT blocks */
        foo=note_error('W',' GOTO only allowed in REPEAT block ')
        return foo
   end 

   WHEN keyword="CASE" | keyword="/CASE"  | keyword="CASES" | keyword="/CASES" then do  /* Error */
        foo=note_error('W',' CASE, /CASE, CASES, and /CASES only allowed in REPEAT block ')
        return foo
   end 


   WHEN keyword="ITER" | keyword="DROP" | keyword="LEAVE" then do  /* Error. ITER, DROP and LEAVE only allowed in REPEAT blocks */
        foo=note_error('W',' ITER, DROP, and LEAVE only allowed in REPEAT block: 'keyword)
        return foo
   end 


   WHEN abbrev(keyword,"ITER")=1 then do  /* Error. Labels only allowed in REPEAT blocks */
        foo=note_error('W',' ITER only allowed in REPEAT block ')
        return foo
   end 

   WHEN abbrev(keyword,"LEAVE")=1 then do  /* Error. Labels only allowed in REPEAT blocks */
        foo=note_error('W',' LEAVE only allowed in REPEAT block ')
        return foo
   end 

/* end of "REPEAT block only" CMS commands */


   otherwise return '<!-- unknown CMS keyword: '||keyword||' --> '

end             /* keyword select */
return '<!-- error in process_cms -->'


/*******************/
do_poll_keyword:
parse arg no_option 
vverror=0
gotopts=0
poll_info.=''

/* is this not the first record? If so, then pull options from rep_stuff */
if rep_stuff.iarf2.!GOT2=1 then do
   varname=rep_stuff.iarf2.!VARNAME
   if varname='' then do
          foo=note_error('E',' CMS  poll variable lookup did not specify a VARNAME ')
          return ''
    end

   idvalue=rep_stuff.iarf2.!ID
   cnumber=rep_stuff.iarf2.!cnumber
   auto=rep_stuff.iarf2.!auto
   wtarget= rep_stuff.iarf2.!wtarget
   the_h3= rep_stuff.iarf2.!the_h3
   the_desc= rep_stuff.iarf2.!the_desc
   the_form= rep_stuff.iarf2.!the_form

   inf_responses= rep_stuff.iarf2.!RESPONSES
   inf_choices=rep_stuff.iarf2.!CHOICES

   inf_allstats= rep_stuff.iarf2.!allstats

   inf_chart= rep_stuff.iarf2.!chart
   inf_chartspec= rep_stuff.iarf2.!chartspec

   inf_mean= rep_stuff.iarf2.!mean
   inf_sum= rep_stuff.iarf2.!sum
   inf_median= rep_stuff.iarf2.!median
   inf_max= rep_stuff.iarf2.!max
   inf_sd= rep_stuff.iarf2.!sd
   inf_weight= rep_stuff.iarf2.!weight

   inf_HEADER=rep_stuff.iarf2.!HEADER
   inf_footer=rep_stuff.iarf2.!FOOTER 
   inf_color=rep_stuff.iarf2.!COLOR
   inf_Ct=rep_stuff.iarf2.!CT
   inf_store=rep_stuff.iarf2.!STORE 
   inf_DISP=rep_stuff.iarf2.!DISP 

   inf_bgcolor=rep_stuff.iarf2.!BGCOLOR

   svarname=rep_stuff.iarf2.!SVARNAME
   svartype=rep_stuff.iarf2.!SVARTYPE



/***  This observations' poll-def file may be the same as before... */
   nopts=strip(rep_stuff.iarf2.!vopts.!nopts)

   if nopts='' then do          /* force lookup next time*/
      rep_stuff.iarf2.!SFILE=''
   end
   else do
     do jj1=1 to nopts
        vopts.jj1=rep_stuff.iarf2.!vopts.jj1
        vopts.jj1.2=rep_stuff.iarf2.!vopts.jj1.2
     end 
     vopts.0=rep_stuff.iarf2.!vopts.0
     do jj1=1 to words(vopts.0)
        aw=strip(word(vopts.0,jj1))
        va1='!'||aw
        vopts.va1=rep_stuff.iarf2.!vopts.va1
     end
     gotopts=1
   end

end

if gotopts=0 then do           /* read the options the hard way, then save */
      call parse_options(options)
      varname=strip(translate(ropts.!VAR))
      if varname='' then do
          foo=note_error('E',' CMS  poll variable lookup did not specify a VARNAME ')
          return ''
       end

      idvalue=ropts.!ID
      cnumber=ropts.!CNUMBER  
      auto=strip(translate(ropts.!AUTO))

      wtarget=ropts.!TARGET
      the_h3=ropts.!h3
      the_desc=ropts.!desc
      the_form=ropts.!form

      stats=strip(translate(ropts.!STATS))

      inf_choices=strip(translate(ropts.!CHOICES))
      if inf_choices='' & wordpos('CHOICES',ropts.0)>0 then inf_choices=1

      inf_responses=strip(translate(ropts.!RESPONSES))
      if inf_responses='' & wordpos('RESPONSES',ropts.0)>0 then inf_responses=1

      inf_allstats=strip(translate(ropts.!ALLSTATS))
      if inf_allstats='' & wordpos('ALLSTATS',ropts.0)>0 then inf_allstats=1

      inf_mean=strip(translate(ropts.!mean))
      if inf_mean='' & wordpos('MEAN',ropts.0)>0 then inf_mean=1

      inf_chart=strip(translate(ropts.!chart))
      if inf_chart='' & wordpos('CHART',ropts.0)>0 then inf_chart=1
      inf_chartspec=strip(translate(ropts.!chartspec))

      inf_sum=strip(translate(ropts.!sum))
      if inf_sum='' & wordpos('SUM',ropts.0)>0 then inf_sum=1

      inf_median=strip(translate(ropts.!median))
      if inf_median='' & wordpos('MEDIAN',ropts.0)>0 then inf_median=1

      inf_sd=strip(translate(ropts.!sd))
      if inf_sd='' & wordpos('SD',ropts.0)>0 then inf_sd=1

      inf_weight=strip(translate(ropts.!weight))
      if inf_weight='' & wordpos('WEIGHT',ropts.0)>0 then inf_weight=1

      inf_max=strip(translate(ropts.!max))
      if inf_max='' & wordpos('MAX',ropts.0)>0 then inf_max=1

      inf_header=strip(translate(ropts.!HEADER))
      if inf_header='' & wordpos('HEADER',ropts.0)>0 then inf_header=1

      inf_footer=strip(translate(ropts.!FOOTER))
      if inf_FOOTER='' & wordpos('FOOTER',ropts.0)>0 then inf_FOOTER=1

      inf_color=strip(translate(ropts.!COLOR))
      if inf_COLOR='' & wordpos('COLOR',ropts.0)>0 then inf_COLOR=1

      inf_CT=strip(translate(ropts.!CT))
      if inf_CT='' & wordpos('CT',ropts.0)>0 then inf_CT=1

      inf_store=strip(translate(ropts.!STORE))
      if inf_STORE='' & wordpos('STORE',ropts.0)>0 then inf_STORE=1

      inf_bgcolor=strip(translate(ropts.!BGCOLOR))
   
      inf_disp=strip(translate(ropts.!DISP))
      if inf_DISP='' & wordpos('DISP',ropts.0)>0 then inf_DISP=1

      rep_stuff.iarf2.!STATS=stats

      rep_stuff.iarf2.!VARNAME=strip(translate(varname))
      rep_stuff.iarf2.!ID =strip(translate(idvalue))
      rep_stuff.iarf2.!cnumber =strip(translate(cnumber))
      rep_stuff.iarf2.!AUTO =auto
      rep_stuff.iarf2.!WTARGET = wtarget
      rep_stuff.iarf2.!THE_H3 = the_h3
      rep_stuff.iarf2.!THE_desc = the_desc
      rep_stuff.iarf2.!THE_form = the_form

      rep_stuff.iarf2.!RESPONSES =strip(translate(inf_responses))
      rep_stuff.iarf2.!CHOICES =strip(translate(inf_choices))

      rep_stuff.iarf2.!allstats=strip(translate(inf_allstats))
      rep_stuff.iarf2.!mean=strip(translate(inf_mean))
      rep_stuff.iarf2.!sum=strip(translate(inf_sum))
      rep_stuff.iarf2.!chart=strip(translate(inf_chart))
      rep_stuff.iarf2.!chartspec=strip(translate(inf_chartspec))

      rep_stuff.iarf2.!median=strip(translate(inf_median))
      rep_stuff.iarf2.!max=strip(translate(inf_max))
      rep_stuff.iarf2.!sd=strip(translate(inf_sd))
      rep_stuff.iarf2.!weight=strip(translate(inf_weight))
      rep_stuff.iarf2.!HEADER =strip(translate(inf_HEADER))
      rep_stuff.iarf2.!FOOTER =strip(translate(inf_footer))
      rep_stuff.iarf2.!COLOR =strip(translate(inf_color))
      rep_stuff.iarf2.!CT =strip(translate(inf_Ct))
      rep_stuff.iarf2.!STORE =strip(translate(inf_store))
      rep_stuff.iarf2.!BGCOLOR=strip(Translate(inf_bgcolor))
      rep_stuff.iarf2.!DISP =strip(translate(inf_DISP))
      rep_stuff.iarf2.!GOT2=1

      call get_svarname
      if vverror=1 then return ''
      rep_stuff.iarf2.!SVARNAME=strip(translate(svarname))
      rep_stuff.iarf2.!SVARTYPE=strip(translate(svartype))
      rep_stuff.iarf2.!GOT2=1
     
/* note: get_polldef_stuff will define the rep_stuff.iarf2.vopts stuff */
 
end

/* note: ID= in a VAR supercedes the "current" id (as provided within a REPEAT) */
/*  we do this to allow ID= to supercede REPEAT loop id */

is7=''            
if idvalue='' then do   /* no ID specified in option list, so use useID7? */
    if  useid7='' then do
          foo=note_error('E',' CMS  poll variable lookup did not specify an ID ')
          return ''
    end
    findid=useid7
    is7=' ID7 '
end 
else do                  /* this is the record provided in the <? POLL ID=... */
     findid=idvalue
end



/* lookup cms dataseries variable */

if findid='NONE'  then do ' '   /* NONE can be used when generic info is required */
   vval='' ; totresponses='' ; counts='' ; poll_misc=''
end
else do
   astat=sre_cache_data_local(info.!series,'FIND 'is7,findid,svarname)
   parse var astat astat0 astat
   if astat0<>0 then do
      amess=sre_cache_data_local('FIND','ERROR',astat0,svarname)
      foo=note_error('E','CMS POLL lookup ('svarname'): '||amess)
      return ''
   end 
   parse var astat . '0d0a'x  vval

   parse var vval totresponses counts ':' poll_misc

end

call get_polldef_stuff

if nopts='' then do
     foo=note_error('W',' CMS poll variable had missing # options: ' findid ','varname)
     return ''
end

chartspec=''
   parse var inf_chartspec chartonly ',' i250 ',' iheight ',' gcolors
   if datatype(i250)<>'NUM' then i250=250
   if datatype(iheight)<>'NUM' then iheight=20
   chartspec='&CHARTSPEC='||chartonly||','||strip(i250)||','||strip(iheight)||','||strip(gcolors)
   if chartonly='' | strip(chartonly)=1 then chartonly='CRS'

/* At last.... we can resolve this element */

/* AUTO display the results? */

if  auto='SHOW' | abbrev(auto,'DISP')=1 then do
 outstuff='<table border=1 bgcolor="'||vopts.!color||'">'||'0d0a'x

 if vopts.!header<>'' then do
    outstuff=outstuff||'<tr><td colspan=2>'||vopts.!HEADER||'</td>'||'0d0a'x
 end 

 outstuff=outstuff||'<tr><td>'
 outstuff=outstuff||'<font size=-1>updated '||date('n')||' '||time('n')||'</font><br>'
 outstuff=outstuff||'<table border=3>'||'0d0a'x
 outstuff=outstuff||'<tr><td colspan=2>'||'0d0a'x
 tt='!?'
 outstuff=outstuff||'<b>'||vopts.tt||'</b></td>'||'0d0a'x
 if pos('R',chartonly)>0 then do
   do cc=1 to nopts
     dval=vopts.cc.2
     cct=strip(word(counts,cc))
     outstuff=outstuff||'<tr align="right"><td >'||dval||'&nbsp;&nbsp;</td><td><tt>'||cct||'</tt></td>'||'0d0a'x
   end
 end

/* 
   Use the various "stats"  INFOoptions: 
          ALL MEAN SUM MEDIAN MAX SD WEIGHT RESPONSES CHART
    MEAN, SUM, and SD (standard deviation) are computed using positions as
    the value (i.e.; first option has value of 1, 2nd option has value of 2,..)
    However, if WEIGHT is specified, then the STORED value is used as the
    value for  MEAN, SUM, and SD.
    MEDIAN is the median position. WEIGHT is not used.
    MAX is the position with the largest count.
*/    


 call do_stats    /* will append to outstuff (in a new <TR> )*/

 outstuff=outstuff||'</table>'||'0da0'x

 if vopts.!footer<>'' then do
    outstuff=outstuff||'<tr><td colspan=2>'||vopts.!footer||'</td>'||'0d0a'x
 end 

 outstuff=outstuff||'</table>'||'0d0a'x
 return outstuff
end                     /*   ----- AUTO=SHOW --- */

/* Create a link to CMS_POLL (to create a questionaire */

if abbrev(auto,'ASK')=1 | abbrev(auto,'QUEST')=1  | auto='?' then do
 atarget='' ;  btarget=''
 if wtarget<>'' then do
    atarget=' target="'||wtarget||'" '
    btarget='&target=1'
  end

 ah3=''
 if the_h3<>'' then do
     ii=pos('$VARNAME=',translate(the_h3))
     ah3=the_h3
     if ii>0 then do
        p1='' ; if ii>1 then p1=left(the_h3,ii-1) 
        p2=substr(the_h3,ii)
        parse var p2 . '=' avarname p3
        avarname=strip(translate(avarname))
        if wordpos(avarname,'!ID '||info.!series_varnames)>0 then do
           vurl=process_cms('VAR ',,findid,avarname)
        end
        else do
           vurl=process_cms('SHOW ',avarname)
        end 
        ah3=p1||vurl||' '||p3
     end 
     ah3='&h3='||translate(ah3,'+',' ')
 end

 adesc=''
 if the_desc<>'' then do
     ii=pos('$VARNAME=',translate(the_desc))
     adesc=the_desc
     if ii>0 then do
        p1='' ; if ii>1 then p1=left(the_desc,ii-1) 
        p2=substr(the_desc,ii)
        parse var p2 . '=' avarname p3
        avarname=strip(translate(avarname))
        if wordpos(avarname,'!ID '||info.!series_varnames)>0 then do
           vurl=process_cms('VAR ',,findid,avarname)
        end
        else do
           vurl=process_cms('SHOW ',avarname)
        end 
        adesc=p1||vurl||' '||p3
     end 
     adesc='&desc='||sre_packur_make(translate(adesc,'+',' '),2,'&#"')
 end

 aform=''
 if the_form<>'' then do
     ii=pos('$VARNAME=',translate(the_form))
     aform=the_form
     aform='&form='||translate(aform,'+',' ')
 end

 abgcolor=''
 if inf_bgcolor<>""  then do
      abgcolor='&bgcolor='||sre_packur_make(inf_bgcolor,2,'#')
 end 

 outstuff='<A '||ropts.!2||' '||atarget||' href="/CMS/CMS_POLL.CMD?action=21'||ah3||btarget||adesc||aform||abgcolor
 
 outstuff=outstuff||'&ID='||findid||'&deffile='||daindx||'&pfile='||sfile0||'&var='||svarname||'">'

 return outstuff

end

/* Create a link to CMS_POLL (to update poll results */
if abbrev(auto,'UPD')=1 then do
 atarget='' ;  btarget=''
 if wtarget<>'' then do
    atarget=' target="'||wtarget||'" '
    btarget='&target=1'
  end


 ah3=''
 if the_h3<>'' then do
     ii=pos('$VARNAME=',translate(the_h3))
     ah3=the_h3
     if ii>0 then do
        p1='' ; if ii>1 then p1=left(the_h3,ii-1) 
        p2=substr(the_h3,ii)
        parse var p2 . '=' avarname p3
        avarname=strip(translate(avarname))
        if wordpos(avarname,'!ID '||info.!series_varnames)>0 then do
           vurl=process_cms('VAR ',,findid,avarname)
        end
        else do
           vurl=process_cms('SHOW ',avarname)
        end 
        ah3=p1||vurl||' '||p3
     end 
     ah3='&h3='||translate(ah3,'+',' ')
 end

 adesc=''
 if the_desc<>'' then do
     ii=pos('$VARNAME=',translate(the_desc))
     adesc=the_desc
     if ii>0 then do
        p1='' ; if ii>1 then p1=left(the_desc,ii-1) 
        p2=substr(the_desc,ii)
        parse var p2 . '=' avarname p3
        avarname=strip(translate(avarname))
        if wordpos(avarname,'!ID '||info.!series_varnames)>0 then do
           vurl=process_cms('VAR ',,findid,avarname)
        end
        else do
           vurl=process_cms('SHOW ',avarname)
        end 
        adesc=p1||vurl||' '||p3
     end 
     adesc='&desc='||sre_packur_make(translate(adesc,'+',' '),2,'&#"')
 end

 abgcolor=''
 if inf_bgcolor<>""  then do
      abgcolor='&bgcolor='||sre_packur_make(inf_bgcolor,2,'#')
 end 

 outstuff='<A '||ropts.!2||' '||atarget||' href="/CMS/CMS_POLL.CMD?action=31'||ah3||btarget||chartspec||adesc||abgcolor
 outstuff=outstuff||'&ID='||findid||'&deffile='||daindx||'&pfile='||sfile0||'&var='||svarname||'">'
 return outstuff
end



/* otherwise, one of the info-options */
if cnumber<>'' then do
   if datatype(cnumber)<>'NUM' then do  /* possibly a SET variable */
        cnumber=process_cms('SHOW ',cnumber)
   end 

   if datatype(cnumber)<>'NUM' then do
      foo=note_error('E','CMS POLL requested bad CNUMBER: '||cnumber)
      return ''
   end        
   if cnumber<0 | cnumber>nopts then do
      foo=note_error('E','CMS POLL requested out-of-bounds CNUMBER: '||cnumber)
      return ''
   end
end

/*  CHoices     : # of choices for this poll
    RESPonses : current total number of responses
    HEADer    : the header (if one was defined)
    FOOTer    : the footer (if one was defined)
    COLor     : the background color (if one was defined)
(choice specific info -- requires CNUMBER)                         
    CT        : current number of responses (that chose this choice)
    STORE     : the "stored" value for this choice
    DISP      : the "displayed value" for this choice 
(stats)
    ALLSTAT   : do all the stats &  responses
    WEIGHT    : used "stored value" when computing MEAN, SUM, and SD
    MEAN      : mean
    MEDIAN    : median (by choice #)
    MAX       : max (by choice #)
    SUM       : sum
    SD        : standard deviation
    CHART     : make a simple bar chart
*/

outstuff=''
if inf_choices<>'' then do
   if inf_choices=1 then do
        outstuff=outstuff||nopts
   end
   else do
      vv='!'||inf_choices
      poll_info.vv=nopts
      poll_info.0=poll_info.0||' '||inf_choices
   end 
end
if inf_responses<>'' then do
   if inf_responses=1 then do
        outstuff=outstuff||totresponses
   end
   else do
      vv='!'||inf_responses
      poll_info.vv=totresponses
      poll_info.0=poll_info.0||' '||inf_responses
   end 
end

if inf_header<>'' then do
   if inf_header=1 then do
        outstuff=outstuff||vopts.!header
   end
   else do
      vv='!'||inf_header
      poll_info.vv=vopts.!header
      poll_info.0=poll_info.0||' '||inf_header
   end 
end


if inf_footer<>'' then do
   if inf_footer=1 then do
        outstuff=outstuff||vopts.!footer
   end
   else do
      vv='!'||inf_footer
      poll_info.vv=vopts.!footer
      poll_info.0=poll_info.0||' '||inf_footer
   end 
end

 
if inf_color<>'' then do
   if inf_color=1 then do
        outstuff=outstuff||vopts.!color
   end
   else do
      vv='!'||inf_color
      poll_info.vv=vopts.!color
      poll_info.0=poll_info.0||' '||inf_color
   end 
end

if inf_store<>'' then do
   if datatype(cnumber)<>'NUM' then do
      foo=note_error('E','CMS POLL. STORE requested,  but no CNUMBER ')
      return ''
   end
   else do
     if inf_store=1 then do
        outstuff=outstuff||vopts.cnumber
     end
     else do
       vv='!'||inf_store
       poll_info.vv=vopts.cnumber
       poll_info.0=poll_info.0||' '||inf_store
     end 
   end
end
       
if inf_disp<>'' then do
   if datatype(cnumber)<>'NUM' then do
      foo=note_error('E','CMS POLL. DISP requested,  but no CNUMBER ')
      return ''
   end
   else do

    if inf_disp=1 then do
        outstuff=outstuff||vopts.cnumber.2
    end
    else do
      vv='!'||inf_disp
      poll_info.vv=vopts.cnumber.2
      poll_info.0=poll_info.0||' '||inf_disp
    end 
  end
end
         
if inf_ct<>'' then do
   if datatype(cnumber)<>'NUM' then do
      foo=note_error('E','CMS POLL. CT requested,  but no CNUMBER ')
      return ''
   end
   else do
     if inf_ct=1 then do
        outstuff=outstuff||word(counts,cnumber)
    end
    else do
      vv='!'||inf_ct
      poll_info.vv=word(counts,cnumber)
      poll_info.0=poll_info.0||' '||inf_ct
    end 
  end
end

call do_stats2
 

return outstuff         /* also, poll_info. is set */


/******************/
/* read poll file info  (for var svarname, element iarf2) */
get_polldef_stuff: 
parse var poll_misc sfile0 ',' squest


sfile0=translate(strip(sfile0))
sfile=sfile0

if words(sfile)<>1 then do
     foo=note_error('W',' Bad or missing poll-def file for 'findid','varname': 'sfile0)
     vopts.=''
     vopts.!nopts=0
     nopts=''
     rep_stuff.iarf2.!SFILE=''
     return ''
end 

/* note that on first read of iarf2, !SFILE will='' */

if rep_stuff.iarf2.!SFILE=sfile0 & rep_stuff.iarf2.!SFILE<>"" then do
  return 1       /* use stored value */
end

usedir=cmsdir_pol
if abbrev(sfile,'~')=1 then do
   sfile=substr(sfile,2)
   usedir=stuff.!filedir||'\POLLS'
end 
pfile=usedir||'\'||strip(sfile)||'.POL'
if stream(pfile,'c','query exists')='' then do
    foo=note_error('E',' No such CMS poll-def file, 'pfile', for variable 'varname)
    return ''
end

rep_stuff.iarf2.!SFILE=sfile0
rep_stuff.iarf2.!POLL_FILE=pfile

polstuff=sre_read_file(pfile,2,2,1)

vopts.=''
do jj=1 to words(rep_stuff.iarf2.!vopts.0)
   vv1=strip(Translate(word(rep_stuff.iarf2.!vopts.0,jj)))
   vv11='!'||vv1
   rep_stuff.iarf2.!vopts.vv11=''
end 
REP_STUFF.IARF2.!VOPTS.0=' '
do until polstuff=''
     parse var polstuff aline '0d0a'x polstuff
     aline=strip(aline)
     if aline='' then iterate
     if abbrev(aline,';')=1 then iterate
     parse var aline a1 ':' a2
     a1=strip(translate(a1))
     va1='!'||a1
     if rep_stuff.iarf2.!vopts.va1='' then do  /* first occurence of this option */
        rep_stuff.iarf2.!vopts.va1=strip(a2)
        vopts.va1= rep_stuff.iarf2.!vopts.va1
        rep_stuff.iarf2.!vopts.0=rep_stuff.iarf2.!vopts.0||' '||a1
        vopts.0=rep_stuff.iarf2.!vopts.0
     end
     else do            /* add 2nd or later line */
        rep_stuff.iarf2.!vopts.va1=rep_stuff.iarf2.!vopts.va1||'0d0a'x||strip(a2)
        vopts.va1=rep_stuff.iarf2.!vopts.va1
     end
end

nopts=0
do jj1=1 to words(vopts.0)
   bw=strip(word(vopts.0,jj1))
   if datatype(bw)<>'NUM' then iterate
   vbw='!'||bw
   dval=vopts.vbw

   nopts=nopts+1
   vopts.!nopts=nopts
   rep_stuff.iarf2.!vopts.!nopts=nopts

   rep_stuff.iarf2.!vopts.nopts=bw
   vopts.nopts=bw

   rep_stuff.iarf2.!vopts.nopts.2=dval
   vopts.nopts.2=dval
end      

return 1


/*****************/
/* compute some stats for current poll-variable values. Append to outstuff */
do_stats:

/* if no stats specified, just return # of responses */
if inf_allstats='' & inf_responses='' & inf_mean=''  & inf_sum='' & ,
    inf_median='' & inf_max='' & inf_sd='' & inf_chart='' then do
    if pos('S',chartonly)>0 then
      outstuff=outstuff||'<tr><td colspan=2><tt>Total # of responses='||totresponses||'</tt></td>'||'0d0a'x
    return 1
end

/* compute the various stats  */  
outstuff=outstuff||'<tr><td align="left" colspan=2>'||'0d0a'x
outstuff=outstuff||'<hr>'||'0d0a'x

doall=0 ; if inf_allstats<>'' then doall=1
do jj=1 to nopts
       dacount.jj=strip(word(counts,jj))
       if datatype(dacount.jj)<>'NUM' then dacount.jj=0
       wt.jj=jj
       if inf_weight<>'' then do
          aw=vopts.jj
          if datatype(aw)='NUM' then wt.jj=aw
       end
end
       
tsum=0 ; tsum2=0 ; totr=0 ; amax=1
do jj=1 to nopts
        totr=totr+dacount.jj
        if dacount.jj>dacount.amax then amax=jj
        runsum.jj=totr
        t1=(dacount.jj*wt.jj)
        tsum=tsum+t1
        tsum2=tsum2+(t1*t1)
end
if totr=0 then do
  amean=0 ;asd=0;amedian=1;
end
else do
  amean=tsum/totr
  if totr>1 then
     asd=(tsum2-(totr*(amean*amean)))/(totr-1)
  else
     asd=0
  med2=totr/2; amedian=nopts
  do jj=1 to nopts
       if runsum.jj>med2 then do
            amedian=jj
            leave
       end
  end
end

/* got the stats. What to output? */
outstuff=outstuff||'<ul>'||'0d0a'x
sayw=''
   if inf_weight<>'' then sayw=' (weighted) '

   if inf_allstats<>'' | inf_responses<>'' then
      if pos('S',chartonly)>0 then outstuff=outstuff||'<li>Total # of responses: <tt>'||totr||'</tt>'||'0d0a'x
   if inf_allstats<>'' | inf_sum<>'' then
      if pos('S',chartonly)>0 then       outstuff=outstuff||'<li>'sayw' Sum: <tt>'||tsum||'</tt>'||'0d0a'x
   if inf_allstats<>'' | inf_mean<>'' then
      if pos('S',chartonly)>0 then       outstuff=outstuff||'<li>'sayw' Mean: <tt>'||amean||'</tt>'||'0d0a'x
   if inf_allstats<>'' | inf_sd<>'' then
      if pos('S',chartonly)>0 then     outstuff=outstuff||'<li>'sayw' StdDev: <tt>'||asd||'</tt>'||'0d0a'x

   if inf_allstats<>'' | inf_median<>'' then
      if pos('S',chartonly)>0 then    outstuff=outstuff||'<li> Median: <tt>'||amedian||' (out of '||nopts||')</tt>'||'0d0a'x
   if inf_allstats<>'' | inf_max<>'' then
      if pos('S',chartonly)>0 then    outstuff=outstuff||'<li> Max: <tt> choice # '||amax||'</tt>'||'0d0a'x
   
      if pos('S',chartonly)>0 then outstuff=outstuff||'</ul>'||'0d0a'x

if (inf_chart<>'' | inf_allstats<>'' ) & (totr>0)  & (pos("C",chartonly)>0 ) then do

   chartspec='&CHARTSPEC='||chartonly||','||strip(i250)||','||strip(iheight)||','||strip(gcolors)

   ithc=0
   if gcolors<>'' then do
     gcolors.=''
     ithc=0
     do until gcolors=''
        parse var gcolors aco ',' gcolors
        if aco='' then leave
        ithc=ithc+1
        gcolors.ithc=aco
     end 
   end 


   outstuff=outstuff||'<p><table align="left" border=1 cellpadding=0 cellspacing=0 >'
   outstuff=outstuff||'<tr><td><em>choices</em></td>'||'0d0a'x
   outstuff=outstuff||'<td><em>count</em>&nbsp;&nbsp;</td>'||'0d0a'x
   outstuff=outstuff||'<td align="middle"><img align="middle" alt="chart_top" height=2 '||,
             ' width='||i250||' src="/cms/cms_poll.cmd?action=GIF&color=BLACK_BOX&file.gif">'||,
             '</td>'||'0d0a'x

   cnn=0
   do nn=1 to nopts
      cnn=cnn+1
      if ithc>0 then do
         if cnn>ithc then cnn=1
         colordo=gcolors.cnn
      end
      else do
         colordo=cnn
      end 
      outstuff=outstuff||'<tr><td >'||vopts.nn.2||'</td>'||'0d0a'x
      outstuff=outstuff||'<td><tt>'||dacount.nn||'</td>'||'0d0a'x 
      ilen=trunc((dacount.nn/totr)*i250)
      if ilen>0 then
        outstuff=outstuff||'<td align="middle"><img align="middle" alt="chart_bar" height='||iheight||' '||,
             ' width='||ilen||' src="/cms/cms_poll.cmd?action=GIF&color='||colordo||'&file.gif">'||,
             '</td>'||'0d0a'x
      else
        outstuff=outstuff||'<td>&nbsp;</td>'||'0d0a'x
   end 
   outstuff=outstuff||'</table><p>'
end

outstuff=outstuff||'</td>'||'0d0a'x


return 1



/*****************/
/* compute some stats for current poll-variable values. WRite to output or save  */
do_stats2:

/* if no stats specified, skip */
if inf_allstats='' & inf_responses='' & inf_mean=''  & inf_sum='' & ,
    inf_median='' & inf_max='' & inf_sd='' then return 1

doall=0 ; if inf_allstats<>'' then doall=1
do jj=1 to nopts
       dacount.jj=strip(word(counts,jj))
       if datatype(dacount.jj)<>'NUM' then dacount.jj=0
       wt.jj=jj
       if inf_weight<>'' then do
          aw=vopts.jj
          if datatype(aw)='NUM' then wt.jj=aw
       end               
end
       
tsum=0 ; tsum2=0 ; totr=0 ; amax=1
do jj=1 to nopts
        totr=totr+dacount.jj
        if dacount.jj>dacount.amax then amax=jj
        runsum.jj=totr
        t1=(dacount.jj*wt.jj)
        tsum=tsum+t1
        tsum2=tsum2+(t1*t1)
end
if totr=0 then do
  amean=0 ;asd=0;amedian=1;
end
else do
  amean=tsum/totr
  if totr>1 then
     asd=(tsum2-(totr*(amean*amean)))/(totr-1)
  else
     asd=0
  med2=totr/2; amedian=nopts
  do jj=1 to nopts
       if runsum.jj>med2 then do
            amedian=jj
            leave
       end
  end
end

/* got the stats. What to output? */

if inf_mean<>'' then do
   if inf_mean=1 then do
        outstuff=outstuff||amean
   end
   else do
      vv='!'||inf_mean
      poll_info.vv=amean
      poll_info.0=poll_info.0||' '||inf_mean
   end 
end

if inf_sum<>'' then do
   if inf_sum=1 then do
        outstuff=outstuff||tsum
   end
   else do
      vv='!'||inf_sum
      poll_info.vv=tsum
      poll_info.0=poll_info.0||' '||inf_sum
   end 
end

if inf_sd<>'' then do
   if inf_sd=1 then do
        outstuff=outstuff||asd
   end
   else do
      vv='!'||inf_sd
      poll_info.vv=asd
      poll_info.0=poll_info.0||' '||inf_sd
   end 
end

if inf_max<>'' then do
   if inf_max=1 then do
        outstuff=outstuff||amax
   end
   else do
      vv='!'||inf_max
      poll_info.vv=amax
      poll_info.0=poll_info.0||' '||inf_max
   end 
end

if inf_median<>'' then do
   if inf_median=1 then do
        outstuff=outstuff||amedian
   end
   else do
      vv='!'||inf_median
      poll_info.vv=amedian
      poll_info.0=poll_info.0||' '||inf_median
   end 
end

if inf_responses<>'' then do
   if inf_responses=1 then do
        outstuff=outstuff||totr
   end
   else do
      vv='!'||inf_responses
      poll_info.vv=totr
      if wordpos(inf_responses,poll_info.0)=0 then 
        poll_info.0=poll_info.0||' '||inf_responses
   end 
end



if inf_chart<>'' & totr>0 then do

   ithc=0
   if gcolors<>'' then do
     gcolors.=''
     ithc=0
     do until gcolors=''
        parse var gcolors aco ',' gcolors
        if aco='' then leave
        ithc=ithc+1
        gcolors.ithc=aco
     end 
   end 

   tt1='<p><table align="left" border=1 cellpadding=0 cellspacing=0 >'
   tt1=tt1||'<tr><td><em>choices</em></td><td align="middle"><img align="middle" alt="chart_top" height=2 '||,
             ' width='||i250||' src="/cms/cms_poll.cmd?action=GIF&color=BLACK_BOX&file.gif">'||,
             '</td>'||'0d0a'x
   cnn=0
   do nn=1 to nopts
      cnn=cnn+1
      if ithc>0 then do
         if cnn>ithc then cnn=1
         colordo=gcolors.cnn
      end
      else do
         colordo=cnn
      end 
      tt1=tt1||'<tr><td >'||vopts.nn.2||'</td>'||'0d0a'x
      ilen=trunc((dacount.nn/totr)*i250)
      if ilen>0 then
        tt1=tt1||'<td align="middle"><img align="middle" alt="chart_bar" height='||iheight||' '||,
             ' width='||ilen||' src="/cms/cms_poll.cmd?action=GIF&color='||colordo||'&file.gif">'||,
             '</td>'||'0d0a'x
      else
        tt1=tt1||'<td>&nbsp;</td>'||'0d0a'x
   end 
   tt1=tt1||'</table><p>'

   if inf_chart=1 then do
        outstuff=outstuff||tt1
   end
   else do
      vv='!'||inf_chart
      poll_info.vv=tt1
      poll_info.0=poll_info.0||' '||inf_chart
   end 
end



return 1



/*******************/
do_var_keyword:
parse arg no_option 

vverror=0
idvalue='' ; vfilename_local='' ; vfilename='' ; tfmt='N'
gotopts=0 ; func_type=0
funcn=''                /* use variable as is (not a var function) */
is_lookup=0
if varname<>'' | no_option=1 then is_lookup=1  /* is_lookup=1 means "this is a simple lookup, so nothing fancy" */
assignto=''             /* returned to caller as a global */
nodisplay=0             /* returned to caller as a global */

/* is this not the first record? If so, then pull options from rep_stuff */
if is_lookup=0 then do 
 if rep_stuff.iarf2.!GOT2=1 then do
   varname=rep_stuff.iarf2.!VARNAME 
   vfilename=rep_stuff.iarf2.!VFILENAME
   vfilename_local=rep_stuff.iarf2.!VFILENAME_LOCAL
   idvalue=rep_stuff.iarf2.!IDVALUE
   funcn=rep_stuff.iarf2.!FUNCN
   vopts=rep_stuff.iarf2.!VOPTS
   assignto=rep_stuff.iarf2.!ASSIGN
   nodisplay=rep_stuff.iarf2.!NODISPLAY
   func_type=rep_stuff.iarf2.!FUNC_TYPE
   gotopts=1
 end
end

/*note: gotopts can only be one if this is NOT the first record processed in the REPEAT block,
        AND if varname was not provided (if this is NOT a "lookup a variable" call */

/* note: ID= in a VAR supercedes the "current" id (as provided within a REPEAT) */
if gotopts=0 then do
if varname='' then do            /* varname not provided by calling routine */
   if pos('=',opt1)=0 then do    /* just the variable name in option list*/
         parse var options varname .
   end 
   else do                               /* possibly other stuff in option list */
      call parse_options(options)
      varname=ropts.!NAME
      vfilename=ropts.!FILENAME
      vfilename_local=ropts.!FILENAME_LOCAL
      if varname<>'' & (vfilename_local<>'' | vfilename<>'' ) then do
          fpp=note_error('W','NAME=xxx and FILENAME=xxx both specified; FILENAME is ignored ')
          vfilename=''
      end 
      if varname='' & vfilename<>'' then varname=vfilename  
      if varname='' & vfilename_local<>'' then varname=vfilename_local  
      assignto=ropts.!ASSIGN
      if assignto='' then assignto=ropts.!SAVE
      nodisplay=ropts.!NODISPLAY
      idvalue=ropts.!ID
   end

   if is_lookup=0 then do
     rep_stuff.iarf2.!VARNAME=varname
     rep_stuff.iarf2.!VFILENAME=vfilename
     rep_stuff.iarf2.!VFILENAME_local=vfilename_local
     rep_stuff.iarf2.!IDVALUE=idvalue
     rep_stuff.iarf2.!GOT2=1
     rep_stuff.iarf2.!ASSIGN=assignto
     rep_stuff.iarf2.!NODISPLAY=nodisplay
     rep_stuff.iarf2.!GOT2=1
   end
end                      /* note: if varname provided, then id7 MUST be provided */
end


varname0=varname
varname=translate(strip(varname))

/* special cases */
if (varname='$ABSTRACT' | varname='$STATUS') &  info.!COMMENTS_MODE=1 then do
  if varname='$ABSTRACT' then  return info.!comments_mode.!ABSTRACT
  return info.!comments_mode.!STATUS
end                      /* a $ variable */


if gotopts=0 then do
  if pos('(',varname)=0 then do
     call get_svarname
     if vverror=1 then return ''
     if is_lookup=0 then do
       rep_stuff.iarf2.!SVARNAME=svarname
       rep_stuff.iarf2.!SVARTYPE=svartype
       rep_stuff.iarf2.!GOT2=1
     end
  end
end
else do         /* read from saved options */
  svarname=rep_stuff.iarf2.!SVARNAME
  svartype=rep_stuff.iarf2.!SVARTYPE
end


is7=''            
if idvalue='' then do   /* no ID specified in option list, so use useID7? */
    if  useid7='' then do
          foo=note_error('E',' CMS variable lookup did not specify an ID ')
          return ''
    end
    findid=useid7
    is7=' ID7 '
end 
else do                  /* this is the record provided in the <? VAR ID=... */
     findid=idvalue
end
if findid=0 | findid='NONE' then return ' '   /* NONE or 0 means suppress */

/* Note: we did this to allow ID= to supercede REPEAT loop id */

if gotopts=0 then do
  if pos('(',varname)>0 then do  /* meta-data, or function on variable's value, for this record */
    parse var varname0 funcn '(' vopts ')'  ; funcn=translate(strip(funcn))
    if wordpos(funcn,'TIME DATE STATUS')=0 then do    /* not a meta-function? */
        parse var vopts t1 ',' t2 ',' .
        select                                       /* use this varname below */
            when funcn="REPLACE" | funcn='LENGTH' | funcn='SUBSTR' | funcn='SUBWORD' | funcn='SUBLINE' then
                varname=strip(Translate(t1))
            when funcn='POS' | funcn='WORDPOS' then varname=strip(Translate(t2))
            when funcn='DATETIME' then  do
                if t2<>'' then
                  varname=strip(Translate(t2))
                else
                  varname='!ID'
            end
            otherwise   varname=strip(Translate(t1))              
        end
        call get_svarname
        if vverror=1 then return ''
        if is_lookup=0 then do
          rep_stuff.iarf2.!VARNAME=varname
          rep_stuff.iarf2.!SVARNAME=svarname
          rep_stuff.iarf2.!SVARTYPE=svartype
          rep_stuff.iarf2.!funcn=funcn
          rep_stuff.iarf2.!VOPTS=vopts
          rep_stuff.iarf2.!GOT2=1
        end
        func_type=1
    end
    else do
        func_type=2
    end
    if is_lookup=0 then  rep_stuff.iarf2.!FUNC_TYPE=func_type
  end
end             /* gotopts */

if func_type=2 then do      /* this is a meta-function (data about record) */
  if gotopts=0 then do
      parse var varname avar '(' aopts0 ')' 
      svarname='!ID'
      if is_lookup=0 then do
         rep_stuff.iarf2.!AVAR=avar
         rep_stuff.iarf2.!AOPTS0=aopts0
         rep_stuff.iarf2.!SVARNAME=svarname
         rep_stuff.iarf2.!GOT2=1
      end
  end
  else do               /* use stored options */
     avar=rep_stuff.iarf2.!AVAR ; aopts0=rep_stuff.iarf2.!AOPTS0 ;
     svarname=rep_stuff.iarf2.!SVARNAME
  end 
  astat=sre_cache_data_local(info.!series,'FIND '||is7,findid,'!ID')  /* !ID is a time stamp */
  parse var astat astat0 aval
  if astat0<>0 then do
           foo=note_error('E','Unable to find !ID: 'errs.astat0)
           return ''
   end 
   parse var astat . '0d0a'x vval
   if info.!comments_mode=1 then do
       vstatus=info.!comments_mode.!status
   end
   else do
       bstat=sre_cache_data_local(info.!series,'FIND '||is7,findid,'$STATUS') 
       parse var bstat . '0d0a'x vstatus
   end
   tt=fig_sysvars2(avar,aopts0,vval,vstatus)  /* vval is id which is a time stamp, vstatus is status var */
   return tt
end                        /* is a  meta-function */


/* else, not meta data (but possibly var-function) */

/* lookup cms dataseries variable */


astat=sre_cache_data_local(info.!series,'FIND 'is7,findid,svarname)

parse var astat astat0 astat
if astat0<>0 then do
    amess=sre_cache_data_local('FIND','ERROR',astat0,svarname)
    foo=note_error('E','CMS VAR lookup ('svarname'): '||amess)
    return ''
end 
parse var astat . '0d0a'x vval

/* is it a file to read? */
if svartype='F' then do  /* read the file */
    parse var vval origname '=' dafile ';' uptype
    if vfilename<>'' then return strip(origname)
    if vfilename_local<>'' then return strip(dafile)

    if abbrev(dafile,'\')=0 & pos(':',dafile)=0 then do     /* relative filename */
       dafile=defdir_up||dafile
    end
    vval=sre_read_file(dafile,2,,1)
    parse var vval astat . '0d0a'x vval ; astat=strip(astat)
    if astat<0 then do
            astat=abs(astat)
            errs.1 = 'no such filename'
            errs.2 = 'filename exists, but size was not produced'
            errs.3 = 'unable to open this file, even after trying'
            foo=note_error('W','Unable to read CMS File Variable ('||svarname||'='||dafile'): '||errs.astat)
            vval=''
    end                 /* error */
end             /* file type */
if svartype='M' then do  /* a long, verbatim string */
     vval=sre_pack64(vval)
end

if funcn='' then return vval

/* else, do the var-function on this value (using VVAL read from this record)  */
vval=do_var_function(funcn,vval,vopts)
return vval


/******************/
/* return svarname and svartype */
get_svarname:
if varname='!ID' then do        
    svarname="!ID"
    svartype='1'
    return ''
end 
jj1=wordpos(varname,info.!series_varnames)
if jj1=0 then do
   vverror=1
   foo=note_error('E',' No such CMS variable = '||varname)
  return ''
end 
Svarname=info.!series_varname.jj1
svartype=info.!series_vartype.jj1
return ''

/*************************/
/* possibly trim old&existing stuff from "archive" do list */
check_dolist:procedure expose gerrs.  update_verb 
parse arg alist
ith=0
do until alist=''
   if alist='' then leave
   parse var alist aline '0d0a'x alist
   ith=ith+1
   lls.ith=aline
end
if ith=0  then return alist  /* nothing to do */

newl='' ;got1=0 ; willdo=0
do mm=ith to 1 by -1
   parse var lls.mm tfile ofile . ',' atype '09'x adisp '09'x  .
   ofile=strip(ofile) ;ofile=strip(ofile,,'+')
   zz=stream(strip(ofile),'c','query exists')
   if zz='' then do             /* create non-existing archive files */
       if newl<>'' then newl=newl||'0d0a'x
       newl=newl||lls.mm
       willdo=willdo+1
       iterate
   end 
   if got1=1 then do            /* already created first "existing" archive file? */
       fpp=note_error('M','   not updating subset-archive file for '||adisp)
       iterate
   end 
   willdo=willdo+1
   got1=1
   if newl<>'' then newl=newl||'0d0a'x
   newl=newl||lls.mm
end
fpp=note_error('M','Creating 'willdo' of 'ith' subset-archive files.',1)
return newl

/*************************/
/* make a list of links to archive files */
make_arc_links:procedure expose  arclist gerrs.  update_verb 
parse arg ll,other_fields
 lst.0=0
       dsp.=''
       ith=0
       do until ll=''
          parse var ll aline '0d0a'x ll

/* message: describe the archived set (i.e.; week bounds)
   message2: from the <? ARCHIVE message="..."> option
   Note that message can be read (in archive template) using <? template desc > 
   and message2 can be read using <? show !arg > 
*/
          parse var aline tfile ofile svar sorder1 stype isarc ',' atype '09'x message '09'x message2 '09'x aurl '09'x aval
          ith=ith+1
          dsp.ith='<a href="'||aurl||'" '||other_fields||' >'||message||'</a>'
       end 
       dsp.0=ith
       i1=ith; i2=1 ; iby=-1
       if right(arclist,1)='A' then do
          i1=1 ; i2=ith ; iby=1
       end
       select
         when abbrev(arclist,'UL')=1  then do
              dsp='<ul>'||'0d0a'x
         end 
         when abbrev(arclist,'OL')=1  then do
              dsp='<ol>'||'0d0a'x
         end 
         when abbrev(arclist,'TABLE')=1  then do
              dsp='<TABLE>'||'0d0a'x
         end 
         otherwise nop
       end

       do jj=i1 to i2 by iby
          select
            when abbrev(arclist,'UL')=1 | abbrev(arclist,'OL')=1 then do
              dsp=dsp||'<li>'||dsp.jj||'0d0a'x
            end 
            when abbrev(arclist,'TABLE')=1  then do
                dsp=dsp||'<Tr><td>'||dsp.jj||'0d0a'x
            end 
            otherwise nop
          end
       end

       select
         when abbrev(arclist,'UL')=1  then do
              dsp=dsp||'</ul>'||'0d0a'x
         end 
         when abbrev(arclist,'OL')=1  then do
              dsp=dsp||'</ol>'||'0d0a'x
         end 
         when abbrev(arclist,'TABLE')=1  then do
              dsp=dsp||'</TABLE>'||'0d0a'x
         end 
         otherwise nop
       end 
       
 return dsp


/*******************/
/* determine list of archive-file instructions
   This will use the contents of recids.
   It might be necessary to reload recids, suppressing deselection info  */
load_arcfiles:procedure  expose info. gerrs. arctype arctfile arcofile recids. outfile tplfile  update_verb ,
                                 def_select def_selectval arcmessage arcurl
parse arg doartype

/* DO a "forced" load of recids
   If archive type is a time (i.e.; WEEK), use !ID, Ascending, Numeric sort
   If archive type =nnn, then use current sort options
   If archive type=SORT, then use current sort options. Note that if current sort
     is on !ID, this will have yielded an error in the LOAD_ARCFILES caller 
*/
old_sort=info.!sorted ;dsort=old_sort
old_sort_order=info.!sorted_order 
old_sort_type=info.!sorted_type ; dtype=old_sort_type
dolist=''

dorder='A'
if arctype<>'SORT' then do  /* do NOT sort on var (sort on userid, instead) */
  dsort='!ID' ; dtype='N' ; 
end

fpp=load_recids(dsort,dorder,dtype,1,1,1)         /* force reload of recids. */

/* now find "unique values" */
oldval=''
ithset0=0
jat=1
if arctype='SORT' then jat=2

do jj=1 to recids.0     
  aval=recids.jj.jat

  select
        when datatype(arctype)='NUM' then do
          aval=trunc( (jj-0.1)/arctype)
          if aval<>oldval then do
            adisp=((aval)*arctype)+1
            adisp2=min(adisp+arctype-1,recids.0)
            adisp=adisp||' to '||adisp2
          end
        end
        when arctype="SORT" then do
          aval=strip(translate(aval))
          if aval<>oldval then  adisp=dsort||': '||aval
        end
        when arctype='HOUR' then do
          aa=aval
          aval=trunc(aval*24)
          if aval<>oldval then do
             adate=trunc(aa)+726467
             tt=strip(dateconv(adate,'B','N'))
             ahr=trunc((aa-trunc(aa))*24)
             adisp=tt||' '||ahr||':00'
          end
       end
       when arctype='DAY' | arctype='DAY2' then do 
           aval=trunc(aval)
           if aval<>oldval then do
             adate=trunc(aval)+726467
             if arctype='DAY' then
                adisp=strip(dateconv(adate,'B','N'))   /* 27 aug 1999 */
             else
                adisp=strip(dateconv(adate,'B','U'))   /* 08/27/99 */
           end
        end 
        when arctype='WEEK' | arctype='WEEK2' then do
          tt=determine_week(aval)
          parse var tt aval ',' . ',' wkb ',' wke
          if aval<>oldval then do
             if arctype='WEEK2' then do
                 wke=dateconv(wke,'N','U')
                 wkb=dateconv(wkb,'N','U')
                 adisp=wkb||'-'||wke
             end
             else do
                adisp=wkb||' to '||wke
             end
          end
        end           
        when arctype='MONTH' | arctype='MONTH2' then do
          adate=trunc(aval)+726467
          tt=strip(dateconv(adate,'B','S'))   /* 19990325 */
          aval=left(tt,6,'0')
          if aval<>oldval then do
             if arctype="MONTH2" then do
                adisp=dateconv(adate,'B','U')
                parse var adisp amo '/' . '/' ayr
                adisp=amo'/'ayr
             end
             else do
                adisp=subword(strip(dateconv(adate,'B','L')),2,2)
             end
           end

        end
        when arctype='YEAR' then do
          adate=trunc(aval)+726467
          tt=strip(dateconv(adate,'B','S'))   /* 19990325 */
          aval=left(tt,4,'0')
          if aval<>oldval then do
             adisp=subword(strip(dateconv(adate,'B','N')),3,1)
           end
        end
        otherwise iterate               /* unallowed */
    end


   if aval=oldval then iterate  /* same id, so no change */
   oldval=aval

/* save specs */
    ithset0=ithset0+1
    ithset='_'||adisp
    ithset=space(translate(ithset,' $$_  ',' /\-:.'),0)
    ithset=translate(ithset,'_',' ')
    if length(ithset)>30 then ithset='_'||ithset0
    arcofile1=fix_file(arcofile,outfile,,ithset,1)   /* choose a file name */

    if dolist<>'' then dolist=dolist||'0d0a'x

    uname=filespec('n',strip(arcofile1))
    aurl=arcurl||uname

    dolist=dolist||arctfile||' +'||arcofile1||' '||dsort||' '||dorder||' '||dtype||' '||doartype||','||arctype||'09'x|| ,
                        adisp||'09'x||arcmessage||'09'x||aurl||'09'x||aval 
end 

/* reload to current sort options (remove effects of this archiving action) */
fpp=load_recids(old_sort,old_sort_order,old_sort_type,1,,1)         /* force reload of recids. */


return dolist  
   

/************/
/* figure out a filename, possibly randomized, possibly relative to a directory of a 
reference file*/
fix_file:procedure
parse arg afile,reffile,noast,ith,doover

afile=strip(afile)
if abbrev(afile,'\')=0 & pos(':',afile)=0 then do     /* relative filename */
  afile=strip(filespec('d',reffile)||filespec('p',reffile),,'\')||'\'||afile
end
if pos('*',afile)+pos('?',afile)=0 then return afile  /* no randomization component */

if noast=1 then return afile

if doover=1  then do                    /* use ITH to determine file name (don't search for unused name)*/
  if pos('*',afile)=0 then return afile||'_'||ith
   parse var afile p1 '*' p2 ; afile=p1||ith||p2
  return afile
end

if pos('*',afile)>0 then do
     parse var afile p1 '*' p2 ; afile=p1||'????'||p2
end 
fpath=filespec('d',afile)||filespec('p',afile)
fname=filespec('n',afile)
aname=sre_tempfile(fname,'?',fpath)

return aname

/************/
/* determine a "system variable" */
fig_sysvar:procedure expose gerrs. info. recids.  update_verb 
parse arg avar,aopts0,useval
select
   when avar='$RETAIN' then do
      vv='!'||avar
      return info.vv
   end 

   when avar='$DATE' then do
        aopts=aopts0
        if aopts='' then aopts='N'
        aopts=strip(translate(left(strip(aopts),3,' ')))
        if wordpos(aopts,'B D E L M N O S U W Y J WN WB WB2 ')=0 then do
           fpp=note_error('W','Bad format in $DATE: 'aopts0)
           return ''
        end 
        select
           when aopts='WB' | aopts='WN' | aopts='WB2' then do
              foo=determine_week(useval) 
              parse var foo wn','ndate','wkb','wke
              if aopts='WN' then return wn
              if aopts='WB2' then do
                 wke=dateconv(wke,'N','U')
                 wkb=dateconv(wkb,'N','U')
                 return wkb||'-'||wke
              end 
              else do
                return wkb||' to 'wke
              end
           end 
           when aopts='J' then do
              if useval='' then
                 dd=time('b')-726467
             else
                 dd=trunc(useval)
              return dd
           end 
           otherwise do
               if useval='' then
                 useval=date('B')
               else
                 useval=trunc(useval+726467)
               if aopts='B' then return useval
               aopts1=aopts
               if aopts1='Y' then aopts1='N'
               dtmp=dateconv(useval,'B',aopts1)
               if aopts='Y' then parse var dtmp . . dtmp
               return dtmp
           end
        end
   end          /* Date */

   when avar='$TIME' then do 
        aopts=aopts0
        if aopts='' then aopts='N'
        aopts=strip(translate(left(strip(aopts),1)))
        if pos(aopts,'LHMSNCF')=0 then do
            foo=note_error('W','Bad $TIME format: '||aopts0)
            return ''
        end 
        zz=digits()
        if useval<>'' then do
                parse var useval . '.' dtmp
                ttmp='0.'||dtmp
        end 
        else do
             tt=time('S')
             numeric digits 16
             tmp1=tt/86400
             parse var tmp1 . '.' tmp2
             tmp2=strip(left(strip(tmp2),8))
             ttmp='0.'||tmp2
             numeric digits zz
        end
        if aopts='F' then do
           numeric digits zz
           return ttmp
        end
        secs=trunc(ttmp*86400)
        tmp0=secs/3600
        parse var tmp0 hour '.' tmp1 ; if tmp1='' then tmp1=0
        hour=right(hour,2,'0')
         tmp1='0.'||tmp1
         tmp2=tmp1*60
         parse var tmp2 minute '.'tmp3 ; ; if tmp3='' then tmp3=0
         minute=right(minute,2,'0')
         tmp3='0.'||tmp3
         tmp3=tmp3*60 ; parse var tmp3 second '.' .
         second=right(second,2,'0')
         numeric digits zz
         select
            when aopts='L'  then return hour||':'||minute||':'||second||'.00'
            when aopts='M' then return trunc(secs/60)
            when aopts='H' then return strip(hour,'l','0')
            when aopts='S' then return secs
            when aopts='C' then do
                ampm='am'
                if hour>11 then ampm='pm'
                if hour>12 then hour=hour-12 ; hour=right(hour,2,' ')
                return hour||':'||minute||':'||second||ampm
            end /* do */
            otherwise return hour||':'||minute||':'||second
         end  /* select */
   end                          /* $TIME */
  
   otherwise do
        fpp=note_error('W','Unknown system-variable: 'avar)
        return ''
   end
end  /* select */

return ''  /* should never get here */


/************/
/* determine a "system variable" */
fig_sysvars2:procedure expose gerrs. info.  recids.  update_verb 
parse arg avar,aopts0,atime0,astatus       

/* atime0 is sent as the user-id, which is a timestamp */

/* parse arg avar0,atime0,astatus        */
/* parse var avar0 avar '(' aopts0 ')'  */

aopts0=translate(strip(aopts0))
avar=strip(translate(avar))

parse var atime0 adate1 '.' atime
adate=adate1+726467

select
   when avar='STATUS' then do  /* atime will be in nnnn.ffffffff format */
        parse var aopts0 svar ',' dfmt ','tfmt
        svar=strip(svar)
        useit=''
        do until astatus=''
           parse var astatus a1 astatus
           parse var a1 a1a '=' a1b
           if strip(a1a)<>svar then iterate
           useit=a1b
           leave
        end 
        if svar='CMT_CNT' then return useit

/* break the nnn.ffff date into components */
        if dfmt='' then dfmt='N'; if tfmt='' then tfmt='N'

        if dfmt=0 | useit=0 then
          oo1=''
        else
           oo1=fig_sysvar('$DATE',dfmt,useit) 
        if tfmt=0 | useit=0 then
           oo2=''
        else
           oo2=fig_sysvar('$TIME',tfmt,useit)  
       return oo1||' '||oo2
   end

   when avar='DATE' then do  /* atime will be in nnnn.ffffffff format */
        aopts=aopts0
        if aopts='' then aopts='N'
        aopts=strip(translate(left(strip(aopts),3,' ')))
        if wordpos(aopts,'B D E L Y M N O S U W J WN WB WB2')=0 then do
           fpp=note_error('W','VAR. Bad format in $TIME: 'aopts0)
           return ''
        end 
        select
           when aopts='WB' | aopts='WN' | aopts='WB2' then do
              foo=determine_week(atime0)
              parse var foo wn','ndate','wkb','wke
              if aopts='WN' then return wn
              if aopts='WB2' then do
                 wke=dateconv(wke,'N','U')
                 wkb=dateconv(wkb,'N','U')
                 return wkb||'-'||wke
              end
              else do
                  return wkb||' to 'wke
              end
           end 
           when aopts='J' then do
              return adate1
           end 
           when aopts='Y' then do
               tt=dateconv(adate,'B','N')
               parse var tt . . adate1
              return strip(adate1)
           end 

           otherwise do
               tt=dateconv(adate,'B',aopts)
               return tt
           end
        end
   end          /* Date */

   when avar='TIME' then do 
        aopts=aopts0
        if aopts='' then aopts='N'

        if pos(aopts,'LHMSNCF')=0 then do
            foo=note_error('W','Bad (VAR) $TIME format: '||aopts0)
            return ''
        end 

        if aopts='F' then return '0.'||atime

/* else, convert to hr min second */
         zz=digits()
         numeric digits 16
         atime='0.'||atime
         secs=trunc(atime*86400)
         tmp0=secs/3600
         parse var tmp0 hour '.' tmp1 ; if tmp1='' then tmp1=0
         hour=right(hour,2,'0')
         tmp1='0.'||tmp1
         tmp2=tmp1*60
         parse var tmp2 minute '.'tmp3 ; ; if tmp3='' then tmp3=0
         minute=right(minute,2,'0')
         tmp3='0.'||tmp3
         tmp3=tmp3*60 ; parse var tmp3 second '.' .
         second=right(second,2,'0')
         numeric digits zz
         select
            when aopts='L'  then return hour||':'||minute||':'||second||'.00'
            when aopts='M' then return trunc(secs/60)
            when aopts='H' then return strip(hour,'l','0')
            when aopts='S' then return secs
            when aopts='C' then do
                ampm='am'
                if hour>11 then ampm='pm'
                if hour>12 then hour=hour-12 ; hour=right(hour,2,'0')
                return hour||':'||minute||':'||second||ampm
            end /* do */
            otherwise return hour||':'||minute||':'||second
         end  /* select */
   end                          /* $TIME */

   otherwise do
        fpp=note_error('W','Unknown (VAR) system-variable: 'avar)
        return ''
   end
end  /* select */

return ''  /* should never get here */


/************/
/* Repetitively replace  <? cms ...> elements (in a REPEAT .. /REPEAT block) */
PROCESS_CMS_REPEAT:procedure expose stuff. rep_stuff. info. recids. cshdata. defdir_up defdir CINFO. gerrs.  ,
                                           def_sortvar def_sortorder def_sorttype def_select def_selectval ,
                                          tplfile dolist_1  dolist_c isarchive outfile root thisaddon daindx optname. ,
                                         cmsdir_files cmsdir_uploads cmsdir_lib cmsdir cmsdir_pol  update_verb 
parse arg repstat
special_vars=' !DEFFILE !ID !LOOP_S !LOOP_E !NUMREC !CURRENT !FILE_VARS !UPLOADDIR !FILEDIR  '

goo=''
findid=''
asked_id7=''   /* set when ID=nnn is specified */
info.!!count_block=0
if_results=''

parse var repstat . opt1

call find_i1_i2


if result<>1 then return result

/*  ---- for each record, go through the repeat instructions  --- */

if datatype(i1)<>'NUM' | datatype(I2)<>'NUM' then do
   fpp=note_error('E','Bad REPEAT range: I1, I2 = 'I1','I2)
   i1=2; i1=1
end
else do
  i1=max(i1,1)
  i2=min(recids.0,i2)
end


call find_gotos         /* add goto information (to REP_STUFF.) for IF and GOTO keywords */
ares=result             /* ares= # of errors */
if ares>0 then return '' 


/* CALL and IF might use the records's variables */
if  asked_id7<>0 & info.!ADD_LOCALS=1 then do        /* skip if ID=NONE specified */
  CINFO.0=CINFO.0||' '||special_vars||' '||info.!series_varnames
  CINFO.!!LOOP_S=i1
  CINFO.!!LOOP_E=i2
  CINFO.!!NUMREC=recids.0
end
else do
   CINFO.0=CINFO.0||' !ID !CURRENT'
end 

/* start processing the selected range of records */
lcontrol=0
willmake=0 ; willnotmake=0

do iarf=i1 to i2
   thisgoo=''
   if asked_id7<>'' then   /* this is set when a specific ID is requested */
      aid7=asked_id7
   else
      aid7=recids.iarf

   info.!!count=info.!!count+1
   info.!!count_block=info.!!count_block+1

   if info.!ADD_LOCALS=1 & aid7<>0 then do   /* load all record components into cinfo. array */
         tt=more_cinfo(aid7,iarf,defdir)     /* AID7=0 if ID=NONE specified */
   end
   else do              /* set !!CURRENT and !!ID (don't bother with the cms variables */
     CINFO.!!CURRENT=iarf
     CINFO.!!ID=recids.iarf.1
     cinfo.!!COUNT=info.!!count
     cinfo.!!COUNT_BLOCK=info.!!COUNT_BLOCK
     if asked_id7=0 then CINFO.!!ID=0
   end 
   iarf2=0 ;   idrop=0 ; icontrol=0
   popif=0   ; stripstart=0
   
/* Start processing the instructions in the REPEAT .. /REPEAT */

   do until iarf2 >= rep_stuff.0           /* iarf2 can also be set by IF and GOTO */
     if iarf2='' then do
        foo=note_error('E',' Bad jump to address ')
        iterate
     end 
      iarf2=iarf2+1

      if popif=1 then do                /* IF condition ended, so POP this "if's" result */
         parse var if_results wasif if_results
         popif=0
      end 

      keyword=rep_stuff.iarf2.1  ; options=rep_stuff.iarf2.2

      gerrs.!CURRENT='Record '||iarf||'. '||keyword||' '||options

      if rep_stuff.iarf2.!cms=0 then do
         thisgoo=thisgoo||rep_stuff.iarf2
         iterate                /* get next instruction */
      end                       /* otherwise, it's a cms */

/* ---  Now process CMS keywords  --- */

      select                    /* do the CMS keyword */

        when abbrev(keyword,'REM')=1 then iterate

        when keyword='!--' then do
           thisgoo=thisgoo||'<!-- '
        end
        when keyword='--' then do
           thisgoo=thisgoo||' -->'
        end

        when keyword="<" then thisgoo=thisgoo||'<'

        when keyword="/<" then thisgoo=thisgoo||'>'

        when abbrev(keyword,"REPEAT")=1 then do
          foo=note_error('E','You can not put a REPEAT inside of a REPEAT .. /REPEAT block ')
        end


        WHEN keyword="/ARCHIVE"  | keyword='/LINK' | keyword='/COMMENT'| keyword='/POLL' then do
           thisgoo=thisgoo||'</a>'
        end

        when abbrev(keyword,"ARCHIVE")=1 then do
           if isarchive='12' | isarchive='13' then do
              foo=note_Error('E','ARCHIVE: archive not permitted when creating comment files ')
              return ''
           end

           auid=translate(cinfo.!!ID,'__',' .')
           call parse_options(options)  

           arctfile=strip(translate(ropts.!FILE))
           arctfile=fix_file(arctfile,tplfile)
           if arctfile='' then do
                foo=note_Error('E','ARCHIVE: the template file not specified: 'arctfile)
                iterate
           end
           if stream(arctfile,'c','query exists')='' then do
              foo=note_Error('E','ARCHIVE: template file does not exist: 'arctfile)
              iterate
           end

           arcofile=strip(translate(ropts.!OUTPUT))
           if arcofile='' then do
              iip=lastpos('.',outfile)
              if iip=0  then iip=length(outfile)+1
              p1=left(outfile,iip-1)
              p2=substr(outfile,iip+1)
              arcofile=p1||'$A*.'||p2
           end

           if abbrev(arcofile,'$')=1 then do    /* use a variable name */
              voof=substr(strip(translate(arcofile)),2)
              if wordpos(voof,'!ID '||info.!series_varnames)>0 then do
                  vurl=process_cms('VAR ',,aid7,voof)  /* cms dataseries variable */
              end
              else do
                  vurl=process_cms('SHOW ',voof)  /* user-defined variable */
              end 
              arcofile=strip(vurl)
           end
           else do
            if pos('*',arcofile)=0 then do
              foo=note_Error('E','ARCHIVE; output file does not contain * : 'arcofile)
              iterate
            end
            parse var arcofile p1 '*' p2
            arcofile=strip(p1||auid||p2)
          end                   /* end of basic outfile determination */

          outdir=stuff.!FILEDIR                 /* add path info */
          if outdir='' then outdir=cmsdir_files
          outdir=strip(strip(outdir),,'\')

           if pos(':',arcofile)=0 &  abbrev(arcofile,'\')=0 then do
              arcofile=outdir||'\'||arcofile
           end 

/*        If UPDATE=AUTO, then use value of UPDATEARCHIVE option  (as specified by client):
                1 means "do NOT overwrite",
                0 or '' means "do overwrite (do NOT update)  */
           isupdate=strip(translate(ropts.!UPDATE))
           if isupdate='AUTO' | isupdate='' then isupdate=optname.!UPDATEARCHIVE
           if isupdate<>1 then isupdate=0

           if update_verb=1 then isupdate=1   /* always update */
           if update_verb=3 then isupdate=0  /* always create */

           isupdate2=1+isupdate

           t1=strip(filespec('d',arcofile)||filespec('p',arcofile),,'\')
           if right(t1,1)=':' then t1=t1||'\'
           if dosisdir(t1)=0 then do
               foo=note_Error('E','ARCHIVE: ii) no such directory for output files: 't1)
               iterate
           end

           arcurl=strip(translate(ropts.!URL))
           arcurl=strip(arcurl,'t','/')||'/'

           arcmessage=strip(ropts.!MESSAGE)

          arcofile1=fix_file(arcofile,outfile,,auid,1)   /* choose a file name */
        
          nomake=0
          if isupdate=1 then do         /* update mode, check for preexisting */
             if stream(arcofile1,'c','query exists')<>'' then do
                fpp=note_error('M','    not overwriting singleton-archive file for 'auid)
                uname=filespec('n',strip(arcofile1))
                aurl=arcurl||uname
                nomake=1
                willnotmake=willnotmake+1
             end 
          end

          if nomake<>1 then do
             uname=filespec('n',strip(arcofile1))
             aurl=arcurl||uname

             adolist=arctfile||' +'||arcofile1||' 1 1 1 '||isupdate2||',1'||'09'x|| ,
                           auid||'09'x||arcmessage||'09'x||aurl||'09'x||aid7

/* now we are ready to make a list of archive file creation instructions */
             dolist_1=adolist||'0d0a'x||dolist_1  /* add to list of template files to process */
             willmake=willmake+1
           end

           showlist='<a href="'||aurl||'" '||ropts.!2||' >'
           thisgoo=thisgoo||showlist

        end  /* ARCHIVE */

        when abbrev(keyword,"COMMENT")=1 then do      
           tt=sre_cache_data_local(stuff.!series,'FIND ID7 ',aid7,'$STATUS ')
           parse var tt . '0d0a'x tt
           tt2=''
           if wordpos('$ABSTRACT',info.!series_varnames)>0 then do 
                tt2=sre_cache_data_local(stuff.!series,'FIND ID7 ',aid7,'$ABSTRACT')
                parse var tt2 . '0d0a'x tt2
           end 
           tgoo=mk_comment1(iarf2,cinfo.!!ID,tt';'||translate(tt2,' ','0d0a0901'x))
           thisgoo=thisgoo||tgoo
        end

        when abbrev(keyword,"/DEFINE")=1 then do
          foo=note_error('E','You can not put a /DEFINE inside of a REPEAT .. /REPEAT block ')
        end

        when abbrev(keyword,"DEFINE")=1 then do
          foo=note_error('E','You can not put a DEFINE inside of a REPEAT .. /REPEAT block ')
        end

        when abbrev(keyword,"/ARCHIVE")=1 then do
          foo=note_error('E','You can not put a ARCHIVE inside of a REPEAT .. /REPEAT block ')
        end

        when abbrev(keyword,'ITER')=1 then do
              icontrol=1                                /* stop processing for this record */
              if wordpos('DROP',translate(options))>0 then idrop=1 /* do not save currently processed stuff for this record */
        end

        when abbrev(keyword,'DROP')=1 then do
              icontrol=1        /* stop processing for this record */
              idrop=1           /* do not save currently processed stuff for this record */
        end

        when abbrev(keyword,'LEAVE')=1  then do      
            icontrol=2                               /* stop processing all records */
            if wordpos('DROP',translate(options))>0 then idrop=2 /* do not save currently processed stuff for this record */
        end

        when abbrev(keyword,"VAR")=1 then do
            varname=''  ; useid7=aid7
            opt1=strip(translate(options))
            call do_var_keyword 0
            ares=result
            if assignto<>'' then do
                 foo=process_cms('SET',assignto,ares)
            end 
            if nodisplay=1 then  ares=''        /* display nothing */
            thisgoo=thisgoo||ares
        end

        when abbrev(keyword,"POLL")=1 then do
            varname=''  ; useid7=aid7
            opt1=strip(translate(options))
            call do_poll_keyword 0
            ares=result
            do kk1=1 to words(poll_info.0)
               cw=strip(translate(word(poll_info.0,kk1)))
               vcw='!'||cw
               aval=poll_info.vcw
               foo=process_cms('SET',cw,aval)
            end 
            thisgoo=thisgoo||ares
        end


        when abbrev(keyword,"TEM")=1 | abbrev(keyword,'TPL')=1 then do
          thisgoo=thisgoo||process_cms('TEMPLATE ',options,aid7)
        end

        when abbrev(keyword,"SERI")=1 then do
          thisgoo=thisgoo||process_cms('SERIES ',options,aid7)
        end

        when abbrev(keyword,"SET")=1 then do
          thisgoo=thisgoo||process_cms('SET ',options)
        end

        when abbrev(keyword,"SHOW")=1 then do
          thisgoo=thisgoo||process_cms('SHOW ',options)
        end

        when abbrev(keyword,"SHRINK")=1 then do
          stripstart=length(thisgoo)
          iterate
        end
 
        when abbrev(keyword,"/SHRINK")=1 then do
           goot=translate(substr(thisgoo,stripstart+1),' ','0d0a0920'x)
           goot=space(goot)
           if stripstart=0 then
             thisgoo=goot
           else
             thisgoo=left(thisgoo,stripstart)||goot
           stripstart=0
           iterate
        end

        when abbrev(keyword,"INC")=1 then do
          thisgoo=thisgoo||process_cms('INC ',options)
        end

        WHEN keyword="/IF" then do  /* note: skip PAST here when IF not satisfied */
          popif=1              /* signal to "pop an if value" */
          iterate
        end 

        WHEN keyword="/ELSE" then do  /* note: skip PAST here when IF not satisfied */
          iterate
        end 

        when keyword='CASES' then do

          if rep_Stuff.iarf2.!VARNAME='' then do
            if pos('=',opt1)=0 then do    /* just the variable name in option list*/
                 parse var options varname .
            end 
            else do                               /* possibly other stuff in option list */
                call parse_options(options)
                varname=ropts.!NAME
            end     
            vv='!'||strip(translate(varname))
            rep_stuff.iarf2.!VARNAME=vv
          end
          else do
             vv=rep_stuff.iarf2.!VARNAME
          end
          rep_stuff.iarf2.!VALUE=strip((cinfo.vv))       /* use cinfo (since this is an "if" like statement */
          rep_stuff.iarf2.!TRUE=0               /* signal which iarf2 had a true case */
          iterate
        end

        when keyword='CASE' then do
             if rep_stuff.iarf2.!CONDS='' then do
                 nconds=read_conds(options,iarf2)       /* set conditions in rep_stuff.iarf2, returns # of conditions */
                 rep_stuff.iarf2.!CONDS=nconds
             end
             else do
                 nconds=rep_stuff.iarf2.!CONDS
             end 
             if  nconds=0 then do         /*  the default */
                  rep_stuff.bindex.!TRUE=iarf2           /* signals "case resolved successfully */
                  iterate
             end

/* avalue can be multiple conditions -- check them */
             bindex=rep_stuff.iarf2.!BEGIN
             rvalue=rep_stuff.bindex.!VALUE    /* value of chosen variable, for this record */
             istrue=check_case(rvalue,nconds,iarf2)
             if istrue=-1 then do
                icontrol=3
             end 
             else do
               if istrue=1 then do         /* this is a true case */
                   rep_stuff.bindex.!TRUE=iarf2           /* signals "case resolved successfully */
                   iterate
               end
               iarf2=rep_stuff.iarf2.!GOTO        /* jump to next CASE */
              iterate
            end
        end

        when keyword='/CASE' then do
             bindex=rep_stuff.iarf2.!BEGIN
             if rep_stuff.bindex.!TRUE>0 then do         /* case resolved? */
                 iarf2=rep_stuff.bindex.!ENDCASE
             end
             iterate
        end

        when keyword='/CASES' then do
             iterate
        end

        
        WHEN keyword="LABEL" then do   /* Note: skip PAST here on a GOTO  */
           iterate
        end 

        when abbrev(keyword,"SORT")=1 then do
          foo=note_error('E','You can not specify SORT within a REPEAT block ')
          iterate
        end

        WHEN keyword="GOTO" then do  /* typically, GOTO is surrounded by IF and /IF */
           iarf2=rep_stuff.iarf2.!GOTO  /* GOTO a label, as set in find_gotos */
           iterate
        end 


/* A stuff */
        when keyword='A' then do
            if rep_stuff.iarf2.!A=0 then do   /* no ?xx modifiers, so nothing fancy to do  */
                 thisgoo=thisgoo||'<A '||options||'>'
            end
            if rep_stuff.iarf2.!A='' then do  /* see if ?URL or ?LINK, and parse it  */
               rep_stuff.iarf2.!A=1
               call parse_options(options)
               rep_stuff.iarf2.!options=ropts.!0
               rep_stuff.iarf2.!options_noamp=ropts.!3
               rep_stuff.iarf2.!URL=strip(ropts.!?URL)
               rep_stuff.iarf2.!BASE=strip(ropts.!?BASE)
               rep_stuff.iarf2.!NAME=strip(ropts.!?NAME)
               rep_stuff.iarf2.!LNAME=strip(ropts.!?LNAME)
               rep_stuff.iarf2.!URL_FILE=strip(ropts.!?URL_FILE)
               gotamps=pos('&',ropts.0)
               iamps=0
               if  gotamps>0 then do    /* save all &optname=varname options */
                   do jj=1 to words(ropts.0)
                       aw0=strip(word(ropts.0,jj)) ;  aw=translate(aw0) 
                       if abbrev(aw,'&')=0 then iterate
                       vaw='!'||aw
                       iamps=iamps+1
                       rep_stuff.iarf2.!!AMPS.iamps=substr(aw0,2)
                       rep_stuff.iarf2.!!AMPS.iamps.1=ropts.vaw
                   end
                   if iamps>0 then rep_stuff.iarf2.!!AMPS.0=iamps
               end               /* !!AMPS build */
               if  rep_stuff.iarf2.!URL_FILE='' & rep_stuff.iarf2.!URL='' & rep_stuff.iarf2.!NAME='' ,
                  & rep_stuff.iarf2.!LNAME=''  & rep_stuff.iarf2.!BASE=''  & iamps=0 then 
                  rep_stuff.iarf2.!A=0
            end

            if rep_stuff.iarf2.!A=1 then do  /* must check a ?URL or ?BASE or a ?NAME */
              aopts=''                         
              if rep_stuff.iarf2.!!AMPS.0<>"" then do  /* make an option list? */
                do mm=1 to rep_stuff.iarf2.!!AMPS.0
                   aopt1=rep_stuff.iarf2.!!AMPS.mm
                   voof=rep_stuff.iarf2.!!AMPS.mm.1
                   if words(voof)>1  then do
                      optval=voof         /* use as is */
                   end 
                   else do
                     voof=strip(translate(voof))
                     if wordpos(voof,'!ID '||info.!series_varnames)>0 then do
                        optval=process_cms('VAR ',,aid7,voof)  /* process ?NAME */
                     end
                     else do
                        optval=process_cms('SHOW ',voof)  /* process ?LNAME */
                     end 
                   end
                   if aopts='' then
                      aopts='?'||aopt1||"="||sre_packur_make(strip(optval),2,' &+~"')
                   else
                      aopts=aopts||'&'||aopt1||"="||sre_packur_make(strip(optval),2,' &+~"')
                end
              end               /* if !!AMPS.0<>'' */

              vurl=''

              select                    /* now check all the other options */

                when rep_stuff.iarf2.!URL<>'' then do   /* note that aopts is ready to be added */
                    voof=strip(translate(rep_stuff.iarf2.!URL))
                    if wordpos(voof,'!ID '||info.!series_varnames)>0 then do
                       vurl=process_cms('VAR ',,aid7,voof)  /* cms dataseries variable */
                    end
                    else do
                        vurl=process_cms('SHOW ',voof)  /* user-defined variable */
                    end 

                    if abbrev(vurl,'<!--')=1 then do
                        vurl=vurl
                    end 
                    else do                  /* not an error comment */
                      abase=''
                      if rep_stuff.iarf2.!URL='!ID' then vurl=translate(vurl,'_','.')
                       if abbrev(translate(vurl),'HTTP://')=0 & rep_stuff.iarf2.!BASE<>'' then do
                            if pos('/',vurl)=0 then do
                              abase=rep_stuff.iarf2.!BASE
                            end
                            else do
                               parse var vurl gooba '/' .
                               gooba=translate(gooba,' ','.')
                               if words(gooba)<3 then abase=rep_stuff.iarf2.!BASE
                            end
                       end 
                       if vurl<>'' then vurl=' href="'||space(abase||vurl||aopts,0)||'"'
                    end
                end            /* when !URL */

/* cinfo.0 : all vars, info.!series_varnames=dataseries vars */

                when rep_stuff.iarf2.!NAME<>'' | rep_stuff.iarf2.!LNAME<>"" then do
                    voof=rep_stuff.iarf2.!name
                    if voof='' then voof=rep_stuff.iarf2.!lname
                    voof=strip(translate(voof))
                    if wordpos(voof,'!ID '||info.!series_varnames)>0 then do
                       vurl=process_cms('VAR ',,aid7,voof)  /* cms dataseries variable */
                    end
                    else do
                        vurl=process_cms('SHOW ',voof)  /* user-defined variable */
                    end 
                     vurl=translate(vurl,'__','. ')
                    if vurl<>'' then vurl=' name="'||rep_stuff.iarf2.!BASE||space(vurl,0)||'"'
                end


                when rep_stuff.iarf2.!URL_FILE<>'' then do
                    vurl=process_cms('VAR ','FILENAME_LOCAL='||rep_Stuff.iarf2.!URL_FILE,aid7)  /* process ?FILENAME */
                    if rep_stuff.iarf2.!NAME='!ID' then vurl=translate(vurl,'_','.')
                    if vurl<>'' then do
                       abase=rep_stuff.iarf2.!BASE
                       if abase='' then
                           vurl=' href="'||space(vurl,0)||'"'
                       else
                           vurl=' href="'||abase||space(vurl,0)||'"'
                    end
                end

                otherwise do
                   nop
                end
              end                       /* select */
              thisgoo=thisgoo||'<a '||vurl||' '||rep_stuff.iarf2.!OPTIONS_noamp||'>'        
           end                      /* rep_stuff.iarf2.!A=1  */ 

           else do              /* sometimes happens (first call has nothing dynamic) */
               thisgoo=thisgoo||'<A '||options||'>'
           end


        end


/* READFILE stuff */

   WHEN keyword="READFILE" then do   /* specify sort parameters */
        val1=''

        if options='' then do
             fpp=note_error('W','You did not specify a varname in a READFILE ')
             return '' 
        end
        if words(options)=1 then do        /* simple form */
             fvarname=translate(strip(options))
             if pos('.',fvarname)>0 then val1=fvarname
             vsave=''; nodisplay=0
             usefdir=strip(stuff.!filedir,,'\')||'\'
        end 
        else do
           call parse_options(options)  /* name=, order= */
           fvarname=translate(strip(ropts.!NAME))
           if fvarname='' then do
             fpp=note_error('W','You did not specify a NAME= in a READFILE')
             iterate 
           end
           if pos('.',fvarname)>0 then val1=fvarname
           fdir=strip(ropts.!DIR)
           if pos(':',fdir)>0 | abbrev(fdir,'\')=1 then do  /* fully qualified */
               usefdir=fdir
           end 
           else do
                usefdir=strip(stuff.!filedir,,'\')||'\'||fdir
           end 
           usefdir=strip(usefdir,,'\')||'\'
           nodisplay=strip(ropts.!NODISPLAY)
           vsave=translate(strip(ropts.!SAVE))
           if vsave='' then
              vsave=translate(strip(ropts.!ASSIGN))

        end

/* get varname's value */
        if val1='' then do
          if wordpos(fvarname,'!ID '||info.!series_varnames)>0 then do
              val1=process_cms('VAR ',,aid7,fvarname)  /* cms dataseries variable */
          end
          else do
            if wordpos(fvarname,cinfo.0)=0 then do
               fpp=note_error('E','No such variable 'fvarname' in a READFILE element ')
               iterate 
            end
            val1=process_cms('SHOW ',fvarname)  /* user-defined variable */
          end 
          if val1='' then do
             fpp=note_error('W','No value for 'fvarname' in a READFILE element ')
             iterate 
          end
        end             /* otherwise, explicit filename in ?readfile */
/* read the file */
        if pos(':',val1)>0 | abbrev(val1,'\')=1 then do  /* fully qualified */
           ffile=val1
        end
        else do
           ffile=usefdir||strip(val1)
        end 

        fstuff=sre_read_file(ffile,2,2,1)
        if vsave<>'' then do
          vv='!'||vsave
          CINFO.vv=fstuff
          if wordpos(vsave,CINFO.0)=0 then   CINFO.0=CINFO.0||' '||vsave
        end
        if nodisplay<>1  then  thisgoo=thisgoo||fstuff
   end



/* LINK stuff */
        when keyword='LINK' then do
           togoo=do_link(options,cinfo.!!ID,iarf2,aid7)
           thisgoo=thisgoo||togoo
        end


/* CALL stuff -- more complicated then when not inside of a REPEAT (tokenize it, since
   it will be called repeatedly!)*/

        WHEN abbrev(keyword,'CAL')=1 then do   /* used for CALL and INTERPRET  */
           if rep_stuff.iarf2.!GOT='' then do  /* tokenize the code */
              k1=info.!nprocs+1 
              info.!nprocs=k1
              userargs=''
              parse var options ctype options
              if ctype='$FILE' then do
                 parse var options dafile userargs ; dafile=strip(dafile)
                 options=read_interp_file(dafile,info.!cmsdir)
              end
              astr=make_interpret_string(options,,0)    /* make the rexxrun'able string */
              ok1=tokenizestring(astr,'t2')
              if ok1=0 then do
                 foo=note_error('E','Unable to compile CALLed procedure ')
                 t2=''
                 rep_stuff.iarf2.!got=0        /* 0 means "busted" */
              end 
              else do
                 rep_stuff.iarf2.!got=k1        /* CALL "ID" */
              end 
              rep_stuff.iarf2.!TOK=t2
              rep_stuff.iarf2.!USERARGS=userargs
          end
          if rep_stuff.iarf2.!GOT>0 then do     /* interpretable file is useable */
              foo=make_interpret_string(options,info.!series_varnames,1)  /* make usevars and useval */
              t2=rep_stuff.iarf2.!TOK
              userargs=rep_stuff.iarf2.!USERARGS
              astat=rexxrun('T',t2,'daresult',usevars,useval,userargs,info.!series_varnames,rep_stuff.iarf2.!got)
              retstuff=''
              if astat=0 then do                /* success */
                parse var daresult rstat ',' rhd '0d0a'x rval
                if rstat=0 then do                   /* error in CALLed code */
                    parse var rhd rcode rline ; rline=rline-29   /* 29 lines of prep material */
                    foo=note_error('E','Error 'rcode' in repeated CALLed procedure at line 'rline)
                end 
                else do
                   parse var daresult rhd '0d0a'x rval
                   parse var rhd . nres rhd
                   if nres<>0 then retstuff=left(rval,nres)
                   iat=nres+1
                   do until rhd=''
                      if rhd='' then iterate
                      parse var rhd rrname rrbytes rhd ; rrname=strip(rrname) ; rrbytes=strip(rrbytes)
                      if rrbytes=0 then do
                          CINFO.rrname=''
                      end
                      else do
                          CINFO.rrname=substr(rval,iat,rrbytes)
                          iat=iat+rrbytes
                          gg=substr(rrname,2)
                          if wordpos(gg,cinfo.0)=0 then cinfo.0=cinfo.0||' '||gg
                      end      /* rrbytes=0 */

                   end       /* rhd ='' */
                end        /* rstat= 0 */
                thisgoo=thisgoo||retstuff
              end       /* astat=0 */
              else do
                 foo=note_error('E','Error ('||astat||' CALLing procedure ')
              end
          end           /* else, bad tokenization, so do nothing */
          else do
                 foo=note_error('M','Unable to call uncompiled procedure ')
          end           

        end             /* CALL */


/* IF must be within REPEAT */
        WHEN keyword="IF" then do            
           parse var options a1 cond a2  
           if pos('=',a1)>0 then do               /* a common goof */
              parse var options a1 '=' a2 ; cond='eq'
           end 
           if cond='=' | cond='==' then cond='eq'
           val1=if_value(a1)
           if a2='' then do      /* an "EVENT" check */
              val2=rep_stuff.iarf2.!VAL1  /* use prior value (on first obs, use default of '' */
              rep_stuff.iarf2.!VAL1=val1
           end
           else do
              val2=if_value(a2)
           end
           if_failed=check_if_failed(val1,cond,val2)   /* if_failed means "IF returned FALSE" */
           if if_failed=1 then iarf2=rep_stuff.iarf2.!GOTO
           if_results=if_failed||' '||if_results               /* can be used by ELSE */
           iterate
        end

/* ELSE must be within REPEAT, and MUST be followed by an IF */
        WHEN keyword="ELSE" then do            
           if wasif='' then do
               foo=note_error('E','ELSE without a preceding IF ')
               t2=''
           end
           if wasif=0 then do
               iarf2=rep_stuff.iarf2.!GOTO  /* 0 means "if did NOT fail (hence do NOT do this ELSE */
           end
           iterate

        end


/* misc html elements, with dynamic arguments */
        when keyword='HTML' then do
           parse var options ealname options
           call parse_options(options)
           thisgoo=thisgoo||'<'||strip(ealname)||' '
           do jj=1 to words(ropts.0)
               qaw=strip(translate(word(ropts.0,jj)))
               vqaw='!'||qaw
               getv=0
               if abbrev(qaw,'?')=1 then do
                  getv=1
                  qaw=substr(qaw,2)
               end
               atval=ropts.vqaw
               if getv==1 then do
                   atval=translate(strip(atval))
                   if wordpos(atval,'!ID '||info.!series_varnames)>0 then do
                       atval=process_cms('VAR ',,aid7,atval)  /* process ?NAME */
                    end
                    else do
                        atval=process_cms('SHOW ',atval)  /* process ?LNAME */
                    end 
                end
                thisgoo=thisgoo||' '||qaw||'="'||atval||'" '
          end
          thisgoo=thisgoo||' >'
        end 


/* Table stuff */
        when keyword='TR' | keyword="TH" | keyword="TD" then do
            if rep_stuff.iarf2.!COLORS.!RD='' then do  /* see if ?colors, and parse it  */
              rep_stuff.iarf2.!COLORS.!RD=0
              call parse_options(options)
              rep_stuff.iarf2.!options=ropts.!0
              colorlist=ropts.!?COLORS
              colorlist=translate(colorlist,' ',',')
              if colorlist<>'' then do
                  rep_stuff.iarf2.!colors.!LEN=words(colorlist)
                  do mm=1 to rep_stuff.iarf2.!colors.!LEN
                     rep_stuff.iarf2.!colors.mm=strip(word(colorlist,mm))
                     if mm=words(colorlist) then            /* a time saving hack */
                         rep_stuff.iarf2.!colors.0=rep_stuff.iarf2.!colors.mm
                  end 
               end              /* colorlist <>'' */
               else do
                  rep_stuff.iarf2.!colors.!LEN=0
               end 
            end         /* parsing ?COLORS for this ?TR (or ?TH or ?TD) instruction */

            if rep_stuff.iarf2.!COLORS.!LEN=0 then do   /* no ?COLORS, so nothing fancy to do  */
                 thisgoo=thisgoo||'<'||keyword||' '||options||'>'
            end
            else do
              icuse=iarf//rep_stuff.iarf2.!colors.!LEN   /* if here, use one of the ?colors */
              kkolor=rep_stuff.iarf2.!colors.icuse
              if abbrev(kkolor,'$')=1 then do
                 kkolor2=substr(kkolor,2)
                 kkolor=process_cms('SHOW  ',kkolor2)        
              end 
              else do
                   kkolor='#'||kkolor
              end

              options2=rep_stuff.iarf2.!options||' bgcolor="'||kkolor||'" '
              thisgoo=thisgoo||'<'||keyword||' '||options2||'>'
            end
        end            

        
        otherwise do
           foo=note_error('W','Unimplemented CMS keyword (REPEAT): '||keyword)
           thisgoo=thisgoo||foo
         end
      end                       /* keyword processing (KEYWORD select) */

/* Iter found? */
     if icontrol>0 then leave           /* stop processing record (possibly all records */

   end                           /* repeat block instruction (IARF2) */

/* drop record? */
   if idrop=0 then do           /* no, add this to output */
        goo=goo||thisgoo
   end 
   else do
      info.!!count=info.!!count-1                 /* de augment "processed records" counter */
      info.!!count_block=info.!!count_block-1
   end 

   if icontrol>1 then  leave    /* stop processing all records */
end                             /* each record (IARFX) */



/* remove the CMS varnames from CINFO. */
lst1=CINFO.0
if info.!ADD_LOCALS=1 then do
  do ww=1 to info.!series_varname.0
     aww=info.!series_varname.ww
     parse var lst1 l1 (aww) l2
     lst1=l1||' '||l2
     vw='!'||strip(aww)
     CINFO.vw=''
  end
  parse var lst1 l1 '!ID' l2
  CINFO.!!ID=''
  lst1=l1||' '||l2
end

if willmake+willnotmake>0 then do
     tt=willmake+willnotmake
     fpp=note_error('M','Creating '||willmake||' of '||tt||' singleton-archive files.',1)
end

ss=special_vars
do until ss=''       /* remove special vars */
       parse var ss avv ss ; avv=strip(avv)
       parse var lst1 l1 (avv) l2
       lst1=l1||' '||l2
       avv2='!'||avv
       CINFO.avv2=''
end
CINFO.0=space(lst1)
return goo

/*********************/
/* read & set conditions */
read_conds:procedure expose rep_stuff. gerrs.  update_verb 
parse arg optlist,iarf2

if optlist='' then do    /* simplest case  */
   rep_stuff.iarf2.ith.!CONDS=0
   return 0
end 

ith=0

/* syntax is:  cond "value" , cond "value" , ....
   values MUST be in quotes! */

optlist=strip(optlist)

if words(optlist)=1 | abbrev(optlist,'"')=1 then do     /* simple case (1 value) */
   if abbrev(optlist,'"')=1 then optlist=strip(optlist,,'"')
   rep_stuff.iarf2.1.!COND='EQ'
   rep_stuff.iarf2.1.!VALUE=optlist
   rep_stuff.iarf2.1.!CONDS=1
   return 1
end

/* possibly multiple, or non-equality, conditions */
do until optlist=''
   parse var optlist acond '"' avalue '"' . ',' optlist
   if avalue='' then iterate

   acond=strip(translate(acond))
   ith=ith+1
   rep_stuff.iarf2.ith.!COND=acond
   rep_stuff.iarf2.ith.!VALUE=avalue
end
rep_stuff.iarf2.ith.!CONDS=ith
return ith
return 0



/*******************/
/* check the case conditions */
check_case:procedure expose rep_stuff. gerrs.  update_verb 
parse arg rvalue,nconds,thisindx                       /* rvalue: val of var for this record */

do mm=1 to nconds
   acond=rep_stuff.thisindx.mm.!COND
   cvalue=rep_stuff.thisindx.mm.!VALUE  /* condition value */
   
   select
      when acond='EQ' then do
         if rvalue<>cvalue then return 0
      end
      when acond='GE' then do
         if rvalue<cvalue then return 0
      end
      when acond='GT' then do
         if rvalue<=cvalue then return 0
      end
      when acond='LE' then do
         if rvalue>cvalue then return 0
      end
      when acond='LT' then do
         if rvalue>=cvalue then return 0
      end
      when acond='NE' then do
         if rvalue=cvalue then return 0
      end
      when acond='IN' then do
         if wordpos(rvalue,cvalue)=0 then return 0
     end
      when acond='INI' then do
         if wordpos(translate(rvalue),translate(cvalue))=0 then return 0
     end

     otherwise do
       foo=note_error('E','Bad condition in CASE : '||acond)
       return -1                 /* bad condition */
     end
  end
end
return 1                /* if here, no failure */


/****************/
/* process a ? COMMENT */
mk_comment1:procedure expose thisgoo rep_stuff.  options tplfile outfile  dolist_c gerrs. recids. daindx root thisaddon  update_verb ,
                                cmsdir_files cmsdir_uploads cmsdir_lib stuff. info. cinfo.  optname. cmsdir cmsdir_pol
parse arg iarf2,uid,miscvars
auid=translate(uID,'__',' .')
if rep_stuff.iarf2.0=0 then do               /* no ?options, or error in an ? option */
    return  ' <A '||options||'>'
end

if rep_stuff.iarf2.0='' then do  /* load some info */

      call parse_options(options)  

      rep_stuff.iarf2.!options=ropts.!0

      cmtmode=wordpos(strip(Translate(ropts.!?MODE)),'SHOW UPDATE CREATE ADD NAME AUTO ')  /* default is "show" */
      if cmtmode=0 then cmtmode=1  /* default is "SHOW" */
      if cmtmode=6 then do               /* update/create depending on user setting */
        if optname.!updatearchive=1 then  /* update is enabled */
           cmtmode=2
        else                    /* updated not enable means "create all" */
           cmtmode=3
      end
      rep_stuff.iarf2.!CMTMODE=cmtmode
      if cmtmode=4 then do              /* add an "add comment" link */

         rep_stuff.iarf2.0=1  /* signal okay */

         cmttfile=strip(translate(ropts.!?FILE))
         if cmttfile=0 then cmttfile=''
         cx=cmttfile
         if cmttfile<>'' then  do
            cmtdir=stuff.!FILEDIR
            if cmtdir<>'' then
               cmttfile=fix_file(cmttfile,strip(cmtdir,,'\')||'\')
            else
               cmttfile=fix_file(cmttfile,strip(cmsdir_lib,,'\')||'\')
            if stream(cmttfile,'c','query exists')='' then do
              cmttfile=fix_file(cx,strip(cmsdir_lib,,'\')||'\')
              if stream(cmttfile,'c','query exists')='' then do
                 foo=note_Error('E','COMMENT: i) template file does not exist: 'cmttfile)
                 rep_stuff.iarf2.0=0 
                 return ''
              end
            end
         end
         rep_stuff.iarf2.!cmttfile=cx

         goo=root||'CMS_CMT.CMD'
         rep_stuff.iarf2.!cms_cmt=goo

         cmtresp=space(translate(ropts.!?RESPFILE),0)
         if cmtresp=0 then cmtresp=''
         foof=cmtresp
         if cmtresp<>''  then do
            cmtdir=stuff.!FILEDIR 
            if cmtdir<>'' then
               foof=fix_file(cmtresp,strip(cmtdir,,'\')||'\')
            else
               foof=fix_file(cmtresp,strip(cmsdir_lib,,'\')||'\')
            if stream(foof,'c','query exists')='' then do
              foofold=foof
              foof=fix_file(cmtresp,strip(cmsdir_lib,,'\')||'\')
              if stream(foof,'c','query exists')='' then do
                foo=note_Error('E','COMMENT: i) no such response file: 'foof ' or ' foofold)
                rep_stuff.iarf2.0=0 
                return ''
              end
            end
         end
         rep_stuff.iarf2.!cmtresp=cmtresp

/* don't bother setting other stuff */
         tline1=''
         if cmtresp<>'' then  tline1='&respfile='||cmtresp
         tline2=''
         if cx<>'' then  tline2='&file='||cx
          t1='<a '||rep_stuff.iarf2.!options||,
                   ' href="'||rep_stuff.iarf2.!cms_cmt||,
                   '?deffile='||daindx||tline1||tline2||,
                   '&action=3&comment='||uid||'">'
         return t1

      end               /* add mode */

      cmttfile=strip(translate(ropts.!?FILE))
      cmttfile=fix_file(cmttfile,tplfile)
   
      if cmttfile='' then do
         foo=note_Error('E','COMMENT: the template file not specified: 'cmttfile)
         rep_stuff.iarf2.0=0 
         return ''
     end
     if stream(cmttfile,'c','query exists')='' then do
        if cmtmode=2 | cmtmode=3 then do
           foo=note_Error('E','COMMENT: template file does not exist: 'cmttfile)
           rep_stuff.iarf2.0=0 
           return ''
        end
      end
      rep_stuff.iarf2.!cmttfile=cmttfile

      cmtofile=strip(translate(ropts.!?OUTPUT))
      if cmtofile='' then do
                iip=lastpos('.',outfile)
                if iip=0  then iip=length(outfile)+1
                p1=left(outfile,iip-1)
                p2=substr(outfile,iip+1)
                cmtofile=p1||'$C*.'||p2
      end
      if pos('*',cmtofile)=0 then do
               foo=note_Error('E','COMMENT; output file does not contain * : 'cmtofile)
               rep_stuff.iarf2.0=0 
               return ''
      end
      if pos(':',cmtofile)=0 &  abbrev(cmtofile,'\')=0 then do
          outdir=stuff.!FILEDIR
          if outdir='' then outdir=cmsdir_files
          outdir=strip(strip(outdir),,'\')
          cmtofile=strip(strip(outdir,,'\'))||'\'||cmtofile
       end 

      parse var cmtofile p1 '*' p2
      cmtofile1=p1||auid||p2
      t1=strip(filespec('d',cmtofile1)||filespec('p',cmtofile1),,'\')
      if right(t1,1)=':' then t1=t1||'\'
      if dosisdir(t1)=0 then do
            foo=note_Error('E','COMMENT: no such directory for comment files: 't1)
            rep_stuff.iarf2.0=0 
            return ''
      end
      rep_stuff.iarf2.!cmtofile=cmtofile

      cmturl=strip(translate(ropts.!?URL))
      cmturl=strip(cmturl,'t','/')||'/'
      rep_stuff.iarf2.!cmturl=cmturl

      retains=strip(translate(ropts.!?RETAIN))
      if retains<>'' then do
        vv=strip(translate(retains))
        if wordpos(vv,info.!SERIES_VARNAMES)=0 then do
               foo=note_Error('E','COMMENT; no such variable to retain: 'retains)
               rep_stuff.iarf2.0=0 
               return ''
        end 
        rep_stuff.iarf2.!RETAINS='!'||vv
      end

      cmtresp=space(translate(ropts.!?RESPFILE),0)
      if cmtresp=0 then cmtresp=''
      if cmtresp<>''  then do
        cmtdir=stuff.!FILEDIR 
        if cmtdir='' then cmtdir=cmsdir_lib
        foof=strip(cmtdir,,'\')||'\'||strip(cmtresp,,'\')
        if stream(foof,'c','query exists')='' then do
            foof=strip(cmsdir_lib,,'\')||'\'||strip(cmtresp,,'\')  /* often redundant, but so what */
            if stream(foof,'c','query exists')='' then do
               foo=note_Error('E','COMMENT: ii) no such response file: 'foof)
               rep_stuff.iarf2.0=0 
               return ''
           end
        end
      end
      rep_stuff.iarf2.!cmtresp=cmtresp

      rep_stuff.iarf2.!QUIET=ropts.!?QUIET 

      rep_stuff.iarf2.0=1  /* signal okay */

end


/*  process this record */
cmtofile=rep_stuff.iarf2.!cmtofile
cmttfile=rep_stuff.iarf2.!cmttfile
cmturl=rep_stuff.iarf2.!cmturl
cmtmode=rep_stuff.iarf2.!cmtmode
cmtresp=rep_stuff.iarf2.!cmtresp
isquiet=rep_stuff.iarf2.!QUIET
retainvar=rep_stuff.iarf2.!RETAINS

if cmtmode=4 then do              /* just add an "add comment" link */
   tline1=''
   if cmtresp<>'' then  tline1='&respfile='||cmtresp
   tline2=''
   if cmttfile<>'' then  tline2='&file='||cmttfile
   t1='<a '||rep_stuff.iarf2.!options||,
             ' href="'||rep_stuff.iarf2.!cms_cmt||,
             '?deffile='||daindx||tline1||tline2||,
             '&action=3&comment='||uid||'">'
   return t1
end 


cmtofile1=fix_file(cmtofile,outfile,,auid,1)   /* choose a file name */
 
uname=filespec('n',strip(cmtofile1))
aurl=cmturl||uname
cstrg.=''
cstrg.2='Update '
cstrg.3='Create'


if cmtmode=2 | cmtmode=3 then do            /* create or update? */

/* tplfile incfile sortvar sortorder sorttype  isarc ',' ,
    vSELECT '09'x desc_tpl '09'x da_arg  '09'x aaurl  '09'x vSELECTVAL 
     '09'x retainval */

/* now make a list of comment file creation instructions -- note that TCC
   signals the  mode */
    tcc=10+cmtmode
    adolist=cmttfile||' '||cmtofile1||' $RECID A I '||tcc||',SORT '||'09'x|| ,
                 miscvars||'09'x|| ,
                 uid||'09'x||aurl||'09'x||uid
    if retainvar<>'' then
        adolist=adolist||'09'x||cinfo.retainvar

/* idea is to retain comments whose $RECID is this uid */
   dolist_c=adolist||'0d0a'x||dolist_c  /* add to list of comment files to create/update */
end

if isQUIET=1 then return ''

if cmtmode=5 then return aurl            /* just the URL, no html */

return '<a href="'||aurl||'" '||rep_stuff.iarf2.!OPTIONS||' >'


/****************/
/* create a link to a permanent location */
do_link:procedure expose rep_stuff. info. outfile cshdata. gerrs. recids.  update_verb  optname.
parse arg options,uid,iarf2,aid7

if rep_stuff.iarf2.!A=-1 then return ''

if rep_stuff.iarf2.!A=0 then do   /* no ?xx modifiers, so nothing fancy to do  */
    return '<A '||options||'>'
end

if rep_stuff.iarf2.!A='' then do  /* see if ?URL or ?LINK, and parse it  */
    rep_stuff.iarf2.!A=1
    call parse_options(options)
    rep_stuff.iarf2.!options=ropts.!0
    rep_stuff.iarf2.!URL=strip(ropts.!?URL)
    rep_stuff.iarf2.!NAME=strip(ropts.!?NAME)
    rep_stuff.iarf2.!TYPE=translate(strip(ropts.!?TYPE))
    rep_stuff.iarf2.!OUTPUT=strip(ropts.!?OUTPUT)


    if rep_stuff.iarf2.!URL='' & rep_stuff.iarf2.!NAME=''  &  rep_stuff.iarf2.!TYPE='' & ,
           rep_stuff.iarf2.!OUTPUT=''  then do
      rep_stuff.iarf2.!A=0
      return '<A '||options||'>'
    end

/* fix up output file */
    ofile=strip(translate(rep_stuff.iarf2.!OUTPUT))
    if ofile='' then do
          iip=lastpos('.',outfile)
          if iip=0  then iip=length(outfile)+1
          p1=left(outfile,iip-1)
          p2=substr(outfile,iip+1)
          ofile=p1||'$A*.'||p2
    end
    if pos('*',ofile)=0 then do
            foo=note_Error('E','LINK; output file does not contain * : 'ofile)
            rep_stuff.iarf2.!A=-1 
            return ''
    end
    ofile=fix_file(ofile,outfile,1)

    t1=strip(filespec('d',ofile)||filespec('p',ofile),,'\')
    if right(t1,1)=':' then t1=t1||'\'
    if dosisdir(t1)=0 then do
            foo=note_Error('E','LINK: no such directory for output files: 't1)
            rep_stuff.iarf2.!A=-1 
            return ''
    end
    rep_stuff.iarf2.!OUTPUT=ofile

/* fix up url */
   curl=strip(translate(rep_stuff.iarf2.!URL))
   curl=strip(curl,'t','/')||'/'
   rep_stuff.iarf2.!URL=curl

/* fix up type */
   ctype=strip(translate(rep_stuff.iarf2.!TYPE))
   select 
     when ctype='SORT' then do
        iis=info.!sorted 
        if iis='' | iis='!ID' then do
          foo=note_Error('E','LINK type=SORT, but SORT variable not specified.')
          rep_stuff.iarf2.!A=-1 
          return ''
        end
     end
     when datatype(ctype)='NUM' then do
        ctype=trunc(ctype)
        if ctype <1 then do
            foo=note_Error('E','LINK: type=nnn must be integer > 0: 'ctype)
            rep_stuff.iarf2.!A=-1 
            return ''
        end
     end 
     otherwise do
       if wordpos(ctype,'HOUR DAY DAY2 WEEK WEEK2 MONTH MONTH2 YEAR')=0 then do
           foo=note_Error('E','LINK: unknown type: 'ctype)
           rep_stuff.iarf2.!A=-1 
           return ''
       end 
     end
   end         /* select */
end             /* !A='' */


arctype=rep_stuff.iarf2.!TYPE
aval=uid
select
  when datatype(arctype)='NUM' then do
     aval=trunc( (jj-0.1)/arctype)
     adisp=((aval)*arctype)+1
     adisp2=min(adisp+arctype-1,recids.0)
     adisp=adisp||' to '||adisp2
  end
  when arctype="SORT" then do
     aval=strip(translate(aval))
     adisp=dsort||': '||aval
   end
   when arctype='HOUR' then do
       aa=aval
       aval=trunc(aval*24)
       adate=trunc(aa)+726467
       tt=strip(dateconv(adate,'B','N'))
       ahr=trunc((aa-trunc(aa))*24)
       adisp=tt||' '||ahr||':00'
   end
   when arctype='DAY' | arctype='DAY2' then do 
       aval=trunc(aval)
       adate=trunc(aval)+726467
       if arctype='DAY' then
           adisp=strip(dateconv(adate,'B','N'))   /* 27 aug 1999 */
       else
           adisp=strip(dateconv(adate,'B','U'))   /* 08/27/99 */
  end 
  when arctype='WEEK' | arctype='WEEK2' then do
      tt=determine_week(aval)
      parse var tt aval ',' . ',' wkb ',' wke
      if arctype='WEEK2' then do
           wke=dateconv(wke,'N','U')
           wkb=dateconv(wkb,'N','U')
           adisp=wkb||'-'||wke
      end
      else do
           adisp=wkb||' to '||wke
      end
  end           
  when arctype='MONTH' | arctype='MONTH2' then do
      adate=trunc(aval)+726467
      tt=strip(dateconv(adate,'B','S'))   /* 19990325 */
      aval=left(tt,6,'0')
      if arctype="MONTH2" then do
           adisp=dateconv(adate,'B','U')
           parse var adisp amo '/' . '/' ayr
           adisp=amo'/'ayr
      end
      else do
          adisp=subword(strip(dateconv(adate,'B','L')),2,2)
      end
  end
  when arctype='YEAR' then do
     adate=trunc(aval)+726467
     tt=strip(dateconv(adate,'B','S'))   /* 19990325 */
     aval=left(tt,4,'0')
     adisp=subword(strip(dateconv(adate,'B','N')),3,1)
  end
  otherwise do 
     nop
  end
end

/* now compute name */
ithset='_'||adisp
ithset=space(translate(ithset,' $$_  ',' /\-:.'),0)
ithset=translate(ithset,'_',' ')
if length(ithset)>30 then ithset='_'||ithset0
ofile=rep_stuff.iarf2.!output

ofile1=fix_file(ofile,outfile,,ithset,1)   /* choose a file name */


/* the #locald */
locald=''
if rep_stuff.iarf2.!NAME<>'' then do
    vurl=process_cms('VAR ',,aid7,rep_stuff.iarf2.!NAME)  
    if abbrev(vurl,'<!--')=1 then return ''             /* error */
    if rep_stuff.iarf2.!NAME='!ID' then vurl=translate(vurl,'_','.')
    if vurl<>'' then locald='#'||vurl
end


uname=filespec('n',strip(ofile1))
aurl=rep_stuff.iarf2.!URL||uname||locald

vv='<A href="'||aurl||'" '||rep_stuff.iarf2.!options||'>'
return vv



/****************/
/* add, to CINFO., the values of the current record's cms variables */
/* Hint:
    info.!series_varname.mm=aname
    info.!series_vartype.mm=strip(translate(word(stuff.!vartypes,mm)))
*/

more_cinfo:procedure expose CINFO. info. cshdata. gerrs.  recids.  update_verb 
parse arg aid7,icc,defdir,defdir_up

/* expects results returned as id tab var1 tab var2 tab ..  */


astat=sre_cache_data_local(info.!series,'FIND ID7',aid7,'*')
file_vars=''
parse var astat . '0d0a'x vlist   /* if failure, then below sets values to '' */
vid='!!ID'
parse var vlist tt '09'x lvals
CINFO.vid=tt

cinfo.!!COUNT=info.!!count
cinfo.!!COUNT_BLOCK=info.!!COUNT_BLOCK


do k1=1 to info.!series_varname.0
  parse var lvals aval '09'x lvals
  vw='!'||info.!series_varname.k1
  ttype=info.!series_vartype.k1
  if ttype='M' & aval<>'' then aval=sre_pack64(aval)

  if tt='F'  & aval<>'' then do
       file_vars=file_vars||' '||info.!series_varname.k1
       parse var aval origname '=' dafile ';' uptype
       if abbrev(dafile,'\')=0 & pos(':',dafile)=0 then do     /* relative filename */
              dafile=defdir_up||dafile                     /* convert to fully qualified */
              aval=origname||'='||dafile||' ;'||uptype
       end
  end
  CINFO.vw=aval
  CINFO.vw.!TYPE=ttype             /* used internally */
end 

CINFO.!!CURRENT=icc

CINFO.!!FILE_VARS=file_vars

return 1



/*********************/
/* find GOTO information for IF and GOTO keywords (as stored in REP_STUFF.) */
find_gotos:procedure  expose rep_Stuff. info. gerrs.  recids.  update_verb 

info.!ADD_LOCALS=0       
if_at=''
labels.=''
gotos.=''
case_at=''
case_list=''

chk_goto.=''

do jj=1 to rep_stuff.0
   akey=rep_stuff.jj.1
   select
     when abbrev(akey,'CAL')=1 then do 
        info.!add_locals=1
     end 
     when abbrev(akey,'HTML')=1 then do   /* html might have ? options */
        info.!add_locals=1
     end 

     when abbrev(akey,'VAR')=1 then do       /* VAR might call a meta function  */
       info.!add_locals=1
     end
     when abbrev(akey,'COMMENT')=1 | akey='A' then do  /* comment might have a RETAIN */
       info.!add_locals=1                              /* A might have a name= */
     end
     when abbrev(akey,'IF')=1 then do       /* remember that jj is location of an IF */
       info.!add_locals=1
       if_at=jj||' '||if_at             /* use if_at list to support embedded IFs (pop off front) */
       rep_stuff.jj.!GOTO=-1     /* -1 means "waiting to be set"  */

  /*     rep_stuff.jj.!GOTO=rep_stuff.0+1  default is end of content */
     end 

/* else is simply an IF with "not IF-result" condition. If-results are stored in a stack */
     when abbrev(akey,'ELSE') then do       
         if_at=jj||' '||if_at             /* use if_at list to support embedded IFs and ELSEs (pop off front) */
         rep_stuff.jj.!GOTO=-1     /* -1 means "waiting to be set"  */

/*         rep_stuff.jj.!GOTO=rep_stuff.0+1  default is end of content */

         iterate
     end

     when abbrev(akey,'/IF')=1  then do       /* this is where the IF should jump beyond, or a ELSE should jump to (if If is true) */
         parse var if_at jj0 if_at ;jj0=strip(jj0)
         rep_stuff.jj0.!GOTO=jj-1      /* jump here, note that /IF is treated as a <? COMMENT ... > */
         iterate
     end

     when abbrev(akey,'/ELSE')=1  then do       /* this is where the ELSE should jump beyond, or a ELSE should jump to (if If is true) */
         parse var if_at jj0 if_at ;jj0=strip(jj0)
         rep_stuff.jj0.!GOTO=jj      /* jump here, note that /ELSE is treated as a <? COMMENT ... > */
         iterate
     end

     when abbrev(akey,'CASES') then do  /* begin case-list (location of /CASES will be set for this iarf2*/
         info.!add_locals=1
         case_at=jj||' '||case_at         /* list of begincase statements */
     end

     when abbrev(akey,'/CASES') then do  /* end of cases */
         info.!add_locals=1
        parse var case_at bcase case_at           /* pop of "iarf2 of CASES" */
        bcase=strip(Bcase)
        rep_stuff.bcase.!ENDCASE=jj
     end

     when abbrev(akey,'CASE') then do  /* begin case (.!GOTO is set by /CASE, !BEGIN  set here) */
        info.!add_locals=1
        parse var case_at bcase . ; bcase=strip(bcase)
        if bcase='' then foo=note_error('E','Missing CASES statement ')
        rep_stuff.jj.!BEGIN=bcase
        case_list=jj||' '||case_list       /* /CASE will pop this off */
        rep_stuff.jj.!GOTO=-1     /* -1 means "waiting to be set"  */

/*         rep_stuff.jj.!GOTO=rep_stuff.0+1  default is end of content */

     end

/* look up icase=iarf2.!CASE. If icase.!TRUE=1, then set ibegin=icase.!BEGIN,  and goto
   ibegin.!ENDCASE. Otherwise, do next line */

     when abbrev(akey,'/CASE') then do          
        info.!add_locals=1
        parse var case_list ccase case_list ; ccase=strip(ccase)
        if ccase='' then foo=note_error('E','Missing CASE statement ')

        rep_stuff.ccase.!GOTO=jj

        parse var case_at bcase . ; bcase=strip(bcase)
        rep_stuff.jj.!BEGIN=bcase
     end
        

     when abbrev(akey,'GOTO')=1 then do  /* remember that jj is location of a GOTO */
        aopt=rep_stuff.jj.2
        parse upper var aopt golabel . ; golabel=strip(golabel)
        vv='!'||golabel
        gotos.vv=jj
        gotos.vv.!IF_AT=if_at
        gotos.vv.!CASE_AT=case_list
        rep_stuff.jj.!GOTO=-1     /* -1 means "waiting to be set"  */
        gotos.!0=gotos.!0||' '||vv
     end 

     when abbrev(akey,'LABEL')=1 then do  /* remember that jj is location of a GOTO */
        aopt=rep_stuff.jj.2
        parse upper var aopt golabel . ; golabel=strip(golabel)
        vv='!'||golabel
        labels.vv=jj
        labels.vv.!IF_AT=if_at
        labels.vv.!CASE_AT=case_list
        labels.!0=labels.!0' '||vv
     end

     otherwise nop
   end          /*select */

end     /* rep_stuff.0 */

/* now assign the labels to the GOTOs */
do until gotos.!0=''
   parse var gotos.!0 vlabel gotos.!0 ; vlabel=strip(vlabel)
   if vlabel='' then leave
   labelat=labels.vlabel
   if labelat='' then do
      iterate  /* if no such label, error (will be caught below) */
   end
   gotoat=gotos.vlabel

   goto_ifat=gotos.vlabel.!IF_AT
   label_ifat=labels.vlabel.!IF_AT
   if goto_ifat<>label_ifat then do     /* check for "goto inside of an if */
      parse var goto_ifat . (label_ifat) mook
      jlen=length(label_ifat)
      if jlen>length(goto_ifat) then do
         foo=note_error('E','Attempt to GOTO a LABEL inside of an IF ('||rep_stuff.gotoat.2||')')
         iterate
      end 
      if strip(right(goto_ifat,jlen))<>strip(label_ifat) then do
         foo=note_error('E','Attempt to GOTO a LABEL inside of an IF ('||rep_stuff.gotoat.2||')')
         iterate
      end 
   end 

   goto_caseat=gotos.vlabel.!CASE_AT
   label_caseat=labels.vlabel.!CASE_AT

   if goto_caseat<>label_caseat then do     /* check for "goto inside of an if */
      parse var goto_caseat . (label_caseat) mook
      jlen=length(label_caseat)
      if jlen>length(goto_caseat) then do
         foo=note_error('E','Attempt to GOTO a LABEL inside of a CASE ('||rep_stuff.gotoat.2||')')
         iterate
      end 
      if strip(right(goto_caseat,jlen))<>strip(label_caseat) then do
            foo=note_error('E','Attempt to GOTO a LABEL inside of a CASE ('||rep_stuff.gotoat.2||')')
            iterate
      end 
   end 

   rep_stuff.gotoat.!GOTO=labelat
end

/* now check for unassigned goto, else, if, and case */
nerrs=0
do mm=1 to rep_stuff.0
    igg=rep_stuff.mm.!GOTO
    if igg<>-1 then iterate
    foo=note_error('E','Unmatched '||rep_stuff.mm.1||', with: '||rep_stuff.mm.2)
    nerrs=nerrs+1
end 

return nerrs

/*********************/
/* convert string (from an IF) into a value */
if_value:procedure expose CINFO. gerrs. info.  recids. defdir_up  update_verb 
parse arg a1
a1=strip(a1)

thisid=cinfo.!!id

/* enclose in ", then it's a literal value */
if abbrev(a1,'"')=1  then do
    parse var a1 . '"' val1 '"' .
    return val1
end

if datatype(a1)='NUM' then return a1

/* is it a "meta-data" function? */
if pos('(',a1)>0 then do
   vval=if_function(a1)
   return vval
end 

/* it's a variable */
a1=translate(a1)
ijj=wordpos(a1,cinfo.0)
if ijj=0 then do
   fpp=note_error('E','Undefined variable in IF: '||a1)
   return ''
end 

va1='!'||translate(a1)
val1=CINFO.va1

return val1

/************************/
/* do a set function */
fig_set_function:procedure expose cinfo. info. gerrs.  update_verb 
parse arg lval

parse var lval afunc '(' lval2  ')' .
funcn=strip(Translate(afunc))
if funcn='' then funcn='IDENTITY'

parse var lval2 t1 ',' t2 ',' .

select                                       /* use this varname below */
    when funcn="REPLACE" | funcn='LENGTH' | funcn='SUBSTR' | funcn='SUBWORD' | funcn='SUBLINE' then
          lname=strip(Translate(t1))
    when funcn='POS' | funcn='WORDPOS' then lname=strip(Translate(t2))
    when funcn='DATETIME' then  do
                if t2<>'' then
                  lname=strip(Translate(t2))
                else
                  lname='!ID'
    end
    otherwise   lname=strip(Translate(t1))              
end
jj1=wordpos(lname,'!ID  '||info.!series_varnames)
if jj1>0 then do
        fpp=note_error('W','Attempting to SET CMS variable (use VAR instead): 'lname)
         return ''            /* use VAR or SERIES to display cms vars, ID, or SERIES name */
end
jj2=wordpos(lname,cinfo.0)
if jj2=0 then do
      fpp=note_error('W','Attempting to SET undefined variable: 'lname)
      return ''            /* use VAR or SERIES to display cms vars, ID, or SERIES name */
end
vv='!'||strip(lname)
varval=CINFO.vv

lval3=do_var_function(funcn,varval,lval2)
return lval3
 
/************************/
/* do a var function */
do_var_function:procedure expose gerrs.  update_verb 
parse arg a1,varval,opts0

select

when a1='IDENTITY' | a1="" then return varval

when a1='LENGTH' then do
  return length(varval)
end

when a1='WORDS' then do
  varval=translate(varval,' ','0d0a09'x)
  return words(varval)
end

when a1='LINES' then do
  icrlf=pos('0d0a'x,varval)
  ilines=1
  do until icrlf=0
       ilines=ilines+1
       icrlf=pos('0d0a'x,varval,icrlf+2)
  end 
  return ilines
end

when a1='SUBSTR' then do
  parse  var opts0 avar ',' istart ',' ilen ;istart=strip(istart) ;ilen=strip(ilen)
  if datatype(istart)<>'NUM' then istart=1
  if datatype(ilen)<>'NUM' then 
     goo2=substr(varval,istart)
  else
     goo2=substr(varval,istart,ilen)
  return goo2
end  

when a1='HTML_MAKE' then do  /* convert plain text to html */
  newval=''
  do until varval=''
      parse var varval aline '0d0a'x varval
      aline=sre_replacestrg(aline,'09'x,'&nbsp;&nbsp;&nbsp;','ALL')
      if aline='' then do
         newval=newval||'<p>'||'0d0a'x
         iterate
      end 
      nonsp=verify(aline,' ')
      if nonsp>3 then do
         aa=''
         do jj=1 to nonsp-2
            aa=aa||'&nbsp;'
         end    
         aline=aa||' '||strip(aline)
      end
      if pos('.',aline)=0 then do
         newval=newval||' '||aline||'0d0a'x
      end 
      else do           /* possibly a url (any word with two .'s is a url */
         newline=''
         do until aline=''
            parse var aline aword aline
            if words(translate(aword,' ','.'))<3 then do
              newline=newline||' '||aword
              iterate
            end
            ahref="href"
            if pos('@',aword)>0 then ahref='mailto'
            ac=' '
            if pos(right(aword,1),',;:')>0 then do
               ac=right(aword,1)
               aword=left(aword,length(aword)-1)
            end 
            aslash='' ; if abbrev(translate(aword),'HTTP://')=0 then aslash='http://'
            aword=' <a '||ahref||'="'||aslash||strip(aword)||'">'||aword||'</a>'||ac
            newline=newline||aword
         end
         newval=newval||' '||newline||'0d0a'x
      end
  end 
  return newval

end

when a1="HTML_REMOVE"  then do  /* remove html tags */
  parse upper var opts0 avar ',' aopt ; aopt=strip(aopt)
  newval=''
  if aopt='DROP' then do
    do until varval=''
       parse var varval p1 '<' p2 '>' varval
       newval=newval||p1
    end 
  end
  else do
     newval=sre_html_encode(varval)
  end
  return newval
end


when a1='SUBWORD' then do
  parse  var opts0 avar ',' istart ',' ilen ;istart=strip(istart) ;ilen=strip(ilen)
  if datatype(istart)<>'NUM' then istart=1
  varval=translate(varval,' ','0d0a09'x)
  if datatype(ilen)<>'NUM' then 
     goo2=subword(varval,istart)
  else
     goo2=subword(varval,istart,ilen)
  return goo2
end  

when abbrev(a1,'INC')=1 then do
   if datatype(varval)<>'NUM' then return ''
   tmp=varval+1
   return tmp
end

/*    newstring=sre_replacestrg(astring,target,putme,type,exact) 
exact:	 if 1, then do not capitalize during search (exact match only)
TYPE:
#  not specified : replace first occurence of target
# BACKWARD : replace last occurence of target
# ALL : replace all occurences of target
*/

when a1='REPLACE' then do
  parse  var opts0 avar ',' needle ',' newthing ',' rtype ',' rexact
  if abbrev(needle,'"')=1 then needle=substr(needle,2)
  if abbrev(newthing,'"')=1 then newthing=substr(newthing,2)
  if right(needle,1)='"' then needle=left(needle,length(needle)-1)
  if right(newthing,1)='"' then newthing=left(newthing,length(newthing)-1)
  rtype=strip(rtype)
  useme=sre_replacestrg(varval,needle,newthing,rtype,rexact)
  return useme
end 

when a1='SUBLINE' then do
  parse  var opts0 avar ',' istart ',' ilen ;istart=strip(istart) ;ilen=strip(ilen)
  if datatype(istart)<>'NUM' then istart=1
  if datatype(ilen)<>'NUM' then  ilen=100000000
  do jj=1 to istart-1   /* skip to desired line */
      parse var varval . '0d0a'x varval
  end
  useme=''

  do jj=1 to ilen
      if varval='' then leave
      parse var varval aline '0d0a'x varval
      if useme='' then
          useme=aline
      else
         useme=useme||'0d0a'x||aline
   end
   return useme
end       

when a1='POS' then do
  parse  var opts0 astring ',' avar ',' istart ',' ci ; istart=strip(istart); 
  astring=strip(astring)
  ci=strip(translate(ci))
  if abbrev(astring,'"')=1 then astring=strip(astring,,'"')
  if ci='I' then do
     varval=translate(varval)
     astring=translate(astring)
  end
  if datatype(istart)<>'NUM' then istart=1
  goo2=pos(astring,varval,istart)
  return goo2
end  

when a1='WORDPOS' then do
  parse  var opts0 astring ',' avar ',' istart ',' ci ; istart=strip(istart); 
  astring=strip(astring)
  ci=strip(translate(ci))
  
  if abbrev(astring,'"')=1 then astring=strip(astring,,'"')
  if ci='I' then do
     varval=translate(varval)
     astring=translate(astring)
  end
  if datatype(istart)<>'NUM' then istart=1
  goo2=wordpos(astring,TRANSLATE(varval,' ','0D0A09'X),istart)
  return goo2
end  


when a1='DATETIME' then do

  parse upper var opts0 opts ',' .
  atime0=varval 
  parse var atime0 adate1 '.' atime1
  atime='0.'||atime1
  adate=adate1+726467
  tt=strip(dateconv(adate,'B','S'))   /* 19990325 */
  opts=strip(opts)
  select
     when (abbrev(opts,'N')=1) | opts='' then do
        rr=fig_time(varval)
        return rr
     end
     when abbrev(opts,'Y')=1 then do
       return left(tt,4,'0')
     end
     when abbrev(opts,'MO')=1 then do
       return left(tt,6,'0')
     end
     when abbrev(opts,'W')=1 then do
         foo=determine_week(adate1)
         parse var foo wn ',' .
         return wn
     end
     when abbrev(opts,'D')=1 then do
       return left(tt,8,'0')
     end

     when abbrev(opts,'H')=1 then do
         zz=digits()
         numeric digits 16
         hour=trunc(86400*atime/3600)
         numeric digits zz
         aa=tt||':'||right(hour,2,'0')
         return aa
    end
    when abbrev(opts,'MI')=1 then do
         zz=digits()
         numeric digits 16
         amin=trunc(86400*atime/60)
         numeric digits zz
         aa=tt||':'||right(amin,4,'0')
         return  aa
    end
    when abbrev(opts,'S')=1 then do
         zz=digits()
         numeric digits 16
         secs=trunc(atime*86400)
         numeric digits zz
         aa=tt||':'||right(secs,5,'0')
         return aa
    end
    otherwise do
         foo=note_error('E','Unknown DATETIME option: 'opts0)
         return ''
    end
  end  /* DATETIME select */
end 
otherwise do
         foo=note_error('E','Unknown function: 'a1)
         return ''
end

end

return ''

/****************/
/* lookup a var-function or meta-function -- uses CINFO. */
if_function:procedure expose cinfo. gerrs. info.  recids. defdir_up  update_verb 
parse arg a1a


parse var a1a a1 '(' opts0 ')' ;
a1=strip(translate(a1))
opts=strip(translate(opts0))


if wordpos(a1,'LENGTH LINES SUBSTR POS SUBWORD WORDPOS')>0 then do
   select
     when a1='LENGTH' | a1='LINES'  then avar=opts
     when a1='SUBSTR' | a1='SUBWORD' then parse var opts avar ',' .
     when a1='POS' | a1='WORDPOS' then parse var opts .','avar ',' . 
     otherwise return ''
   end
   if avar='' then return ''
   vopts='!'||strip(translate(avar))
   goo=cinfo.vopts
   if cinfo.vopts.!TYPE='F' then do              /* read file */
       parse var goo . '=' valfile . ; valfile=strip(valfile)
       if abbrev(valfile,'\')=0 & pos(':',valfile)=0 then do     /* relative filename */
         valfile=defdir_up||valfile
       end
       goo=sre_read_file(valfile,2,2,1)
   end
   goo2=do_var_function(a1,goo,opts0)
   return goo2
end  

/* if here, not a "var-function" type of if-condition */

select 

when a1='COUNT' then do
  parse upper var opts icount ',' isblock
  if icount='' then do
      fpp=note_error('E','No count-size in COUNT(): 'opts)
      return ''
  end 
  icount=strip(icount); isblock=strip(isblock)
  bnum=info.!!count
  if isblock='B' then bnum=info.!!count_block
  if icount<1 then do
      fpp=note_error('E','Bad count-size in COUNT(): 'opts)
      return ''
  end 
  jj=1+trunc((bnum-0.1)/icount)

  return jj
 
end

when a1='DATETIME' then do
  parse var opts opts0 ',' avarname
  if avarname<>'' then do
     vv='!'||strip(translate(avarname))
     atime0=cinfo.vv
     if datatype(atime0)<>'NUM' then do
        fpp=note_error('E','Non-numeric time ('||atime0||') for: '||opts)
        return ''
     end
  end
  else do
     atime0=cinfo.!!ID
  end
  goo2=do_var_function(a1,atime0,opts)
  return goo2
end


when a1='STATUS' then do
  vv='!$STATUS'
  astatus=cinfo.vv
  parse var opts svar ',' dfmt ','tfmt
  svar=strip(svar)
  useit=''
  do until astatus=''
           parse var astatus a1 astatus
           parse var a1 a1a '=' a1b
           if strip(a1a)<>svar then iterate
           useit=a1b
           leave
  end 
  if svar='CMT_CNT' then return useit
/* break the nnn.ffff date into components */
  if dfmt='' then dfmt='N'; if tfmt='' then tfmt='N'
  if dfmt=0 then
          oo1=''
        else
           oo1=fig_sysvar('$DATE',dfmt,useit)  
  if tfmt=0 then
           oo2=''
        else
           oo2=fig_sysvar('$TIME',tfmt,useit)  
  return oo1||' '||oo2
end


otherwise do
   fpp=note_error('E','Unknown Meta-variable: '||a1a)
   return ''
end

end

/* should never get here */


/****************/
/* do a comparison of two values, return 0 (IF is true, so IF is ignored met) or 1 
  (IF is false, so skip to next /IF) */

check_if_failed:procedure expose gerrs.  recids.  update_verb 
parse upper arg val1,cond,val2
val1=strip(val1); val2=strip(val2)
cond=strip(translate(cond))

doif=1          /* assume if condition failed to be true */
select
         when cond='EQ' then do
            if val1=val2 then doif=0 
         end
         when cond='NE' then do
            if val1<>val2 then doif=0 
         end
         when cond='GE' then do
            if val1>=val2 then doif=0 
         end
         when cond='GT' then do
            if val1>val2 then doif=0 
         end
         when cond='LE' then do
            if val1<=val2 then doif=0 
         end
         when cond='LT' then do
            if val1<val2 then doif=0 
         end
         otherwise do
            fpp=note_error('E','Unknown condition in IF statement ')
         end
end  /* select */
return doif



/************/
/* get first and last record to process */
find_i1_i2:

call parse_options(opt1)

/* RANGE or DATERANGE specified, and sort variable selected, error. */
if info.!sortit_name<>'' then do
  if ropts.!AGE<>'' | ropts.!DATERANGE<>'' then do
    fpp=note_error('E','SORT selected, but AGE or DATERANGE specified in REPEAT: 'opt1)
    return ''
  end
end


/* check for NONE, RANGE, ID, AGE, or DATERANGE -- RANGE trumps AGE trumps DATERANGE */
select

  when ropts.!RANGE<>'' then do
     parse var ropts.!RANGE i1 '-' i2
     i1=strip(i1) ; i2=strip(i2)
     if i1='' then i1=1
     if i2='' then i2=recids.0
  end
  when ropts.!AGE<>'' then do   /* search from recids.0 down, stop when age is exceeded */
      nowtime=sre_gmt(,,'JS',0)
      daage=ropts.!AGE
      if datatype(daage)<>'NUM' then daage=1    /* default of 1 day */
      target=nowtime-daage
      i2=recids.0 ; i1=1                /* assume all records */
      do jm=recids.0 to 1 by -1         /* look for an older id */  
           if recids.jm.1<target then do  
             i1=jm+1
             i2=recids.0
             leave
           end 
      end
  end
  when ropts.!ID<>'' then do

     if strip(translate(ropts.!ID))='NONE' then do
        i1=1 ; i2=1 ; asked_id7=0
     end
     else do
          hh=sre_cache_data(info.!series,'FIND',strip(ropts.!ID ))
          parse var hh astat aid7 i1
          if strip(astat)<>0 then do
             return '<!-- Error. Bad ID in REPEAT: '||ropts.!ID|| '--> '
           end 

          vv='!'||strip(Translate(aid7))   /* check for deselection */
          if recids.vv='' then do
             return ' <!-- Error. No such ID (possibly deselected) in REPEAT: '||ropts.!ID|| '--> '
          end 
  
          i1=1 ; i2=1
          asked_id7=strip(aid7)
     end
  end

  when ropts.!DATERANGE<>'' then do
      parse var ropts.!DATERANGE d1 '-' d2 ; d1=strip(d1) ; d2=strip(d2)

/* convert date range to cms format */
      if pos('.',d1)>0 then do
          sdate=d1              /* use this "native" value */
      end
      else do           /* convert from yyyymmdd:hh:mm to dddd.ffff */
         parse var d1 d1d ':' d1t
         d1j=dateconv(d1d,'S','B')
         if d1j='' then do              /* bad format, so use first */
             sdate='-1'            /* a hack to force "use first record  */
         end
         else do
           parse var d1t hh ':' mm ; 
           if datatype(hh)<>'NUM' then hh=0 
           if datatype(mm)<>'NUM' then mm=0 
           mins=(hh*60)+mm 
           frac=mins/1440
           tmp=sre_gmt(,d1j,'JS',0)
           sdate=tmp+frac
        end
      end

      if pos('.',d2)>0 then do
          edate=d2              /* use this "native" value */
      end
      else do           /* convert from yyyymmdd:hh:mm to dddd.ffff */
         parse var d2 d2d ':' d2t
         d2j=dateconv(d2d,'S','B')
         if d2j='' then do              /* bad format, so use first */
             edate=99999999       /* a hack to force "use last  record  */
         end
         else do
           parse var d2t hh ':' mm ; 
           if datatype(hh)<>'NUM' then hh=23
           if datatype(mm)<>'NUM' then mm=59.9
           mins=(hh*60)+mm 
           frac=mins/1440
           edate=sre_gmt(,d2j,'JS',0)+frac
        end
      end
/* now find top and bottom of range */
      i1=recids.0+1 ; i2=0   /* default is do nothing */
      if edate>sdate then do
         nowtime=sre_gmt(,,'JS',0)
         do jm=1 to recids.0                /* look for an older id */  
           if recids.jm.1<sdate then iterate
           i1=jm
           i2=recids.0          /* assume we'll use most recent */
           do jm2=i1 to recids.0
              if recids.jm2.1>edate then do
                 i2=jm2-1
                 leave
              end
           end             /* jm2 */
           leave
        end             /* jm */
     end     /* edate>sdate */

  end
  otherwise do
      i1=1
      i2=recids.0
  end
end

return 1

/******************/
/* obtain record ids of all records in dataseries. save to recids. */
load_recids:procedure expose recids. info. def_select def_selectval gerrs.  update_verb 

parse arg usesort,useorder,atype,no_shortcut,no_select,quiet


doing_comment=0
if info.!COMMENTS_MODE=1 then doing_comment=1
if doing_comment=1 then do              /* this is sort of a hack */
  usesort0=usesort              /* use these to sort AFTER selection of comments */
  useorder0=useorder
  atype0=atype
  usesort='$RECID'    /* sort on this to pull out comments for def_selectval record */
  useorder='A'
  atype='I'
  DEF_SELECT='SORT'

end

/* no_shortcut=1 means "don't do the "no change in sort instructions" shortcut 
   no_select=1 means "do NOT deselect records

error messages
      1  = Data series is not open
        2  = Bad range (misordered)
        3  = Bad range (smallest<1, or largest>#records in series)
        5  = Record could not be read
        8  = Bad data series name
*/

/* special case: extract 1 record */
if strip(usesort)=1 then do
  aid7=strip(translate(def_selectval))
  if aid7='' then do
          fpp=note_Error('E','No such record id: '||def_selectval)
          recids.0=0
          return 1
  end
  astat=sre_cache_data(info.!series,'FIND ID7 ',aid7,'*')  
  if abbrev(astat,'0')=0 then do
          fpp=note_Error('E','Bad record id: '||astat)
          recids.0=0
          return 1
  end 
  parse var astat . '0d0a'x  allval
  drop recids. ; recids.=''
  recids.0=1
  recids.1=aid7
  parse var allval recids.1.1 '09'x .
/*  recids.1.1=strip(uid) */
  recids.1.3=allval             /* .3 is ONLY used when usesort=1 */
  v7='!'||aid7
  recids.v7=1
  return 1
end  /* of special case */

if usesort='' then usesort='!ID'
if wordpos(strip(translate(usesort)),'!ID '||info.!series_sortables)=0 then do
              fpp=note_Error('E','No such sort variable: '||usesort','info.!series_sortables)
              recids.0=0
              return 1
end

useorder=strip(translate(useorder))
if useorder='' then useorder='A'
if pos(useorder,'AD')=0 then do
          fpp=note_Error('E','No such sort order: '||useorder)
          recids.0=0
          return 1
end

if quiet<>1 & usesort<>'' then fpp=note_error('M','Note -- sorting on variable: 'usesort||' (order='||useorder||')')

if no_shortcut<>1 & doing_comment=0 then  do          /* if sort order not changed, don't bother sorting*/
   wassort=info.!sorted ;
   wasorder=info.!sorted_order
   if usesort=wassort & wasorder=useorder then return 0     /* no change in sort order, so don't recreate index */
end

info.!sorted=usesort

sorder=useorder            /* default is ascending */
if sorder<>'D' then sorder='A'
info.!sorted_order=sorder

stype=atype
if stype<>'N' then stype='I' /* n=numeric, i=case insensitive */
info.!sorted_type=stype

aabs=''
if usesort<>'!ID' then do
  aabs=' abstract='||usesort
end

flen=16
if usesort<>'!ID' then flen=50

drop recids.
recids.=''


astat=sre_cache_data(info.!series,'EXTRACT ID7 ID  '||aabs,'1-'||info.!series_records)

parse var astat astat0 . '0d0a'x astat ; astat0=strip(astat0)

if astat0<>0 then do
   recids.0=0
   return 1
end

ii=0

doselect=0
if def_select<>'' & def_select<>0 & no_select<>1 then doselect=1

/* convert RECENT into a def_selectval */
if doselect=1 then do /* de-selection? */
  def_selectval=strip(Translate(def_selectval))
  if def_selectval='RECENT' then do   /* get most recent entry (last record) */
     igoo=lastpos('0d0a'x,astat)
     farf=substr(astat,igoo+2)
     parse var farf . recentid '09'x recentval
     select
        when def_select="SORT" then def_selectval=recentval
        when def_select='HOUR' then def_selectval=trunc(recentid*24)
        when def_select='DAY' | def_select='DAY2' then def_selectval=trunc(recentid)
        when def_select='WEEK' |def_select='WEEK2' then do
          tt=determine_week(recentid)
          parse var tt def_selectval ',' .
        end           
       when def_select='MONTH' | def_select='MONTH2' then do
          adate=trunc(recentid)+726467
          tt=strip(dateconv(adate,'B','S'))   /* 19990325 */
          def_selectval=left(tt,6,'0')
        end

        when def_select='YEAR' then do
          adate=trunc(recentid)+726467
          tt=strip(dateconv(adate,'B','S'))   /* 19990325 */
          def_selectval=left(tt,4,'0')
        end
        otherwise def_selectval=''
    end

    def_selectval=strip(translate(def_selectval))
    if quiet<>1 then     fpp=note_error('M','Selecting ('||def_select||')= '||def_selectval)
  end                   /* RECENT */
end             /* DEF_SELECT */

iat=1

do until astat=''               /* parse the extract of records */
   iat1=pos('0d0a'x,astat,iat)
   if iat1=0 then do            /* end of string */
      aline=strip(substr(astat,iat))
      astat='' 
   end 
   else do
     iget=(iat1-iat)
     aline=strip(substr(astat,iat,iget))
   end
   iat=iat1+2

   if abbrev(aline,'**')=1 then iterate
   parse var aline aid7 uid '09'x aaval  
   aid7=strip(aid7) ; aid=strip(uid)
   if aid7=uid then iterate             /* skip the cms meta-data records */

   if doselect=1 then do  /* de-selection? */
     select
 
        when datatype(def_select)='NUM' & def_select<>0 then nop  /* do numeric after sort */
        when def_select="SORT" then do        /* on current "sort" variable */
           aa2val=strip(translate(aaval))
           if def_selectval<>aa2val then iterate /* no match, don't use */
        end 
        when def_select='HOUR' then do
          ahr=trunc(uid*24)
          if ahr<>def_selectval then iterate
        end

        when def_select='DAY' | def_select='DAY2' then do
            if trunc(uid)<>def_selectval then iterate
        end

        when def_select='WEEK'  | def_select='WEEK2' then do
/* Example: input--  29 sep 03      (this is a Monday)
            output --  717,29 Sep 2003,28 Sep 2003,4 Oct 2003 */
          tt=determine_week(uid)
          parse var tt wnum ',' .
          if wnum<>def_selectval then iterate
        end

       when def_select='MONTH' | def_select='MONTH2' then do
          adate=trunc(uid)+726467
          tt=strip(dateconv(adate,'B','S'))   /* 19990325 */
          tt=left(tt,6,'0')
          if tt<>def_selectval then iterate
        end

        when def_select='YEAR' then do
          adate=trunc(uid)+726467
          tt=strip(dateconv(adate,'B','S'))   /* 19990325 */
          tt=left(tt,4,'0')
          if tt<>def_selectval then iterate
        end
        otherwise do
           iterate      /* unknown condition -- drop it */
        end
     end                /* select */ 
   end          /* def_select */

   ii=ii+1
   if usesort='!ID' then
      recids.ii=left(aid7,10,' ')||left(uid,14,' ')
   else
      recids.ii=left(aid7,10,' ')||left(uid,14,' ')||left(aaval,min(50,length(aaval)),' ')
end

recids.0=ii

if recids.0=0 then do
   foo=note_error('W','0 records (after de-selection)')
   return 0  /* don't try to sort empty list */
end

/* now sort them */

f11=11
if usesort<>'!ID' then f11=25

foo=arraysort('recids',1,recids.0,f11,flen,sorder,stype)
if foo=0 then do
   foo=note_error('E','Problem sorting records in CMS dataseries')
   return 0
end

jj0=0
do jj=1 to recids.0                     /* store id7 in linear list, and create id7 lookup */
  if datatype(def_select)='NUM' & def_select<>0 then do
       aval=trunc( (jj-0.1)/def_select)
       if aval<>def_selectval then iterate
   end
   abb=recids.jj
   jj0=jj0+1
   recids.jj0.1=strip(substr(abb,11,13))    /* user id */
   recids.jj0=left(abb,7)         /* id7 */
   if usesort<>'!ID' then recids.jj0.2=substr(abb,25)
   vv='!'||strip(translate(recids.jj0))
   recids.vv=jj0
end 
recids.0=jj0

if doing_comment=0 then return 1

do jj=1 to recids.0
   ape='!'||recids.jj ; ape2=recids.ape
end /* do */

/* secondary sort (needed if comment processing */
if usesort=usesort0 & useorder=useorder0 & atype=atype0 then return 1 /* no second sort */
if recids.0<2 then return 1             /* no point in sorting 1 element */

do kk=1 to recids.0
   if usesort0='!ID' then
     doit.kk=left(kk,10,' ')||recids.kk.1
   else
     doit.kk=left(kk,10,' ')||recids.kk.2
   nurec.kk=recids.kk           /* and copy recids to nurec */
   nurec.kk.1=recids.kk.1
   nurec.kk.2=recids.kk.2
   vv='!'||recids.kk
   nurec.vv=recids.vv
end 
nurec.0=recids.0
doit.0=nurec.0

drop recids.
recids.=''
/* sort on 11th char (1-10, the kk'th, travel along ) */
foo=arraysort('doit',1,doit.0,11,25,useorder0,atype0)  /* sort */
do mm=1 to nurec.0                     /* and copy nurec to recids */
   kk=strip(left(doit.mm,10))
   recids.mm=nurec.kk
   recids.mm.1=nurec.kk.1
   recids.mm.2=nurec.kk.2
   vv='!'||nurec.kk
   recids.vv=nurec.vv
end 
recids.0=nurec.0
do jj=1 to recids.0
   ape='!'||recids.jj ; ape2=recids.ape
end /* do */
return 1



/***************/
/* parse a cms elements 'options' (as in  <? xx options >)
   returns values in ropts. array, where
     ropts.0=optname list (space delimited)
     ropts.!o1 , where o1 to oN are the options listed in ropts.0, and ropts.!o1 contains "o1's" value
     ropts.!0=concatenation of optname=optval for optnames that do not begin with ?
     ropts.!1=concatenation of optname=optval for optnames that begin with ?
     ropts.!2=concatenation of optname=optval for optnames that begin with # (# is stripped)
     ropts.!3=concatenation of optname=optval for optnames that do not begin with ?, #, or &
    (ropts.!0 will include items in ropts.!2)

Optlist can contain:  opt1  ?opt2 opt1=abv ?opt1a=def opt2="ab cd" ?opt3="ef gh"
*/
parse_options:procedure expose ropts. gerrs.  recids.  update_verb 
parse arg optlist

optlist=strip(optlist)
ropts.=''

optlist=translate(optlist,' ','0d0a009'x)
do forever
   optlist=strip(optlist)
   if optlist='' then leave
   isp=pos(' ',optlist) 
   addspace=''
   if isp=0 then do
      a1=optlist ;  optlist=''
   end 
   else do
      a1=left(optlist,isp-1)
      optlist=substr(optlist,isp)
   end 
   
   if pos('=',a1)=0  then do  /* perhaps space between optname and = sign ? */
       optname=a1
       if abbrev(strip(optlist),'=')=1 then do
          parse var optlist . '=' tt
          TT=STRIP(TT)
          PARSE VAR TT OPTVAL1 ' ' optlist
          addspace=' '
      end
      else do
          optval1=''
     end
   end 
   else do
       parse var a1 optname '=' optval1
       IF OPTVAL1='' then DO
          if abbrev(strip(optlist),'"')=1 then do
              parse var optlist . '"' optlist
              optval1='"'
          end 
       END
   end 
   optname=strip(translate(optname))
   gotq=0
/* if start of optval1 is ", then look for terminating " */
   optval1=strip(optval1,'l',' ')
   if abbrev(optval1,'"')=1 then do
      gotq=1
      optval1=substr(optval1,2)
      if right(optval1,1)<>'"' then do /* doesn't end with quote, get from optlist */
         parse var optlist a2 '"' optlist
         optval1=optval1||addspace||a2
      end
      else do
          optval1=strip(optval1,'t','"')
      end 
   end
  aa='!'||optname   
  ropts.0=ropts.0||' '||optname
  ropts.aa=optval1 

  if abbrev(optname,'?')=0 & abbrev(optname,'#')=0 & abbrev(optname,'&')=0 then do
     if optval1='' then do
        ropts.!3=ropts.!3||' '||optname||'=""'
     end
     else do
        if gotq=0 then
           ropts.!3=ropts.!3||' '||optname||'='||optval1
        else
           ropts.!3=ropts.!3||' '||optname||'="'||optval1||'"'
     end
  end

  if abbrev(optname,'?')=0 then do
     if optval1='' then do
        ropts.!0=ropts.!0||' '||optname||'=""'
     end
     else do
        if gotq=0 then
           ropts.!0=ropts.!0||' '||optname||'='||optval1
        else
           ropts.!0=ropts.!0||' '||optname||'="'||optval1||'"'
     end
     if abbrev(optname,'#')=1 then do
       oopt2=substr(optname,2)
       if optval1='' then do
           ropts.!2=ropts.!2||' '||oopt2||'=""'
       end
       else do
          if gotq=0 then
              ropts.!2=ropts.!2||' '||oopt2||'='||optval1
          else
             ropts.!2=ropts.!2||' '||oopt2||'="'||optval1||'"'
       end
     end

  end
  else do
     if optval1='' then do
        ropts.!1=ropts.!1||' '||optname
     end
     else do
        if gotq=0 then
           ropts.!1=ropts.!1||' '||optname||'='||optval1
        else
           ropts.!1=ropts.!1||' '||optname||'="'||optval1||'"'
     end
  end
end 

/* Oo='opt1  ?opt2 opt1b= abv ?opt1a=def opt2 ="ab cd" ?opt3="ef gh"' */
return 1


/***************/
/* lookup value using SRE_CACHE_DATA, but check to see if stored in a local cache 
 returns stat_code '0d0a'x value  [emulate sre_cache_data]
*/

SRE_CACHE_DATA_LOCAL:procedure expose cshdata. info. gerrs. recids.   update_verb 


parse arg aseries,is7,findid,svarname
atonce=500                /* number or records to read per gulp */
svarname=strip(Translate(svarname))

id7=findid

/* This works with the ID7.
   If user-id is being sought, then find its ID7.
   Note: user-id FINDS are infrequent, so we use this non-optimized code */

if pos('ID7',translate(is7))=0 then do   /* need to find id7 for this id */
   aline=sre_cache_data(aseries,'FIND ',findid,'%ID')
   if svarname='!ID' then  return aline  
   if abbrev(aline,'0')=0 then return aline
   parse var aline . id7 . '0d0a'x  .
   id7=strip(translate(id7))
   vv0='!'||id7                 
   irecid=cshdata.!1 
end
else do
  id7=strip(translate(id7))
  vv0='!'||id7      
  irecid=recids.vv0               /* okay record? */
  if irecid='' then  do
     fpp=note_error('E','No such record: 'id7','findid)
     return ''              /* '' signals no such id */
  end
end


/* cshdata empty? If so, read the first atonce records (or less) */
if cshdata.!0='01'x then do
  ifrom=1
  ito=min(ifrom+atonce,info.!series_records )
  zz= read_cshdata(ifrom,ito)
end

/* 0 #RECORDS 
   ID7 , id '09'x  packed_values '09'x  packed_index '09'x sortable1 '09'x .. */

vv1='!'||Svarname

astat=cshdata.vv0.vv1
if astat<>'01'x then do  /* got a cached value */
 return '0 '||'0d0a'x||astat
end

/* astat='01'x. Perhaps record in cache, but no such variable */
if cshdata.vv0<>'01'x then do
   fpp=note_error('M','No such variable: 'svarname','findid)
   return ''    
end

/* not in cache, but record does exist. Get next atonce records near this record */
if datatype(irecid)<>'NUM' then irecid=cshdata.!1

ifrom=max(1,trunc(4+irecid-(atonce/2)))
ito=min(ifrom+atonce,info.!series_records)
acall=read_cshdata(ifrom,ito)

astat=cshdata.vv0.vv1
if astat='01'x then do  /* still not in cache, so just get it from sre_cache_data */
  if wordpos(svarname,info.!series_sortables)>0 then do  /* read special variables */
     aline=sre_cache_data(aseries,'FIND ID7',id7,svarname)
     return aline
  end
  if svarname='!ID' then do
     aline=sre_cache_data(aseries,'FIND ID7',id7,'%ID')
     return aline
  end

/*else, need to unpack (a packed var, or a *) */
  aline=sre_cache_data(aseries,'FIND ID7',id7,'*')
  uval=unpack_rec(aline,svarname)
 
  return uval
end
else do                 /* read_cshdata got it */
   return '0 '||'0d0a'x||astat
end


/***************************/
/* unpack values (after stripping id7 and user_id from beginning of a sre_cache_data(xx) record */
unpack_rec:procedure expose info.  update_verb 
parse arg aline0,svarname

parse var aline0 a7 . '0d0a'x aline
parse var aline  aid '09'x vals 

ssv=strip(Translate(svarname))
vssv='!'||ssv

newvals=''
if svarname='*' then newvals=aid

if svarname='!ID' then return '0 '||'0d0a'x||aid


select  

    when  wordpos(ssv,info.!series_sortables)>0 then do   /* it is a packed variable */
        parse var aline uid '09'x . '09'x . '09'x svals
        do k1=1 to info.!spvars.0
           ww1=info.!spvars.k1
           parse var Svals aval '09'x Svals
           if ww1<>vssv then iterate
           return '0 '||'0d0a'x||aval
        END
        return '1 '               /* error if here */
    end

     when  wordpos(ssv,info.!series_pack)>0 then do   /* it is a packed variable */
        parse var aline . '09'x cvals '09'x cindex '09'x .
        if cindex='' then return '2 '             /* an error! */
        inrec=1
        do mm2=1 to info.!cmpvars.0
          blen=strip(word(cindex,mm2))
          av=info.!cmpvars.mm2
          if av<>vssv then do
             inrec=inrec+blen
             iterate
          end
          aval=substr(cvals,inrec,blen)
          return '0 '||'0d0a'x||aval
       end
       return 1         /* should never get here */
     end

    when ssv='*' then do                /* all values, starting with id */
        inrec=1
        parse var aline uid '09'x cvals '09'x cindex '09'x svals
        
        do mm2=1 to info.!cmpvars.0
           blen=strip(word(cindex,mm2))
           apack=info.!cmpvars.mm2
           vs.apack=substr(cvals,inrec,blen)
           inrec=inrec+blen
        end
        do k1=1 to info.!spvars.0
           parse var Svals aval '09'x Svals
           apack=info.!spvars.k1
           vs.apack=aval
        END
        do jj=1 to info.!bothvars.0
           apack=info.!bothvars.jj
           newvals=newvals||'09'x||vs.apack
        end
        return '0 '||'0d0a'x||newvals
    end
    OTHERWISE DO
       RETURN 1
    END
end

/*   should never get here */
   
/******************/
/* re fill cshdata. array */
read_cshdata:Procedure expose  cshdata. info.  recids. gerrs.

parse arg ifrom,ito
aseries=info.!series
drop cshdata.
cshdata.='01'x


/* Special case: only 1 record (typically, ARCHIVE in REPEAT block ) */
if recids.1.3<>'' then do
  recs=recids.1||','||recids.1.3
end
else do
  recs=sre_cache_data(aseries,'EXTRACT ',ifrom||'-'||ito)
  parse var recs astat recread '0d0a'x recs ; astat=strip(astat)
  if astat<>0  then do         /* error, give up */
     return astat        /* return value is ignored */
  end
end

cshdata.!0=ifrom
iat=ifrom-1

/* read values from each record, store in cshdata */

do until recs=''
       parse  var recs aline  '0d0a'x recs 
       parse var aline a7 ',' arfy    ; a7=strip(a7)
       if pos('*',a7)>0 then iterate
       if info.!cmpvars.0=-1 then    /* -1 signals "comment processing */
           parse var arfy uid '09'x svals
       else
           parse var arfy uid '09'x cvals '09'x cindex '09'x svals
       a7=strip(translate(a7))
       if a7=strip(translate(uid)) then iterate

       va7='!'||a7
       if recids.va7='' then iterate  /* possibly deselected */

       iat=iat+1
       cshdata.iat=a7
       cshdata.va7=iat
       vid='!!ID'
       inrec=1

       cshdata.va7.vid=uid

        do mm2=1 to info.!cmpvars.0
           blen=strip(word(cindex,mm2))
           apack=info.!cmpvars.mm2
           cshdata.va7.apack=substr(cvals,inrec,blen)
           inrec=inrec+blen
        end
        do k1=1 to info.!spvars.0
           parse var Svals aval '09'x Svals
           apack=info.!spvars.k1
           cshdata.va7.apack=aval
        END
        newvals=uid
        do jj=1 to info.!bothvars.0
           apack=info.!bothvars.jj
           newvals=newvals||'09'x||cshdata.va7.apack
        end
        vast='!*'
        cshdata.va7.vast=newvals
        cshdata.!1=iat
end
return iat



/***************/
/* interpret a code block  using REXXRUN */
make_interpret_string:procedure expose CINFO. usevars useval gerrs.  recids.  update_verb 
parse arg cmdlist,cmsvars,useonly


/* first, build a string containing the values of CINFO., with an index */

if useonly<>0 then do
  useval=cmsvars 
  usevars='!VARNAMES '||length(cmsvars)||' '
  do mm=1 to words(CINFO.0)
    aww=strip(word(CINFO.0,mm))
    vaww='!'||aww
    vval=CINFO.vaww
    useval=useval||vval 
    usevars=usevars||aww||' '||length(vval)||' '
  end 
end

if useonly=1 then return 1      /* just make the first two arguments for rexxrun */

crlf='0d0a'x

aa=''
aa=aa||'parse arg varlist,vallist,userargs,nochange,procid ; '||crlf 
aa=aa||'INFO.="" ; '||crlf
aa=aa||'tlist=varlist ; iat=1 ;'||crlf
aa=aa||'do until tlist="" ; '||crlf
aa=aa||'   parse var tlist aww vlen tlist ; vname=strip(aww) ;vlen=strip(vlen) ;'||crlf
aa=aa||'   if vlen=0 then iterate ;'||crlf
aa=aa||'   v1=substr(vallist,iat,vlen) ; iat=iat+vlen; '||crlf
aa=aa||'   vaww="!"||aww  ;'||crlf
aa=aa||'   INFO.vaww=v1 ; INFO.0=INFO.0||aww||" " ;'||crlf
aa=aa||'end  ;'||crlf

aa=aa||'okay=1 ; ares=doit1(userargs,procid) ; ' ||crlf 
aa=aa||' if okay=0 then return "0,"||ares ; '||crlf   /* error: return 0,error_code error_line */
/* this part is for returning actual results of the INTERPRETable code */

/* first, the results returned by the CALLed procedure */
aa=aa||'rhd="return "||length(ares) ;'||crlf 
aa=aa||'ret1=ares ;'||crlf 

/* might have some INFO. to change */
aa=aa||'dd=cvtails("info",datails) ;'||crlf 
aa=aa||'do mm=1 to datails.0 ;'||crlf       /* check each tail */
aa=aa||'   e1=datails.mm ;'||crlf 
aa=aa||'   if e1="0" then iterate ;'||crlf      /* don't change several types of info. variables */
aa=aa||'   e1a=substr(datails.mm,2) ;'||crlf 
aa=aa||'   if abbrev(e1a,"!")=1 then iterate ;'||crlf
aa=aa||'   if wordpos(e1a,nochange)>0 then iterate ; '||crlf    
/* if here, include this is a changeable INFO. variable, include it's value in the return stuff */
aa=aa||'   v1=INFO.e1 ;'||crlf 
aa=aa||'   lv1=length(v1) ;'||crlf 
aa=aa||'   rhd=rhd||" "||e1||" "||lv1 ;'||crlf 
aa=aa||'   ret1=ret1||v1 ;'||crlf 
aa=aa||'end ;'||crlf 
aa=aa||'return "1,"||rhd||"0d0a"x||ret1 ;'||crlf    /* note that "1," signals "success" */

aa=aa||'doit1:procedure expose INFO. okay  ; '||crlf 
aa=aa||' signal on syntax name _cmserr_ ; signal on error name _cmserr_ ; '||crlf
aa=aa||cmdlist||crlf 
aa=aa||crlf||'; return " " '||crlf  /* default is return nothing */
aa=aa||'_cmserr_:  ares=rc||" "||sigl ; okay=0 ; return ares ;'||crlf

return aa



/*******************/
/* read and process a CMS procedure file */
read_interp_file:procedure expose  gerrs. info.    update_verb 
parse arg dafile,cmsdir

dafile=strip(dafile)
if pos('.',dafile)=0 then dafile=dafile||'.CMS'

if pos(':',dafile)=0 then do            /* drive info present, treat as fully qualified */
   if abbrev(dafile,'\')=0 then do      /* begins with \, treat as fully qualified */
       if pos('\',dafile)=0 then do   /* NO path, assume its in the CMS\LIB dir */
          dafile=cmsdir||'\LIB\'||dafile
       end
       else do
          dafile=cmsdir||'\'||dafile
       end
   end 
end

aa=stream(dafile,'c','query size')
if aa='' | aa=0 then do
    fpp=note_Error('E','No such CMS procedure file: '||dafile)
    return ' '          /* no such file */
end 

astring=sre_read_file(dafile,2,2,1)


return astring


/*********************/
/* record error ? */
note_error: procedure expose gerrs. info.  update_verb 
parse arg severe,message,noloc

severe=strip(translate(severe))     /* severity (most to least) E W M  (error, warning, message) */

/*
gerrs.=''
gerrs.!DO=debug   /* 0 or ''=do not report errors, 1=major errors, 2=verbose errors, 3=too verbose */
gerrs.0=0 gerrs.!W=0 gerrs.!E=0 gerrs.!M=0
*/

ii=gerrs.0+1
gerrs.ii=message
gerrs.ii.!LINE=gerrs.!CURRENT
gerrs.ii.!noloc=noloc
gerrs.ii.!TYPE=severe
gerrs.0=ii

if wordpos(severe,'W M E')=0 then do
  call sre_pmprintf(' ERROR calling note_error ('severe')')
  message=severe ; severe='E'
end 

vv='!'||severe
gerrs.vv=gerrs.vv+1

aret=''
if gerrs.!DO>1 then aret='<!-- '||message||' -->'

return aret


/*********************************************************************************/
/* COmpute first and last day of a week, given Normal Date (i.e; 27 Aug 2002), or
   given SRE_GMS 'JS' style date, or a 'J' (rexx 'B') date.
   Returns:   
          weeknumber,thisdate, thisweek_sunday_date, thisweek_saturday_date
   where:
        weeknumber: the week since Jan 7, 1990
        each _date is in N format (day month year).
   Note that for N format input, a  year <70 is assumed to be in the 1900s,
   a year gt 70 (and less then 200) is assumed to be in the 2000s,
   otherwise the year is read as entered.


   Example: input--  29 sep 03      (this is a Monday)
            output --  717,29 Sep 2003,28 Sep 2003,4 Oct 2003

 If no argument, use today's date

*/
determine_week:procedure
parse arg a

if a='' then a=date('b')

wd.!SUNDAY=0
wd.!MONDAY=1
wd.!TUESDAY=2
Wd.!WEDNESDAY=3
Wd.!THURSDAY=4
WD.!FRIDAY=5
WD.!SATURDAY=6

if datatype(a)='NUM' then do
   if a<40000 then do           /* JS style */
        b=a+726467.
   end 
   else do
     b=a
   end 
end
else do
   parse var a dy mn yr
   if yr<70 then yr=2000+yr
   if yr>70 & yr < 200 then yr=yr+1900
   mn=left(mn,3)
   a=dy' 'mn' 'yr
   b=translate(dateconv(a,'N','B'))
   if b='' then return ''               /* bad format */
end
b=trunc(b)
chkdate=dateconv(b,'B','N')
wd=translate(dateconv(b,'B','W'))
vv='!'||wd
inweek=wd.vv
newb=b-inweek
wd2=dateconv(newb,'B','W')
ndate=dateconv(newb,'B','N')
ndate2=dateconv(newb+6,'B','N')

/* now compute week #, with week starting on Sunday, 7 Jan 1990, as week # 1 */
jsday=(b-726467)-5

nwk=1+((jsday-.5)%7)
return nwk||','||chkdate||','||ndate||','||ndate2



/******************************/
/* check access rights. Return 1 if okay. Otherwise, return a string to be
  sent back to SREhttp/2 (an authorization response string) */
check_access:procedure expose privset host_nickname idinfo uri  aiter  update_verb 
parse  upper arg owners,series

if owners='' then return 1      /* no owners listed, means open access */

tpr=translate(privset)
pneeded=space(translate('CMS:'||strip(series)),0)
if wordpos(pneeded,tpr)>0 then return 1          /* explicit privilege for this */
if wordpos('SUPERUSER',tpr)>0 then return 1


/* see if a cookie exists */
look_cook='CMS_'||series
aa=sreh2_get_cookie(lookcook)
if aa<>''  then return 1

if aiter=''  then do            /* first time, so recheck privs */
  return 'PRIVS   Access to this dataseries is restricted to specific users '
end 

/* otherwise, see if a username has been sent  */
username=''
authh=sre_reqfield('Authorization')
if authh<>'' then username=sre_extract_username(authh) 
if username='' then do             /* no username available */
  return  'AUTH  Access to this dataseries is restricted to specific users '
end

/* is this one of the OWNERS? */
if pos('*',owners)=0 then do            /* * means "any user" */
if wordpos(translate(username),translate(owners))=0 then do
   return 'AUTH  Access to this dataseries is restricted to specific users '
end
end

/* One of the OWNERS, now check password */
stuff=sreh2_get_userinfo(authh,host_nickname,idinfo)
if stuff='' then do
  return 'AUTH   Access to this dataseries is restricted to specific users. Please enter a new username/password.'
end

parse var uri thisaddon '?' .
jj=lastpos('/',thisaddon)
pp=left(thisaddon,jj)

foo=sreh2_set_cookie(lookcook,1,'30m',pp,,1)

/* if here, legit user.  Record a cookie */
return 1
 

/************/
/* convert from CMS id  (JS time stamp) to "N" time stamp */
fig_time:procedure
parse arg atime0

if datatype(atime0)<>'NUM' then return ''

parse var atime0 adate1 '.' atime
adate=adate1+726467
atime='0.'||atime

dd=dateconv(adate,'B','N')

/* else, convert to hr min second */
zz=digits()
numeric digits 16
secs=trunc(atime*86400)
tmp0=secs/3600
parse var tmp0 hour '.' tmp1 ; if tmp1='' then tmp1=0
hour=right(hour,2,'0')
tmp1='0.'||tmp1
tmp2=tmp1*60
parse var tmp2 minute '.'tmp3 ; ; if tmp3='' then tmp3=0
minute=right(minute,2,'0')
tmp3='0.'||tmp3
tmp3=tmp3*60 ; parse var tmp3 second '.' .
second=right(second,2,'0')
numeric digits zz

dd=dd||' '||hour||':'||minute||':'||second
return dd

/******************/
/* write to tempfile. But check for UPDATE "verb", in which case don't write. */
lineout1:

if update_verb<>0 then return 0   /* UPDATE_verb IS global */

parse arg aa

if aa='1a'x then do
   call lineout tempfile
   return 1
end 
call lineout tempfile, aa
return 2

